// ECDLeadFieldElekta() -- computes lead field matrix for realistic head shape for Elekta Vectorview sensors
//
//	S.E. Robinson (using information supplied by J. Nenonen)
//	MEG Core Facility
//	NIMH
//

#include <lfulib.h>
#include <samlib.h>

// global arrays
double	gradbasis[BASIS+1][3];
double	basis[BASIS+1];
int		CoeffFlag;						// flag for computing coefficients -- needed for changes in head frame

// ECDLeadFieldElekta()
void	ECDLeadFieldElekta(
	double		Org[3],					// ECD coordinate
	HeaderInfo	*Header,				// MEG dataset header information
	ChannelInfo	*Channel,				// Channel[M] -- MEG channel information
	HULL		*OrigHull,				// cortical hull model
	double		**L						// L[M][3] -- output ECD primary sensor lead field matrix
) {
	static HULL		Hull;				// local inflated hull copy so that we don't compute voxels on the hull boundary
	static INT_PNT	*detector;			// detector[S+1] -- integration points & forward solutions for each cardinal moment
	double			Bd[3][3];
	double			x[3];				// x' vector
	double			y[3];				// y' vector
	double			z[3];				// z' vector
	double			Vd[3];				// dipole location
	double			Vs[3];				// detector location
	static double	**coeffs;			// coeffs[BASIS+1][S+1] -- correction coefficients
	static double	**fieldori;			// fieldori[S+1][3]
	static double	**correction;		// correction[S+1][3]
	static double	**coil;				// coil[C][3]
	static double	**Bp;				// Bp[M][3] -- primary sensor lead field
	static double	Wg[] = {			// planar gradiometer integration weights (in order of integration points)
		0.125,		// if T/m, use 1. / (4. * 0.01669),
		0.125,
		0.125,
		0.125,
		0.125,
		0.125,
		0.125,
		0.125
	};
	static double	Wm[] = {			// magnetometer integration weights (in order of integration points)
		16. / 81.,
		25. / 324.,
		25. / 324.,
		25. / 324.,
		25. / 324.,
		10. / 81.,
		10. / 81.,
		10. / 81.,
		10. / 81.
	};
	register double	tmp;				// useful for just about everything!
	double			r;					// magnitude
	double			r2;					// magnitude^2
	int				c;					// coil index
	int				i;
	int				j;
	int				m;					// sensor index
	int				mm;					// alternate sensor index
	int				n;					// integration/detection point index
	int				found;				// found magnetometer
	int				v;					// vector index
	int				w;					// another vector index
	extern int		CoeffFlag;			// flag for computing coefficients -- needed for changes in head frame
	static int		C;					// number of coils
	static int		S = 0;				// number of integration points

	// one-time memory allocation
	if(S == 0) {

		// count integration points & coils
		for(mm=S=C=0; mm<Header->NumPri; mm++) {
			m = Header->PsIndex[mm];
			if(Channel[m].Flag == TRUE) {							// count only good sensors!
				if(Channel[m].Geom.MEGSensor.NumCoils == 2) {		// planar gradiometer
					C+= 2;											//	2 coils
					S+= 16;											//	8 integration points/coil
				} else {											// magnetometer
					C += 1;											//	1 coil
					S += 9;											// 	9 integration points/coil
				}
			}
		}

		// allocate memory
		if((detector = (INT_PNT *)malloc((size_t)(S + 1) * sizeof(INT_PNT))) == NULL)
			allocfailed("detector[]");
		if((coeffs = (double **)malloc((size_t)(BASIS + 1) * sizeof(double *))) == NULL)
			allocfailed("coeffs[]");
		for(i=0; i<=BASIS; i++)
			if((coeffs[i] = (double *)malloc((size_t)(S + 1) * sizeof(double))) == NULL)
				allocfailed("coeffs[][]");
		if((fieldori = (double **)malloc((size_t)(S + 1) * sizeof(double *))) == NULL)
			allocfailed("fieldori[]");
		for(i=0; i<=S; i++)
			if((fieldori[i] = (double *)malloc((size_t)3 * sizeof(double))) == NULL)
				allocfailed("fieldori[][]");
		if((correction = (double **)malloc((size_t)(S + 1) * sizeof(double *))) == NULL)
			allocfailed("correction[]");
		for(i=0; i<=S; i++)
			if((correction[i] = (double *)malloc((size_t)3 * sizeof(double))) == NULL)
				allocfailed("correction[][]");
		if((coil = (double **)malloc((size_t)C * sizeof(double *))) == NULL)
			allocfailed("coil[]");
		for(i=0; i<C; i++)
			if((coil[i] = (double *)malloc((size_t)3 * sizeof(double))) == NULL)
				allocfailed("coil[][]");
		if((Bp = (double **)malloc((size_t)Header->NumPri * sizeof(double *))) == NULL)
			allocfailed("Bp[]");
		for(i=0; i<Header->NumPri; i++)
			if((Bp[i] = (double *)malloc((size_t)3 * sizeof(double))) == NULL)
				allocfailed("Bp[][]");

		// make inflated hull for Nolte solution
		Hull.nv = OrigHull->nv;
		for(v=X_; v<=Z_; v++)
			Hull.Vo[v] = OrigHull->Vo[v];
		Hull.scale = OrigHull->scale;
		if((Hull.vertex = (FIELD *)malloc((size_t)(Hull.nv + 1) * sizeof(FIELD))) == NULL)
			allocfailed("Hull.vertex[]");
		for(i=1; i<=Hull.nv; i++)
			for(v=X_; v<=Z_; v++) {
				Hull.vertex[i].p[v] = OrigHull->vertex[i].p[v] + INFLATE * OrigHull->vertex[i].v[v];	// assumes that normals are all outward-pointing
				Hull.vertex[i].v[v] = OrigHull->vertex[i].v[v];
			}


		// on 1st call, assure that CoeffFlag is TRUE
		CoeffFlag = TRUE;
	}

	// compute integration points (one-time, unless sensor frame changes)
	if(CoeffFlag == TRUE) {
		for(mm=0; mm<Header->NumPri; mm++) {
			m = Header->PsIndex[mm];
			if(Channel[m].Flag == TRUE) {

				// compulsively make certain that the coil normal vector is unit length
				for(v=X_, r2=0.; v<=Z_; v++) {
					z[v] = Channel[m].Geom.MEGSensor.Coil[0].origin.v[v];
					r2 += z[v] * z[v];
				}
				r = sqrt(r2);
				for(v=X_; v<=Z_; v++)
					z[v] /= r;	// there now... I feel better already!

				// if planar gradiometer, determine gradiometer baseline unit vector
				if(Channel[m].Geom.MEGSensor.NumCoils == 2) {			// a planar gradiometer
					for(v=X_, r2=0.; v<=Z_; v++) {
						y[v] = Channel[m].Geom.MEGSensor.Coil[0].origin.p[v] - Channel[m].Geom.MEGSensor.Coil[1].origin.p[v];
						r2 += y[v] * y[v];
					}
					r = sqrt(r2);
					for(v=X_; v<=Z_; v++)
						y[v] /= r;

					// construct x-vector at right angle to y & z (y cross z)
					x[X_] = y[Y_] * z[Z_] - y[Z_] * z[Y_];
					x[Y_] = y[Z_] * z[X_] - y[X_] * z[Z_];
					x[Z_] = y[X_] * z[Y_] - y[Y_] * z[X_];

					// make x' a unit vector
					for(v=X_, r2=0.; v<=Z_; v++)
						r2 += x[v] * x[v];
					r = sqrt(r2);
					for(v=X_; v<=Z_; v++)
						x[v] /= r;

					// set up eight integration points in plane of each rectangular loop
					for(c=0; c<2; c++) {
						for(v=X_; v<=Z_; v++) {
							Channel[m].Geom.MEGSensor.Coil[c].B[0].p[v] = Channel[m].Geom.MEGSensor.Coil[c].origin.p[v] + x[v] * 0.00589 + y[v] * 0.00671 + z[v] * 0.0003;
							Channel[m].Geom.MEGSensor.Coil[c].B[1].p[v] = Channel[m].Geom.MEGSensor.Coil[c].origin.p[v] + x[v] * 0.00589 - y[v] * 0.00671 + z[v] * 0.0003;
							Channel[m].Geom.MEGSensor.Coil[c].B[2].p[v] = Channel[m].Geom.MEGSensor.Coil[c].origin.p[v] - x[v] * 0.00589 + y[v] * 0.00671 + z[v] * 0.0003;
							Channel[m].Geom.MEGSensor.Coil[c].B[3].p[v] = Channel[m].Geom.MEGSensor.Coil[c].origin.p[v] - x[v] * 0.00589 - y[v] * 0.00671 + z[v] * 0.0003;
							Channel[m].Geom.MEGSensor.Coil[c].B[4].p[v] = Channel[m].Geom.MEGSensor.Coil[c].origin.p[v] + x[v] * 0.01080 + y[v] * 0.00671 + z[v] * 0.0003;
							Channel[m].Geom.MEGSensor.Coil[c].B[5].p[v] = Channel[m].Geom.MEGSensor.Coil[c].origin.p[v] + x[v] * 0.01080 - y[v] * 0.00671 + z[v] * 0.0003;
							Channel[m].Geom.MEGSensor.Coil[c].B[6].p[v] = Channel[m].Geom.MEGSensor.Coil[c].origin.p[v] - x[v] * 0.01080 + y[v] * 0.00671 + z[v] * 0.0003;
							Channel[m].Geom.MEGSensor.Coil[c].B[7].p[v] = Channel[m].Geom.MEGSensor.Coil[c].origin.p[v] - x[v] * 0.01080 - y[v] * 0.00671 + z[v] * 0.0003;
						}
						Channel[m].Geom.MEGSensor.Coil[c].NumInt = 8;
						Channel[m].Geom.MEGSensor.Coil[c].Evaluated = TRUE;
					}

					// compute center of current gradiometer coil
					for(v=X_; v<=Z_; v++)
						Vs[v] = (Channel[m].Geom.MEGSensor.Coil[0].origin.p[v] + Channel[m].Geom.MEGSensor.Coil[1].origin.p[v]) / 2.;

					// find the magnetometer coil for this planar gradiometer (so we can re-use x, y, & z vectors)
					for(i=0, found=FALSE; i<Header->NumPri; i++) {
						j = Header->PsIndex[i];
						if(Channel[j].Geom.MEGSensor.NumCoils == 1 && Channel[j].Geom.MEGSensor.Coil[0].Evaluated == FALSE) {
							for(v=X_, r2=0.; v<=Z_; v++) {
								tmp = Vs[v] - Channel[j].Geom.MEGSensor.Coil[0].origin.p[v];
								r2 += tmp * tmp;
							}
							r = sqrt(r2);
							if(r < 0.005) {
								n = j;
								found = TRUE;
								break;
							}
						}
					}
					if(found == TRUE) {			// found magnetometer origin within 5 mm of gradiometer origin
						for(v=X_; v<=Z_; v++) {
							Channel[n].Geom.MEGSensor.Coil[0].B[0].p[v] = Channel[n].Geom.MEGSensor.Coil[0].origin.p[v] + z[v] * 0.0003;
							Channel[n].Geom.MEGSensor.Coil[0].B[1].p[v] = Channel[n].Geom.MEGSensor.Coil[0].origin.p[v] + x[v] * 0.00813 + y[v] * 0.00813 + z[v] * 0.0003;
							Channel[n].Geom.MEGSensor.Coil[0].B[2].p[v] = Channel[n].Geom.MEGSensor.Coil[0].origin.p[v] + x[v] * 0.00813 - y[v] * 0.00813 + z[v] * 0.0003;
							Channel[n].Geom.MEGSensor.Coil[0].B[3].p[v] = Channel[n].Geom.MEGSensor.Coil[0].origin.p[v] - x[v] * 0.00813 + y[v] * 0.00813 + z[v] * 0.0003;
							Channel[n].Geom.MEGSensor.Coil[0].B[4].p[v] = Channel[n].Geom.MEGSensor.Coil[0].origin.p[v] - x[v] * 0.00813 - y[v] * 0.00813 + z[v] * 0.0003;
							Channel[n].Geom.MEGSensor.Coil[0].B[5].p[v] = Channel[n].Geom.MEGSensor.Coil[0].origin.p[v] + y[v] * 0.00813 + z[v] * 0.0003;
							Channel[n].Geom.MEGSensor.Coil[0].B[6].p[v] = Channel[n].Geom.MEGSensor.Coil[0].origin.p[v] - y[v] * 0.00813 + z[v] * 0.0003;
							Channel[n].Geom.MEGSensor.Coil[0].B[7].p[v] = Channel[n].Geom.MEGSensor.Coil[0].origin.p[v] + x[v] * 0.00813 + z[v] * 0.0003;
							Channel[n].Geom.MEGSensor.Coil[0].B[8].p[v] = Channel[n].Geom.MEGSensor.Coil[0].origin.p[v] - x[v] * 0.00813 + z[v] * 0.0003;
						}
						Channel[n].Geom.MEGSensor.Coil[0].NumInt = 9;
						Channel[n].Geom.MEGSensor.Coil[0].Evaluated = TRUE;
					}	// end if(found...
				}
			}
		}

		// copy detection points to array (in sensor order)
		for(mm=0, n=1; mm<Header->NumPri; mm++) {
			m = Header->PsIndex[mm];
			if(Channel[m].Flag == TRUE) {
				if(Channel[m].Geom.MEGSensor.NumCoils == 2) {	// planar gradiometer sensor with 8 integration points
					for(c=0; c<2; c++)
						for(i=0; i<8; i++) {
							for(v=X_; v<=Z_; v++) {
								detector[n].Vi.p[v] = Channel[m].Geom.MEGSensor.Coil[c].B[i].p[v];
								detector[n].Vi.v[v] = Channel[m].Geom.MEGSensor.origin.v[v];
							}
							n++;
						}
				} else {										// magnetometer sensor with 9 integration points
					for(i=0; i<9; i++) {
						for(v=X_; v<=Z_; v++) {
							detector[n].Vi.p[v] = Channel[m].Geom.MEGSensor.Coil[0].B[i].p[v];
							detector[n].Vi.v[v] = Channel[m].Geom.MEGSensor.origin.v[v];
						}
						n++;
					}
				}
			}
		}

		// compute coefficients
		GetCoeffs(detector, S, &Hull, coeffs);
		CoeffFlag = FALSE;			// don't repeat unless sensor frame changes
	}

	// translate dipole source location to volume origin & compute correction for this dipole position
	for(v=X_; v<=Z_; v++)
		Vd[v] = Org[v] - Hull.Vo[v];
	GetBasis(Vd, Hull.scale, basis, gradbasis);

	// for all detection points
	for(i=1; i<=S; i++) {

		// translate detector location to conducting volume origin
		for(v=X_; v<=Z_; v++)
			Vs[v] = detector[i].Vi.p[v] - Hull.Vo[v];

		// solve field for spherical volume conductor (origin of volume conductor)
		xbd(Vd, Vs, Bd);

		// for each cardinal moment (v)
		for(v=X_; v<=Z_; v++) {

			// field is dot product of field Bd with detector normal vector (w)
			for(w=X_, fieldori[i][v]=0.; w<=Z_; w++)
				fieldori[i][v] += Bd[v][w] * detector[i].Vi.v[w];

			// final result
			for(j=1, correction[i][v]=0.; j<=BASIS; j++)
				correction[i][v] -= gradbasis[j][v] * coeffs[j][i];
			detector[i].Bi[v] = fieldori[i][v] + correction[i][v];
		}
	}

	// integrate field within each sensor coil (note that the order of the coils is obtained from the sensor order)
	for(mm=c=0, i=1; mm<Header->NumPri; mm++) {
		m = Header->PsIndex[mm];
		if(Channel[m].Flag == TRUE) {
			if(Channel[m].Geom.MEGSensor.NumCoils == 2) {	// integrate planar gradiometer field (T/m)
				for(v=X_; v<=Z_; v++) {						// 1st planar gradiometer coil
					for(j=0, coil[c][v]=0.; j<8; j++)
						coil[c][v] += detector[i+j].Bi[v] * Wg[j];
				}
				i += 8;
				c++;
				for(v=X_; v<=Z_; v++) {						// 2nd planar gradiometer coil
					for(j=0, coil[c][v]=0.; j<8; j++)
						coil[c][v] += detector[i+j].Bi[v] * Wg[j];
				}
				i += 8;
				c++;
			} else {										// integrate magnetometer field (T)
				for(v=X_; v<=Z_; v++) {
					for(j=0, coil[c][v]=0.; j<9; j++)
						coil[c][v] += detector[i+j].Bi[v] * Wm[j];
				}
				i += 9;
				c++;
			}
		}
	}

	// assemble sensors from coils, computing scalar field (or gradient) for each sensor (maintaining same sensor order as previously)
	for(mm=c=0; mm<Header->NumPri; mm++) {
		m = Header->PsIndex[mm];
		if(Channel[m].Flag == TRUE) {
			if(Channel[m].Geom.MEGSensor.NumCoils == 2) {	// planar gradiometer sensor
				for(v=X_; v<=Z_; v++)
					L[mm][v] = coil[c][v] - coil[c+1][v];
				c += 2;
			} else {										// magnetometer sensor
				for(v=X_; v<=Z_; v++)
					L[mm][v] = coil[c][v];
				c++;
			}
		}
	}
}
