// ECDSolve() - solve for beamformer optimal forward solution
//  for current dipole in realistic model. See SAMLIB/DIPSolve3D.c
//  & SAMLIB/DIPSolve2D.c for the multi-sphere & single sphere
//  models, respectively.
//
//  Author: Stephen E. Robinson
//      MEG Core Facility
//      NIMH
//

#include <lfulib.h>
#include <siglib.h>
#include <voxel.h>

#define QV      3                   // 3 dipole moment solution

void ECDSolve(
    VOXELINFO   *Voxel,             // dipole position, orientation, & flags
    double      *B,                 // B[M] -- optimal target forward solution
    double      **Cinv,             // Cinv[M][M] -- inverse covariance matrix
    HeaderInfo  *Header,            // header information with sensor list
    ChannelInfo *Channel,           // sensor information
    HULL        *Hull,              // cortical hull model
    double      *span               // eigenvalue span of 3x3 solution (return value)
) {
    static double   **K1;           // K1[QV][QV] -- numerator matrix
    static double   **K2;           // K2[QV][QV] -- denominator matrix
    static double   **A;            // A[M][QV] -- intermediate matrix
    static double   *Vec;           // solution eigenvector
    static double   *Val;           // eigenvalues of this problem
    static double   **Cinv2;        // Cinv2[M][M] -- inverse covariance matrix squared
    static double   **L;            // L[M][QV] -- primary sensor lead field
    static double   **Cp = NULL;    // copy of most recent Cinv
    double          m;              // vector magnitude
    double          m2;             // m^2
    double          tmp;            // accumulator
    int             i;              // i-index
    int             j;              // j-index
    int             k;              // k-index
    static int      M = 0;          // primary sensor rank

    // if req'd, allocate memory
    if(M == 0) {
        M = Header->NumPri;

        // allocate matrices & arrays
        if((Cinv2 = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
            allocfailed("Cinv2[]");
        for(i=0; i<M; i++)
            if((Cinv2[i] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
                allocfailed("Cinv2[][]");
        if((K1 = (double **)malloc((size_t)QV * sizeof(double *))) == NULL)
            allocfailed("K1[]");
        for(i=0; i<QV; i++)
            if((K1[i] = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
                allocfailed("K1[][]");
        if((K2 = (double **)malloc((size_t)QV * sizeof(double *))) == NULL)
            allocfailed("K2[]");
        for(i=0; i<QV; i++)
            if((K2[i] = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
                allocfailed("K2[][]");
        if((A = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
            allocfailed("A[]");
        for(i=0; i<M; i++)
            if((A[i] = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
                allocfailed("A[][]");
        if((Vec = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
            allocfailed("Vec[]");
        if((Val = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
            allocfailed("Val[]");
        if((L = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
            allocfailed("L[]");
        for(i=0; i<M; i++)
            if((L[i] = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
                allocfailed("L[][]");
    }

    if (Cp != Cinv) {
        Cp = Cinv;
        // Cinv changed, recompute Cinv2
        for(i=0; i<M; i++)
            for(j=0; j<M; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv[i][k] * Cinv[k][j];
                Cinv2[i][j] = tmp;
            }
    }

    // compute ECD lead field matrix - compute forward solution matrix for unit dipole terms
#if ELEKTA
    ECDLeadFieldElekta(Voxel->p, Header, Channel, Hull, L);
#else
    ECDLeadField(Voxel->p, Header, Channel, Hull, L);
#endif

    // linear solution of dipole moment vector
    if(Voxel->Solve == TRUE) {

        // compute K1 = L' C^(-1) L
        for(i=0; i<M; i++)              // A = C^(-1) * L
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv[i][k] * L[k][j];
                A[i][j] = tmp;
            }
        for(i=0; i<QV; i++)             // K1 = L' A
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += L[k][i] * A[k][j];
                K1[i][j] = tmp;
            }

        // compute K2 = L' C^(-2) L
        for(i=0; i<M; i++)              // A = C^(-2) * L
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv2[i][k] * L[k][j];
                A[i][j] = tmp;
            }
        for(i=0; i<QV; i++)             // K2 = L' A
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += L[k][i] * A[k][j];
                K2[i][j] = tmp;
            }

        // solve generalized eigensystem using "TRUE"
        // TRUE solves for eigenvector of maximum eigenvalue
        // FALSE for minimum eigenvalue
        if(SolveEigen(K1, K2, Vec, Val, QV, TRUE) == 0) {

            // normalize moment to unit magnitude
            for(i=0, m2=0.; i<QV; i++)
                m2 += Vec[i] * Vec[i];
            m = sqrt(m2);
            for(i=0; i<QV; i++)     // output moment vector
                Voxel->v[i] = Vec[i] / m;
            *span = Val[0] / Val[QV-1];
        } else {        // SolveEigen failed -- solution is zero
            for(i=0; i<QV; i++)
                Voxel->v[i] = 0.;
        }
    } else {
        *span = 0.;
    }

    // compute forward solution
    for(i=0; i<M; i++) {
        for(j=0, tmp=0.; j<QV; j++)
            tmp += Voxel->v[j] * L[i][j];
        B[i] = tmp;
    }
}
