// GetHull() -- read cortical hull model from text file in dataset directory
//      note that vertices are indexed base 1, for now...
//
//      Author: Stephen E. Robinson
//              MEG Core Facility
//              NIMH
//

#include <lfulib.h>

void GetHull(
    char    *hullName,          // path of hull.shape file
    HULL    *Hull)              // cortical hull model with position & normals of vertices, including origin & scale
{
    double  Vp[3];              // cartesian position vector
    double  Vo[3];              // cartesian orientation vector
    double  r;                  // normal vector magnitude
    double  r2;                 // r^2
    double  sum;                // simple sum
    double  tmp;                // what else?
    int     i;
    int     j;
    int     v;
    FILE    *fp;

    // read number of vertices & allocate memory
    if((fp = fopen(hullName, "r")) == NULL)
        cleanup("can't open hull file %s", hullName);
    if(fscanf(fp, "%d", &Hull->nv) != 1)
        cleanup("can't read number of vertices");
    if((Hull->vertex = (FIELD *)malloc((size_t)(Hull->nv + 1) * sizeof(FIELD))) == NULL)
        allocfailed("vertex[]");

    // read vertices & normals
    for(j=0, i=1; j<Hull->nv; j++) {
        if(fscanf(fp, "%lf%lf%lf", &Vp[X_], &Vp[Y_], &Vp[Z_]) != 3)
            cleanup("can't read vertex position");
        for(v=X_; v<=Z_; v++)
            if(Vp[v] > 0.2)                     // if position > 20 cm, something is very wrong!
                cleanup("vertex > 20 cm from origin -- corrupted 'hull.shape'?");
        if(fscanf(fp, "%lf%lf%lf", &Vo[X_], &Vo[Y_], &Vo[Z_]) != 3)
            cleanup("can't read vertex orientation");
        for(v=X_, r2=0.; v<=Z_; v++)            // enforce unit normal vector
            r2 += Vo[v] * Vo[v];
        if(r2 > 0.) {
            r = sqrt(r2);
            for(v=X_; v<=Z_; v++) {
                Hull->vertex[i].p[v] = Vp[v];   // hull.shape position is already metres
                Hull->vertex[i].v[v] = Vo[v] / r;
            }
            i++;
        }
    }
    fclose(fp);

    // test good vertices
    i--; // i is now # of good vertices
    if(i < (Hull->nv-100))  // missing more than 100 vertices!
        cleanup("more than 100 zero-length vertices -- corrupted 'hull.shape'?");
    if(i != Hull->nv) {
        fprintf(stderr, " - note: %d hull vertices have zero-length normals", Hull->nv - i);
        Hull->nv = i;       // revise number of vertices with non-zero normals
    }

    // compute hull geometric center
    for(i=1, Vp[X_]=Vp[Y_]=Vp[Z_]=0.; i<=Hull->nv; i++)
        for(v=X_; v<=Z_; v++)
            Vp[v] += Hull->vertex[i].p[v];
    for(v=X_; v<=Z_; v++)
        Hull->Vo[v] = Vp[v] / (double)Hull->nv;

    // compute hull scale
    for(i=1, sum=0.; i<=Hull->nv; i++) {
        for(v=X_, r2=0.; v<=Z_; v++) {
            tmp = Hull->vertex[i].p[v] - Hull->Vo[v];
            r2 += tmp * tmp;
        }
        sum += sqrt(r2);
    }
    Hull->scale = sum / (double)Hull->nv;
}
