// NSolveFwd() - solve for beamformer optimal forward solution
//  for any N current dipole in realistic head model.
//
//  Author: Stephen E. Robinson
//      MEG Core Facility
//      NIMH
//

#include <siglib.h>
#include <lfulib.h>

void NSolveFwd(
    VOXELINFO   *Voxel,             // Voxel[N] -- voxel structure -- req'd for solving orientation
    double      **L,                // L[M][3N] -- lead field matrix
    double      **B,                // B[M][N] -- beamformer forward solutions
    double      **C,                // C[M][M] -- orientation covariance matrix
    int         N,                  // number of sources
    int         M,                  // number of primary sensors
    int         NumVec              // dimensions (either 2 or 3)
) {
    static double   **K1;           // K1[QV][QV] -- numerator matrix
    static double   **K2;           // K2[QV][QV] -- denominator matrix
    static double   **A;            // A[M][QV] -- intermediate matrix
    static double   *Vec;           // solution eigenvector
    static double   *Val;           // eigenvalues of this problem
    static double   **Cinv;         // Cinv[M][M] -- inverse orientation covariance matrix
    static double   **Cinv2;        // Cinv2[M][M] -- inverse orientation covariance matrix squared
    double          r;              // eigenvector normalization constant
    double          tmp;            // accumulator
    int             solve;          // TRUE | FALSE -- flag for whether to solve for orientation
    register int    i;              // i-index
    register int    j;              // j-index
    register int    k;              // k-index
    register int    m;              // sensor index
    register int    v;              // vector index
    static int      QV;             // number of components (3N)
    static int      rank = 0;       // number of primary sensors

    // if req'd, allocate memory & invert covariance
    if(rank == 0) {
        rank = M;                   // make it so, Number One...
        QV = NumVec * N;

        // allocate matrices & arrays
        if((Cinv = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
            allocfailed("Cinv[]");
        for(i=0; i<M; i++)
            if((Cinv[i] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
                allocfailed("Cinv[][]");
        if((Cinv2 = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
            allocfailed("Cinv2[]");
        for(i=0; i<M; i++)
            if((Cinv2[i] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
                allocfailed("Cinv2[][]");
        if((K1 = (double **)malloc((size_t)QV * sizeof(double *))) == NULL)
            allocfailed("K1[]");
        for(i=0; i<QV; i++)
            if((K1[i] = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
                allocfailed("K1[][]");
        if((K2 = (double **)malloc((size_t)QV * sizeof(double *))) == NULL)
            allocfailed("K2[]");
        for(i=0; i<QV; i++)
            if((K2[i] = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
                allocfailed("K2[][]");
        if((A = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
            allocfailed("A[]");
        for(i=0; i<M; i++)
            if((A[i] = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
                allocfailed("A[][]");
        if((Vec = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
            allocfailed("Vec[]");
        if((Val = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
            allocfailed("Val[]");

        // inversion of orientation covariance matrix & its inverse^2
        for(i=0; i<M; i++)
            for(j=0; j<M; j++)
                Cinv[i][j] = C[i][j];
        invert(Cinv, M);
        for(i=0; i<M; i++)
            for(j=0; j<M; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv[i][k] * Cinv[k][j];
                Cinv2[i][j] = tmp;
            }
    }

    // determine whether vector solution is req'd
    switch(NumVec) {
        case 2:     // 2d solution (dipole in sphere)
            solve = TRUE;
            break;
        case 3:     // 3d solution (either multisphere or Nolte)
            for(i=0, solve=TRUE; i<N; i++)      // test voxel flags for whether to solve dipole orientation
                if(Voxel[i].Solve == FALSE) {   // if any one flag is false, don't solve for orientation
                    solve = FALSE;
                    break;
                }
            break;
        default:
            cleanup("NSolveFwd(): NumVec must be 2 or 3");
            break;
    }

    // test voxel flags for whether to solve dipole orientation
    for(i=0, solve=TRUE; i<N; i++)
        if(Voxel[i].Solve == FALSE) {   // if any one flag is false, don't solve for orientation
            solve = FALSE;
            break;
        }

    if(solve == TRUE) {

        // compute numerator matrix -- 1/S^2
        //  K1 = L' C^(-1) L
        for(i=0; i<M; i++)                  // A = C^(-1) * L
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv[i][k] * L[k][j];
                A[i][j] = tmp;
            }
        for(i=0; i<QV; i++)                 // K1 = L' A
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += L[k][i] * A[k][j];
                K1[i][j] = tmp;
            }

        // compute denominator matrix -- 1/N^2 (without sigma^2)
        //  K2 = L' C^(-1) L
        for(i=0; i<M; i++)                  // A = C^(-2) * L
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv2[i][k] * L[k][j];
                A[i][j] = tmp;
            }
        for(i=0; i<QV; i++)                 // K2 = L' A
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += L[k][i] * A[k][j];
                K2[i][j] = tmp;
            }

        // solve generalized eigensystem
        if(SolveEigen(K1, K2, Vec, Val, QV, TRUE) == 0) {       // TRUE solves for eigenvector of maximum eigenvalue, FALSE for minimum eigenvale

            // independently normalize all source orientation vectors
            for(i=0; i<N; i++) {
                for(j=0, k=NumVec*i, tmp=0.; j<NumVec; j++, k++)
                    tmp += Vec[k] * Vec[k];
                r = sqrt(tmp);
                for(j=0, k=NumVec*i; j<NumVec; j++, k++)
                    Vec[k] /= r;
            }

            // compute N-source forward solutions (L[M][QV] & B[M][N])
            for(i=0; i<N; i++)
                for(m=0; m<M; m++) {
                    for(j=0, k=NumVec*i, tmp=0.; j<NumVec; j++, k++)
                        tmp += Vec[k] * L[m][k];
                    B[m][i] = tmp;
                }

        } else {    // SolveEigen fails, set all B elements to zero
            for(i=0; i<N; i++)
                for(m=0; m<M; m++)
                    B[m][i] = 0.;
        }

    } else {        // compute N-Source forward solutions using voxel dipole vector

        // copy voxel orientation to Vec
        for(i=j=0; i<N; i++)
            for(v=X_; v<=Z_; v++, j++)
                Vec[j] = Voxel[i].v[v];

        // compute forward solutions for all sources
        for(i=0; i<N; i++)
            for(m=0; m<M; m++) {
                for(j=0, k=NumVec*i, tmp=0.; j<NumVec; j++, k++)
                    tmp += Vec[k] * L[m][k];
                B[m][i] = tmp;
            }
    }
}
