// NSolvewts() - solve for independent beamformer weights
//	Forward solutions are orthogonalized
//
//	Author:	Stephen E. Robinson
//		MEG Core Facility
//		NIMH
//

#include <siglib.h>
#include <lfulib.h>


void	NSolveWts(
	double		**B,				// B[M][N] -- N dipole forward solutions
	double		**C,				// C[M][M] -- solution covariance matrix
	double		**W,				// W[M][N] -- optimal beamformer coefficients
	int			N,					// number of sources
	int			M					// number of primary sensors
)
{
	static double	**K1;			// K1[M][N] -- numerator matrix
	static double	**K2;			// K2[N][N] -- denominator matrix
	static double	**Cinv;			// Cinv[M][M] -- inverse of solution covariance matrix
	register double	tmp;			// accumulator
	register int	i;				// i-index
	register int	j;				// j-index
	register int	k;				// k-index
	static int		rank = 0;		// number of primary sensors

	// if req'd, allocate memory & invert covariance
	if(rank != M) {
		rank = M;			// make it so, Number One...

		// allocate matrices & invert covariance
		if((Cinv = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
			allocfailed("Cinv[]");
		for(i=0; i<M; i++)
			if((Cinv[i] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
				allocfailed("Cinv[][]");
		if((K1 = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
			allocfailed("K1[]");
		for(i=0; i<M; i++)
			if((K1[i] = (double *)malloc((size_t)N * sizeof(double))) == NULL)
				allocfailed("K1[][]");
		if((K2 = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
			allocfailed("K2[]");
		for(i=0; i<N; i++)
			if((K2[i] = (double *)malloc((size_t)N * sizeof(double))) == NULL)
				allocfailed("K2[][]");

		// inversion of solution covariance matrix
		for(i=0; i<M; i++)
			for(j=0; j<M; j++)
				Cinv[i][j] = C[i][j];
		invert(Cinv, M);
	}

	// compute K1 = C^-1 B
	for(i=0; i<M; i++)
		for(j=0; j<N; j++) {
			for(k=0, tmp=0.; k<M; k++)
				tmp += Cinv[i][k] * B[k][j];
			K1[i][j] = tmp;
		}

	// compute K2 = (B' C^-1 B)^-1 = (B' K1)^-1
	for(i=0; i<N; i++)
		for(j=0; j<N; j++) {
			for(k=0, tmp=0.; k<M; k++)
				tmp += B[k][i] * K1[k][j];
			K2[i][j] = tmp;
		}
	invert(K2, N);

	// compute W = K1 * K2^-1
	for(i=0; i<M; i++)
		for(j=0; j<N; j++) {
			for(k=0, tmp=0.; k<N; k++)
				tmp += K1[i][k] * K2[k][j];
			W[i][j] = tmp;
		}
}
