// Data2CondRVE() -- computes rank vector entropy
//  Valid for E=3..6
//
//  arguments to subroutine:
//      x           -- data time-series (input)
//      h           -- entropy time-series (output)
//      T           -- number of samples in x (input)
//      LPFreq      -- data lowpass frequency (input)
//      tau         -- 1/e time - seconds (input)
//      advance     -- advance - milliseconds (input)
//      rate        -- sample rate (input)
//
//  Author:
//      Stephen E. Robinson, PhD
//      MEG Core Group
//      NIMH
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <rvelib.h>
#include <lfulib.h>


void    Data2CondRVE(
    double  *x,                 // x[T] -- input time-series
    double  *h,                 // h[T] -- output normalized entropy
    int     T,                  // total samples in x[T]
    int     E,                  // phase space dimension
    double  LPFreq,             // lowpass frequency (Hz)
    double  tau,                // 1/e time (seconds)
    double  advance,            // advance (milliseconds)
    double  rate                // sample rate (Hz)
) {
    static double   alpha;      // decay rate constant
    static double   norm;       // entropy normalization factor
    static double   **CXx;      // CXx[N][S] -- counts of x (X) advanced & x
    static double   **PXx;      // PXx[N][S] -- joint probability of x-advanced & x
    static double   *Cx;        // Cx[S] -- counts of x
    static double   *Px;        // Px[S] -- probability of x
    static double   *vector;    // vector[E+1] -- vector of lagged measurement values
    double          sum;        // sum of states
    double          Hx;         // entropy of x
    double          HXx;        // joint entropy of x-advanced & x
    static int      D;          // stride (lags)
    static int      S = 0;      // number of phase-space symbols
    static int      N;          // number of advance symbols
    int             n;          // state index
    int             nn;         // alternate state index
    int             s;          // state index
    int             ss;         // alternate state index
    int             w;          // word index
    static int      delta;      // number of samples advanced
    static int      TE;         // ending sample
    int             t;          // time index
    int             tt = 0;     // alternate time index

    // one-time initialization of constants
    if(S == 0) {
        switch(E) {
            case 3:     // 3-dimensional phase space + advance
                S = 6;
                N = 4;
                break;
            case 4:     // 4-dimensional phase space + advance
                S = 24;
                N = 5;
                break;
            case 5:     // 5-dimensional phase space + advance
                S = 120;
                N = 6;
                break;
            case 6:     // 6-dimensional phase space + advance
                S = 720;
                N = 7;
                break;
            default:
                cleanup("embedding space dimension out of range");
                break;
        }
        if((Cx = (double *)malloc((size_t)S * sizeof(double))) == NULL)
            allocfailed("Cx[]");
        if((CXx = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
            allocfailed("CXx[]");
        for(n=0; n<N; n++)
            if((CXx[n] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
                allocfailed("CXx[][]");
        if((Px = (double *)malloc((size_t)S * sizeof(double))) == NULL)
            allocfailed("Px[]");
        if((PXx = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
                        allocfailed("PXxx[]");
        for(n=0; n<N; n++)
            if((PXx[n] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
                allocfailed("PXxx[][]");
        if((vector = (double *)malloc((size_t)(E+1) * sizeof(double))) == NULL)
            allocfailed("vector[]");
        D = (int)floor(rate / (2. * LPFreq));   // the lags depend upon the sample rate & upper frequency limit
        norm = log2((double)S);                 // maximum entropy for S states
        if(tau > 0.)
            alpha = exp(-1. / (tau * rate));    // compute decay constant for 1/e time of tau
        else
            alpha = 1.;                         // no decay time constant if tau is zero
        delta = (int)floor(0.001 * advance * rate);
        TE = T - (E - 1) * D - delta - 1;
    }

    // initialize counts of Cx & CXx with all states populated
    for(s=0; s<S; s++) {
        Cx[s] = 1.;
        for(n=0; n<N; n++)
            CXx[n][s] = 1.;
    }

    // for the specified segment of data x(t)...
    for(t=0; t<TE; t++) {

        // extract measurement vector from x & convert to symbol
        for(w=0; w<E; w++) {
            tt = t + w * D;
            vector[w] = x[tt];
        }
        tt += delta;
        vector[E] = x[tt];
        nn = (int)Vec2Symbol(vector, E);
        ss = (int)Vec2Advance(vector, E);

        // allow state counts to decay & update symbol count
        for(s=0; s<S; s++) {
            Cx[s] *= alpha;
            for(n=0; n<N; n++)
                CXx[n][s] *= alpha;
        }
        Cx[ss] += 1.;
        CXx[nn][ss] += 1.;

        // compute marginal probability Px
        for(s=0, sum=0.; s<S; s++)
            sum += Cx[s];
        for(s=0; s<S; s++)
            Px[s] = Cx[s] / sum;

        // compute entropy Hx
        for(s=0, Hx=0.; s<S; s++)
            Hx -= (Px[s] > 1.0e-12)? Px[s] * log2(Px[s]): 0.;

        // compute joint probability PXx
        for(n=0, sum=0.; n<N; n++)
            for(s=0; s<S; s++)
                sum += CXx[n][s];
        for(n=0; n<N; n++)
            for(s=0; s<S; s++)
                PXx[n][s] = CXx[n][s] / sum;

        // compute entropy HXx
        for(n=0, HXx=0.; n<N; n++)
            for(s=0; s<S; s++)
                HXx -= (PXx[n][s] > 1.0e-12)? PXx[n][s] * log2(PXx[n][s]): 0.;

        // normalize to theoretical maximum entropy as upper bound.
        h[t] = (HXx - Hx) / norm;
// printf("HXx=%f, Hy=%f, H=%f\n", HXx, Hx, HXx-Hx);
    }

    // pad remaining h[t] with last computed value
    for(t=TE; t<T; t++)
        h[t] = h[TE-1];
}
