// Data2RVE() -- computes rank vector entropy for W=5
//
//	arguments to subroutine:
//		x			-- data time-series (input)
//		h			-- entropy time-series (output)
//		T			-- number of samples in x[T] (input)
//		E			-- phase space dimension (input)
//		lowpass		-- data lowpass frequency (input)
//		tau			-- 1/e time - seconds (input)
//		rate		-- sample rate (input)
//
//	Author:
//		Stephen E. Robinson, PhD (based upon philosophy of A. Mandell)
//		MEG Core Group
//		NIMH
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <rvelib.h>


void	Data2RVE(
	double	*x,						// x[T] -- input time-series
	double	*h,						// h[T-P] -- output normalized entropy
	int		T,						// number of samples in x[T]
	int		E,						// phase-space dimension
	double	LPFreq,					// lowpass frequency (Hz)
	double	tau,					// 1/e time (seconds)
	double	rate					// sample rate (Hz)
) {
	static double	*Cx;			// Cx[S] -- state counts
	static double	*Px;			// Px[S] -- state probabilities
	static double	alpha;			// decay rate constant
	static double	norm;			// entropy normalization factor
	static double	*vector;		// vector[E] -- vector of lagged measurement values
	double			sum;			// sum of states
	static int		S = 0;			// number of symbols
	static int		D;				// stride (lags)
	static int		TE;				// samples used
	int				s;				// states index
	int				w;				// word index
	short			state;			// symbolic state index
	int				t;				// time index
	int				tt;				// alternate time index

	// one-time initialization of constants
	if(S == 0) {
		switch(E) {
			case 3: S = 6; break;
			case 4: S = 24; break;
			case 5: S = 120; break;
			case 6: S = 720; break;
			default:
				cleanup("embedding space dimension out of range");
				break;
		}
		if((Cx = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Cx[]");
		if((Px = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Px[]");
		if((vector = (double *)malloc((size_t)E * sizeof(double))) == NULL)
			allocfailed("vector[]");
		D = (int)floor(rate / (2. * LPFreq));	// the lags depend upon the sample rate & upper frequency limit
		TE = T - (E - 1) * D;					// number of samples to be evaluated
		norm = log2((double)S);					// maximum entropy for S states
		if(tau > 0.)
			alpha = exp(-1. / (tau * rate));	// compute decay constant for 1/e time of tau
		else
			alpha = 1.;							// no decay time constant if tau is zero
	}

	// initialize histogram with all states populated
	for(s=0; s<S; s++)
		Cx[s] = 1.;

	// for the specified segment of data x(t)...
	for(t=0; t<TE; t++) {

		// extract measurement vector from x & convert to symbol
		for(w=0; w<E; w++) {
			tt = t + w * D;
			vector[w] = x[tt];
		}
		state = Vec2Symbol(vector, E);

		// allow state counts to decay & update symbol count
		for(s=0; s<S; s++)
			Cx[s] *= alpha;
		Cx[state] += 1.;

		// compute probabilities of each state
		for(s=0, sum=0.; s<S; s++)
			sum += Cx[s];
		for(s=0; s<S; s++)
			Px[s] = (sum > 0.)? Cx[s] / sum: 0.;

		// compute entropy
		for(s=0, h[t]=0.; s<S; s++)
			h[t] -= (Px[s] > 1.0e-12)? Px[s] * log2(Px[s]): 0.;

		// normalize to theoretical maximum entropy as upper bound.
		h[t] /= norm;
	}

	// pad remaining h[t] with last computed value
	for(t=TE; t<T; t++)
		h[t] = h[TE-1];
}
