// Data2SymbolsArray-- computes symbols from rank vectors
//  & outputs symbol time series. Works for E=3..6
//
//  arguments to subroutine:
//      x           -- data time-series (input)
//      SymPhs      -- phase space symbol time-series (output)
//      SymAdv      -- advamce time-series (output)
//      T           -- number of samples in x (input)
//      TE          -- number of symbols derived from time-series (output)
//      advance     -- number of samples advanced
//      lowpass     -- data lowpass frequency (input)
//      rate        -- sample rate (input)
//
//  Author:
//      Stephen E. Robinson, PhD
//      MEG Core Group
//      NIMH
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <rvelib.h>


void    Data2SymbolsArray(
    double  *x,                     // x[T] -- input time-series (input)
    short   *SymPhs,                // SymPhs[TE] -- phase space symbol array (output)
    short   *SymAdv,                // SymAdv[TE] -- advance symbol array (output)
    int     T,                      // number of samples in x
    int     *TE,                    // number of symbols computed
    int     E,                      // phase space size
    int     advance,                // advance (samples) from last lag
    double  LPFreq,                 // lowpass frequency (Hz)
    double  rate                    // sample rate (Hz)
) {
    double          vector[7];      // vector of lagged measurement values (large enough for E=6)
    static int      Stride = 0;     // stride (phase-space lags)
    int             w;              // word index
    int             t;              // time index
    int             tt = 0;         // alternate time index

    // test!
    if(E < 3 || E > 6)
        cleanup("bad phase space dimension");

    // one-time initialization of lag constant
    if(Stride == 0)
        Stride = (int)floor(rate / (2. * LPFreq));  // the lags depend upon the sample rate & upper frequency limit
    *TE = T - ((E-1) * Stride + advance);           // the output time-series is truncated by the lag space & advance

    // for the specified segment of data x(t)...
    for(t=0; t<*TE; t++) {

        // extract measurement vector from x & convert to symbol
        for(w=0; w<E; w++) {
            tt = t + w * Stride;
            vector[w] = x[tt];
        }
        tt += advance;
        vector[E] = x[tt];
        SymPhs[t] = Vec2Symbol(vector, E);
        SymAdv[t] = Vec2Advance(vector, E);
    }
}
