// Sym2MI() -- compute mutual information from a pair of symbolic time-series for W=5
//
//	arguments to subroutine:
//		SymPhsX		-- symbolic time-series X
//		SymPhsY		-- symbolic time-series Y
//		TE			-- length of symbolic time-series (input)
//		advance		-- advance of Y wrt X (input)
//		E			-- phase space dimension
//
//	Author:
//		Stephen E. Robinson, PhD
//		MEG Core Group
//		NIMH
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <rvelib.h>


double	Sym2MI(
	short	*SymPhsX,				// symbolic time-series X
	short	*SymPhsY,				// symbolic time-series Y
	int		TE,						// length of symbol time-series
	int		advance,				// advance of Y wrt X
	int		E						// phase space dimension
) {
	static double	*Px;			// Px[S] -- counts/probability of X
	static double	*Py;			// Py[S] -- counts/probability of Y
	static double	**Pxy;			// Pxy[S][S] -- counts/probabiity of XY
	static double	norm;			// normalization constant
	double			info;			// mutual information
	double			sx;				// sum of x
	double			sy;				// sum of y
	double			sxy;			// sum of xy
	double			tmp;			// no need to explain
	static int		*Cx;			// Cx[S] -- x-counts
	static int		*Cy;			// Cy[S] -- y-counts
	static int		**Cxy;			// Cxy[S][S] -- xy-counts
	static int		S = 0;			// number of phase-space states
	int				s;				// states index
	int				i;				// i-index
	int				j;				// j-index
	int				x;				// x-symbol index
	int				y;				// y-symbol index
	int				z;				// advanced y-symbol index
	int				t;				// time index

	// one-time computation of norm
	if(S == 0) {
		switch(E) {
			case 3: S = 6; break;
			case 4: S = 24; break;
			case 5: S = 120; break;
			case 6: S = 720; break;
			default: cleanup("embedding space dimension out of range"); break;
		}
		norm = log2((double)S);
		if((Px = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Px[]");
		if((Py = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Py[]");
		if((Pxy = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
			allocfailed("Pxy[]");
		for(i=0; i<S; i++)
			if((Pxy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("Pxy[][]");
		if((Cx = (int *)malloc((size_t)S * sizeof(int))) == NULL)
			allocfailed("Cx[]");
		if((Cy = (int *)malloc((size_t)S * sizeof(int))) == NULL)
			allocfailed("Cy[]");
		if((Cxy = (int **)malloc((size_t)S * sizeof(int *))) == NULL)
			allocfailed("Cxy[]");
		for(i=0; i<S; i++)
			if((Cxy[i] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
				allocfailed("Cxy[][]");
	}

	// initialize counts Cx, Cy, & Cxy
	for(i=0; i<S; i++) {
		Cx[i] = 0;
		Cy[i] = 0;
		for(j=0; j<S; j++)
			Cxy[i][j] = 0;
	}

	// integrate counts Cx, Cy, & Cxy	
	for(t=0; t<(TE-advance); t++) {
		x = SymPhsX[t];
		y = SymPhsY[t];
		z = SymPhsY[t+advance];
		Cx[x]++;
		Cy[y]++;
		Cxy[x][z]++;		// only Cxy uses advance
	}

	// compute probabilities
	for(s=0, sx=sy=0.; s<S; s++) {
		sx += (double)Cx[s];
		sy += (double)Cy[s];
	}
	for(s=0; s<S; s++) {
		Px[s] = (sx > 0.)? (double)Cx[s] / sx: 0.;
		Py[s] = (sy > 0.)? (double)Cy[s] / sy: 0.;
	}
	for(i=0, sxy=0.; i<S; i++)
		for(j=0; j<S; j++)
			sxy += (double)Cxy[i][j];
	for(i=0; i<S; i++)
		for(j=0; j<S; j++)
			Pxy[i][j] = (sxy > 0.)? (double)Cxy[i][j] / sxy: 0.;

	// calculate mutual information
	for(i=0, info=0.; i<S; i++)
		for(j=0; j<S; j++) {
			tmp = (Px[i] > 0. && Py[j] > 0.)? Pxy[i][j] / (Px[i] * Py[j]): 0.;
			info += (tmp > 1.0e-12)? Pxy[i][j] * log2(tmp): 0.;
		}

	// at last!
	return(info / norm);
}
