// Sym2RVE() -- computes instantaneous rank vector entropy
//	(requires lowpass filtering!).
//
//	arguments to subroutine:
//		x			-- data time-series (input)
//		h			-- entropy time-series (output)
//		T			-- number of samples in x[T] (input)
//		E			-- phase space dimension (input)
//
//	Author:
//		Stephen E. Robinson, PhD (based upon philosophy of A. Mandell)
//		MEG Core Group
//		NIMH
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <rvelib.h>


void	Sym2RVE(
	short	*Sym,					// symbolic time-series
	double	*h,						// h -- output normalized entropy
	int		T,						// number of samples in x[T]
	int		E						// phase-space dimension
) {
	static double	*Cx;			// Cx[S] -- state counts
	static double	*Px;			// Px[S] -- state probabilities
	double			sum;			// sum of states
	double			suminv;			// inverse sum
	static int		S = 0;			// number of symbols
	int				s;				// states index
	int				t;				// time index
	int				x;				// symbol index

	// one-time initialization of constants
	if(S == 0) {
		switch(E) {
			case 3: S = 6; break;
			case 4: S = 24; break;
			case 5: S = 120; break;
			case 6: S = 720; break;
			default:
				cleanup("embedding space dimension out of range");
				break;
		}
		if((Cx = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Cx[]");
		if((Px = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Px[]");
	}

	// initialize histogram with all states populated
	for(s=0; s<S; s++)
		Cx[s] = 0.;

	// accumulate counts
	for(t=0; t<T; t++) {
		x = Sym[t];
		Cx[x]++;
	}

	// compute marginal probability Px
	for(s=0, sum=0.; s<S; s++)
		sum += (double)Cx[s];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(s=0; s<S; s++)
			Px[s] = suminv * (double)Cx[s];
	} else {
		for(s=0; s<S; s++)
			Px[s] = 0.;
	}

	// compute instantaneous RVE (bits)-- requires filetering!
	for(t=0; t<T; t++) {
		s = Sym[t];
		h[t] = (Px[s] > 1.0e-12)? -Px[s] * log2(Px[s]): 0.;
	}
}
