// Sym2STE() -- computes symbolic transfer entropy between two time series.
//	This subroutine assumes symbols were generated by Data2SymbolsArray().
//	Observe caution for the value of T. It is less than that for the
//	length of the measurement time-series by W*D, where D is the number
//	of sample lags for the embedding space of dimension W.
//
//	arguments to subroutine:
//		SymPhsX		-- X phase-space symboltime-series (input)
//		SymAdvX		-- X advance symbol time-series (input)
//		SymPhsY		-- Y phase-space symbol time-series (input)
//		SymAdvY		-- Y adance symbol time-series (input)
//		Hx2y		-- forward transfer entropy (output)
//		Hy2x		-- reverse transfer entropy (output)
//		E			-- phase-space dimension (input)
//		T			-- number of samples in X & Y symbols (input)
//
//	Author:
//		Stephen E. Robinson, PhD (inspired by A. Mandell)
//		MEG Core Group
//		NIMH
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <rvelib.h>
#include <lfulib.h>


void	Sym2STE(
	short	*SymPhsX,				// X phase-space symbol time series
	short	*SymAdvX,				// X advance symbol time-series
	short	*SymPhsY,				// Y phase-space symbol time series
	short	*SymAdvY,				// Y advance symbol time series
	float	*Hx2y,					// forward transfer entropy
	float	*Hy2x,					// reverse transfer entropy
	int		E,						// phase-space dimension
	int		T						// number of samples in X & Y symbol time series
) {
	static double	*Px = NULL;		// Px[S] -- probability of an X symbol
	static double	*Py = NULL;		// Py[S] -- probability of a Y symbol
	static double	**Pxy = NULL;	// Pxy[S][S] -- probability of coincidence of X & Y symbols
	static double	**PXx = NULL;	// PXx[N][S] -- probability of coincidence of X+advance and X
	static double	**PYy = NULL;	// PYy[N][S] -- probability of coincidence of Y+advance and Y
	static double	***PXxy = NULL;	// PXxy[N][S][S] -- probability of coincidence of X+advance, X, & Y
	static double	***PYxy = NULL;	// PYxy[N][S][S] -- probability of coincidence of Y+advance, X, & Y
	double			sum;			// simple integrator
	double			suminv;			// 1/sum
	double			num;			// numerator
	double			den;			// denominator
	static int		*Cx = NULL;		// Cx[S] -- counts of x states
	static int		*Cy = NULL;		// Cy[S] -- counts of y states
	static int		**Cxy = NULL;	// Cxy[S][S] -- counts of x,y states
	static int		**CXx = NULL;	// CXx[N][S] -- counts of X+advance,x
	static int		**CYy = NULL;	// CYy[N][S] -- counts of Y+advance,y
	static int		***CXxy = NULL;	// CXxy[N][S][S] -- counts of X+advance,x,t
	static int		***CYxy = NULL;	// CYxy[N][S][S] -- counts of Y+advance),x,y
	static int		S;				// number of phase-space symbols
	static int		N;				// number of advance symbols
	int				i;				// i-index
	int				j;				// j-index
	int				k;				// k-index
	int				x;				// x phase-space symbol index
	int				xx;				// x advance symbol index
	int				y;				// y phase-space symbol index
	int				yy;				// y advance symbol index
	int				t;				// symbolic time-series sample index

	// initialize constants & memory
	if(S == 0) {
		switch(E) {
			case 3: S = 6; N = 4; break;
			case 4: S = 24; N = 5; break;
			case 5: S = 120; N = 6; break;
			case 6: S = 720; N = 7; break;
			default: cleanup("phase space dimension out of range"); break;
		}

		// allocate memory
		if((Cx = (int *)malloc((size_t)S * sizeof(int))) == NULL)
			allocfailed("Cx[]");
		if((Cy = (int *)malloc((size_t)S * sizeof(int))) == NULL)
			allocfailed("Cy[]");
		if((Cxy = (int **)malloc((size_t)S * sizeof(int *))) == NULL)
			allocfailed("Cxy[]");
		for(i=0; i<S; i++)
			if((Cxy[i] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
				allocfailed("Cxy[][]");
		if((Px = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Px[]");
		if((Py = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Py[]");
		if((Pxy = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
			allocfailed("Pxy[]");
		for(i=0; i<S; i++)
			if((Pxy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("Pxy[][]");
		if((CXx = (int **)malloc((size_t)N * sizeof(int *))) == NULL)
			allocfailed("CXx[]");
		for(i=0; i<N; i++)
			if((CXx[i] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
				allocfailed("CXx[][]");
		if((CYy = (int **)malloc((size_t)N * sizeof(int *))) == NULL)
			allocfailed("CYy[]");
		for(i=0; i<N; i++)
			if((CYy[i] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
				allocfailed("CYy[][]");
		if((CXxy = (int ***)malloc((size_t)N * sizeof(int **))) == NULL)
			allocfailed("CXxy[]");
		for(i=0; i<N; i++) {
			if((CXxy[i] = (int **)malloc((size_t)S * sizeof(int *))) == NULL)
				allocfailed("CXxy[][]");
			for(j=0; j<S; j++)
				if((CXxy[i][j] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
					allocfailed("CXxy[][][]");
		}
		if((CYxy = (int ***)malloc((size_t)N * sizeof(int **))) == NULL)
			allocfailed("CYxy[]");
		for(i=0; i<N; i++) {
			if((CYxy[i] = (int **)malloc((size_t)S * sizeof(int *))) == NULL)
				allocfailed("CYxy[][]");
			for(j=0; j<S; j++)
				if((CYxy[i][j] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
					allocfailed("CYxy[][][]");
		}
		if((PXx = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
			allocfailed("PXx[]");
		for(i=0; i<N; i++)
			if((PXx[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("PXx[][]");
		if((PYy = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
			allocfailed("PYy[]");
		for(i=0; i<N; i++)
			if((PYy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("PYy[][]");
		if((PXxy = (double ***)malloc((size_t)N * sizeof(double **))) == NULL)
			allocfailed("PXxy[]");
		for(i=0; i<N; i++) {
			if((PXxy[i] = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
				allocfailed("PXxy[][]");
			for(j=0; j<S; j++)
				if((PXxy[i][j] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
					allocfailed("PXxy[][][]");
		}
		if((PYxy = (double ***)malloc((size_t)N * sizeof(double **))) == NULL)
			allocfailed("PYxy[]");
		for(i=0; i<N; i++) {
			if((PYxy[i] = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
				allocfailed("PYxy[][]");
			for(j=0; j<S; j++)
				if((PYxy[i][j] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
					allocfailed("PYxy[][][]");
		}
	}

	// initialize & integrate symbol counts for Cx, Cy, & Cxy
	for(i=0; i<S; i++) {
		Cx[i] = 0;
		Cy[i] = 0;
		for(j=0; j<S; j++)
			Cxy[i][j] = 0;
	}
	for(t=0; t<T; t++) {
		x = SymPhsX[t];
		y = SymPhsY[t];
		Cx[x]++;
		Cy[y]++;
		Cxy[x][y]++;
	}

	// initialize & integrate CXx, CYy, CXxy, & CYxy
	for(i=0; i<N; i++) {
		for(j=0; j<S; j++) {
			CXx[i][j] = 0;
			CYy[i][j] = 0;
			for(k=0; k<S; k++) {
				CXxy[i][j][k] = 0;
				CYxy[i][j][k] = 0;
			}
		}
	}
	for(t=0; t<T; t++) {
		x = SymPhsX[t];
		xx = SymAdvX[t];
		y = SymPhsY[t];
		yy = SymAdvY[t];
		CXx[xx][x]++;
		CYy[yy][y]++;
		CXxy[xx][x][y]++;
		CYxy[yy][x][y]++;
	}

	// compute marginal probability p(x)
	for(i=0, sum=0.; i<S; i++)
		sum += (double)Cx[i];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<S; i++)
			Px[i] = suminv * (double)Cx[i];
	} else {
		for(i=0; i<S; i++)
			Px[i] = 0.;
	}

	// compute marginal probability p(y)
	for(i=0, sum=0.; i<S; i++)
		sum += (double)Cy[i];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<S; i++)
			Py[i] = suminv * (double)Cy[i];
	} else {
		for(i=0; i<S; i++)
			Py[i] = 0.;
	}

	// compute 2D joint probability p(xy)
	for(i=0, sum=0.; i<S; i++)
		for(j=0; j<S; j++)
			sum += (double)Cxy[i][j];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<S; i++)
			for(j=0; j<S; j++)
				Pxy[i][j] = suminv * (double)Cxy[i][j];
	} else {
		for(i=0; i<S; i++)
			for(j=0; j<S; j++)
				Pxy[i][j] = 0.;
	}

	// compute 2D joint probability p(Xx)
	for(i=0, sum=0.; i<N; i++)
		for(j=0; j<S; j++)
			sum += (double)CXx[i][j];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				PXx[i][j] = suminv * (double)CXx[i][j];
	} else {
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				PXx[i][j] = 0.;
	}

	// compute 2D joint probability p(Yy)
	for(i=0, sum=0.; i<N; i++)
		for(j=0; j<S; j++)
			sum += (double)CYy[i][j];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				PYy[i][j] = suminv * (double)CYy[i][j];
	} else {
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				PYy[i][j] = 0.;
	}

	// compute 3D joint probability p(Xxy)
	for(i=0, sum=0.; i<N; i++)
		for(j=0; j<S; j++)
			for(k=0; k<S; k++)
				sum += CXxy[i][j][k];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					PXxy[i][j][k] = suminv * (double)CXxy[i][j][k];
	} else {
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					PXxy[i][j][k] = 0.;
	}

	// compute 3D joint probability p(Yxy)
	for(i=0, sum=0.; i<N; i++)
		for(j=0; j<S; j++)
			for(k=0; k<S; k++)
				sum += CYxy[i][j][k];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					PYxy[i][j][k] = suminv * (double)CYxy[i][j][k];
	} else {
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					PYxy[i][j][k] = 0.;
	}

	// forward transfer entropy
	for(i=0, sum=0.; i<S; i++) {	// integral over xn
		for(j=0; j<S; j++) {		// integral over yn
			for(k=0; k<N; k++) {	// integral over yn+advance
				num = PYxy[k][i][j] * Py[j];
				den = Pxy[i][j] * PYy[k][j];
				if(num > 0. && den > 0.)
					sum += PYxy[k][i][j] * log2(num / den);
			}
		}
	}
	*Hx2y = sum;

	// reverse transfer entropy
	for(i=0, sum=0.; i<S; i++) {	// integral over xn
		for(j=0; j<S; j++) {		// integral over yn
			for(k=0; k<N; k++) {	// integral over xn+advance
				num = PXxy[k][i][j] * Px[i];
				den = Pxy[i][j] * PXx[k][i];
				if(num > 0. && den > 0.)
					sum += PXxy[k][i][j] * log2(num / den);
			}
		}
	}
	*Hy2x = sum;
}
