// Sym2STEi() -- computes symbolic transfer entropy between two time series.
//	Observe caution for the value of T. The entropy units for Hxy are bits/sec.
//	Works for E=3..5
//
//	arguments to subroutine:
//		SymPhsX		-- X symbol time-series (input)
//		SymPhsY		-- Y symbol time-series (input)
//		SymAdvY		-- Y symbol time-series (input)
//		Hxy			-- transfer entropy (output)
//		T			-- number of symbols in X & Y (input)
//		E			-- phase space dimension (input)
//
//	Author:
//		Stephen E. Robinson, PhD (inspired by A. Mandell)
//		MEG Core Group
//		NIMH
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <siglib.h>
#include <rvelib.h>


void	Sym2STEi(
	short	*SymPhsX,				// X phase space symbol time series
	short	*SymPhsY,				// Y phase space symbol time series
	short	*SymAdvY,				// Y advance symbol time series
	double	tau,					// integration window duration
	double	rate,					// sample rate
	float	*Hxy,					// forward transfer entropy
	int		T,						// number of samples in X & Y symbol time series
	int		E						// phase space dimension
) {

	static double	*py;			// py[S] -- probability of a Y symbol
	static double	**pxy;			// pxy[S][S] -- probability of coincidence of X & Y symbols
	static double	**pYy;			// pYy[N][S] -- probability of coincidence of Y+advance and Y
	static double	***pYxy;		// pYxy[N][S][S] -- probability of coincidence of Y+advance, X, & Y
	static double	*window;		// window[W] -- hanning window
	static double	norm;			// normalization to bits/second
	double			sum;			// integrator
	double			suminv;			// 1/sum
	double			num;			// numerator
	double			den;			// denominator
	double			plog;			// log(p)
	double			tmp;			// what else?
	static int		*Cy;			// Cy[S] -- counts of y states
	static int		**Cxy;			// Cxy[S][S] -- counts of x,y states
	static int		**CYy;			// CYy[N][S] -- counts of Y+advance,y
	static int		***CYxy;		// CYxy[N][S][S] -- counts of Y+advance),x,y
	static int		N;				// number of advance symbols
	static int		S;				// number of phase-space symbols
	static int		W = 0;			// integrator window duration (samples)
	int				i;				// i-index
	int				j;				// j-index
	int				k;				// k-index
	int				x;				// x-symbol index
	int				y;				// y-symbol index
	int				yy;				// Y-symbol index
	int				t;				// sample index
	int				tt;				// alternate sample index
	int				w;				// integration window index
	int				n;				// window sample count

	// initialize arrays, matrices, tensors, & constants
	if(W == 0) {
		switch(E) {
			case 3:		// 3-dimensional embedding space + advance
				S = 6;
				N = 4;
				break;
			case 4:		// 4-dimensional embedding space + advance
				S = 24;
				N = 5;
				break;
			case 5:		// 5-dimensional embedding space + advance
				S = 120;
				N = 6;
				break;
			default:
				cleanup("embedding space dimension out of range");
				break;
		}
		i = (int)rint(rate * tau);
		j = i % 2;
		W = i + j;
		if(W < 10)
			cleanup("integration window is <10 samples");
		if((window = (double *)malloc((size_t)W * sizeof(double))) == NULL)
			allocfailed("window[]");
		Hanning(window, W);
		if((py = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("py[]");
		if((pxy = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
			allocfailed("pxy[]");
		for(i=0; i<S; i++)
			if((pxy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("pxy[][]");
		if((pYy = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
			allocfailed("pYy[]");
		for(i=0; i<N; i++)
			if((pYy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("pYy[][]");
		if((pYxy = (double ***)malloc((size_t)N * sizeof(double **))) == NULL)
			allocfailed("pYxy[]");
		for(i=0; i<N; i++) {
			if((pYxy[i] = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
				allocfailed("pYxy[][]");
			for(j=0; j<S; j++)
				if((pYxy[i][j] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
					allocfailed("pYxy[][][]");
		}
		if((Cy = (int *)malloc((size_t)S * sizeof(int))) == NULL)
			allocfailed("Cy[]");
		if((Cxy = (int **)malloc((size_t)S * sizeof(int *))) == NULL)
			allocfailed("Cxy[]");
		for(i=0; i<S; i++)
			if((Cxy[i] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
				allocfailed("Cxy[][]");
		if((CYy = (int **)malloc((size_t)N * sizeof(int *))) == NULL)
			allocfailed("CYy[]");
		for(i=0; i<N; i++)
			if((CYy[i] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
				allocfailed("CYy[][]");
		if((CYxy = (int ***)malloc((size_t)N * sizeof(int **))) == NULL)
			allocfailed("CYxy[]");
		for(i=0; i<N; i++) {
			if((CYxy[i] = (int **)malloc((size_t)S * sizeof(int *))) == NULL)
				allocfailed("CYxy[][]");
			for(j=0; j<S; j++)
				if((CYxy[i][j] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
					allocfailed("CYxy[][][]");
		}
	}

	// initialize all counts to zero
	for(i=0; i<S; i++)
		Cy[i] = 0;
	for(i=0; i<S; i++)
		for(j=0; j<S; j++)
			Cxy[i][j] = 0;
	for(i=0; i<N; i++)
		for(j=0; j<S; j++)
			CYy[i][j] = 0;
	for(i=0; i<N; i++)
		for(j=0; j<S; j++)
			for(k=0; k<S; k++)
				CYxy[i][j][k] = 0;

	// integrate symbol counts for Cy & compute py
	for(t=0; t<T; t++) {
		y = SymPhsY[t];
		Cy[y]++;
	}
	for(i=0, sum=0.; i<S; i++)
		sum += (double)Cy[i];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<S; i++)
			py[i] = suminv * (double)Cy[i];
	} else {
		for(i=0; i<S; i++)
			py[i] = 0.;
	}

	// integrate symbol counts for Cxy & compute pxy
	for(t=0; t<T; t++) {
		x = SymPhsX[t];
		y = SymPhsY[t];
		Cxy[x][y]++;
	}
	for(i=0, sum=0.; i<S; i++)
		for(j=0; j<S; j++)
			sum += (double)Cxy[i][j];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<S; i++)
			for(j=0; j<S; j++)
				pxy[i][j] = suminv * (double)Cxy[i][j];
	} else {
		for(i=0; i<S; i++)
			for(j=0; j<S; j++)
				pxy[i][j] = 0.;
	}

	// integrate symbol counts for CYy & compute pYy
	for(t=0; t<T; t++) {
		y = SymPhsY[t];
		yy = SymAdvY[t];
		CYy[yy][y]++;
	}
	for(i=0, sum=0.; i<N; i++)
		for(j=0; j<S; j++)
			sum += (double)CYy[i][j];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				pYy[i][j] = suminv * (double)CYy[i][j];
	} else {
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				pYy[i][j] = 0.;
	}

	// integrate symbol counts for CYxy & compute pYxy
	for(t=0; t<T; t++) {
		x = SymPhsX[t];
		y = SymPhsY[t];
		yy = SymAdvY[t];
		CYxy[yy][x][y]++;
	}
	for(i=0, sum=0.; i<N; i++)
		for(j=0; j<S; j++)
			for(k=0; k<S; k++)
				sum += (double)CYxy[i][j][k];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					pYxy[i][j][k] = suminv * (double)CYxy[i][j][k];
	} else {
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					pYxy[i][j][k] = 0.;
	}

	// pad the first W/2 samples with zeros
	for(t=0; t<W/2; t++)
		Hxy[t] = 0.;

	// compute entropy for each sample (centered on t)
	for(t=0; t<T; t++) {
		for(w=n=0, tt=t-W/2, sum=0.; w<W; w++, tt++) {
			if(tt > 0 && tt < T-1) {

				// get current symbols
				x = SymPhsX[tt];
				y = SymPhsY[tt];
				yy = SymAdvY[tt];

				// compute temporo-local entropy based on accumulated probabilites
				num = pYxy[yy][x][y] * py[y];
				den = pxy[x][y] * pYy[yy][y];
				if(den > 0.) {
					tmp = num / den;
					plog = (tmp > 1.)? log2(tmp): 0.;
					sum += window[w] * pYxy[yy][x][y] * plog;
					n++;
				}
			}
		}
		norm = rate / (double)n;
		Hxy[t] = sum * norm;
//		printf("Hxy[%d]=%f\n", t, Hxy[t]);
	}
}
