// Sym2STEint() -- computes symbolic transfer entropy between two time series.
//	Observe caution for the value of T. The entropy units for Hfwd & Hrev are bits/sec.
//	This works for E=3..6.
//
//	arguments to subroutine:
//		SymPhsX		-- X symbol time-series (input)
//		SymPhsY		-- Y symbol time-series (input)
//		SymAdvX		-- X advance symbol time-series (input)
//		SymAdvY		-- Y advance symbol time-series (input)
//		tau			-- integration window (input)
//		H			-- transfer entropy time-series (output)
//		Direction	-- forward or reverse transfer entropy (output)
//		E			-- dimension of phase space (input)
//		T			-- number of symbols in X & Y time-series (input)
//
//	Author:
//		Stephen E. Robinson, PhD (inspired by A. Mandell)
//		MEG Core Group
//		NIMH
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <siglib.h>
#include <rvelib.h>

#define FWD	0
#define REV	1


void	Sym2STEint(
	short	*SymPhsX,				// x phase space symbol time series
	short	*SymAdvX,				// x advance symbolic state time series
	short	*SymPhsY,				// y phase space symbol time series
	short	*SymAdvY,				// y advance symbolic state time series
	double	tau,					// integration window duration
	double	rate,					// sample rate
	float	*H,						// transfer entropy
	int		Direction,				// 0=Fwd, 1=Rev
	int		E,						// dimension of phase space
	int		T						// number of samples in x & y symbolic time series
) {
	static double	*Px;			// Px[S] -- probability of an x phase-state symbol
	static double	*Py;			// Py[S] -- probability of a y phase-state symbol
	static double	**Pxy;			// Pxy[S][S] -- joint probability of x & y symbols
	static double	**PXx;			// PXx[N][S] -- joint probability of X (advance) & x
	static double	**PYy;			// PYy[N][S] -- joint probability of Y (advance) & y
	static double	***PXxy;		// PYxy[N][S][S] -- joint probability of X(advance), x, & y
	static double	***PYxy;		// PYxy[N][S][S] -- joint probability of Y (advance), x, & y
	static double	*window;		// window[W] -- hanning window
	double			sum;			// integrator
	double			suminv;			// 1/sum
	double			num;			// numerator
	double			den;			// denominator
	double			tmp;			// what else?
	static int		*Cx;			// Cx[S] -- counts of x phase-space states
	static int		*Cy;			// Cy[S] -- counts of y phase-space states
	static int		**Cxy;			// Cxy[S][S] -- counts of x & y states
	static int		**CXx;			// CXx[N][S] -- counts of X (advance) & x
	static int		**CYy;			// CYy[N][S] -- counts of Y (advance) & y
	static int		***CXxy;		// CXxy[N][S][S] -- counts of X (advance), x, & y
	static int		***CYxy;		// CYxy[N][S][S] -- counts of Y (advance), x, & y
	static int		N;				// number of advance symbolic states
	static int		S = 0;			// number of embedding window symbolic states
	static int		W = 0;			// integrator window duration (samples)
	int				i;				// i-index
	int				j;				// j-index
	int				k;				// k-index
	int				x;				// x-symbol index
	int				y;				// y-symbol index
	int				xx;				// X-symbol index
	int				yy;				// Y-symbol index
	int				t;				// sample index
	int				tt;				// alternate sample index
	int				w;				// integration window index
	int				n;				// window sample count

	// initialize constants
	if(W == 0) {
		switch(E) {
			case 3: S = 6; N = 4; break;
			case 4: S = 24; N = 5; break;
			case 5: S = 120; N = 6; break;
			case 6: S = 720; N = 7; break;
			default: cleanup("phase space dimension out of range"); break;
		}
		i = (int)rint(rate * tau);
		j = i % 2;
		W = i + j;
		if(W < 10)
			cleanup("integration window is <10 samples");
		if((window = (double *)malloc((size_t)W * sizeof(double))) == NULL)
			allocfailed("window[]");
		Hanning(window, W);

		// allocate memory dependent upon S -- unchanged between calls
		if((Px = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Px[]");
		if((Py = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Py[]");
		if((Pxy = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
			allocfailed("Pxy[]");
		for(i=0; i<S; i++)
			if((Pxy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("Pxy[][]");
		if((Cx = (int *)malloc((size_t)S * sizeof(int))) == NULL)
			allocfailed("Cx[]");
		if((Cy = (int *)malloc((size_t)S * sizeof(int))) == NULL)
			allocfailed("Cy[]");
		if((Cxy = (int **)malloc((size_t)S * sizeof(int *))) == NULL)
			allocfailed("Cxy[]");
		for(i=0; i<S; i++)
			if((Cxy[i] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
				allocfailed("Cxy[][]");
		if((CXx = (int **)malloc((size_t)N * sizeof(int *))) == NULL)
			allocfailed("CXx[]");
		for(i=0; i<N; i++)
			if((CXx[i] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
				allocfailed("CXx[][]");
		if((CYy = (int **)malloc((size_t)N * sizeof(int *))) == NULL)
			allocfailed("CYy[]");
		for(i=0; i<N; i++)
			if((CYy[i] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
				allocfailed("CYy[][]");
		if((CXxy = (int ***)malloc((size_t)N * sizeof(int **))) == NULL)
			allocfailed("CXxy[]");
		for(i=0; i<N; i++) {
			if((CXxy[i] = (int **)malloc((size_t)S * sizeof(int *))) == NULL)
				allocfailed("CXxy[][]");
			for(j=0; j<S; j++)
				if((CXxy[i][j] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
					allocfailed("CXxy[][][]");
		}
		if((CYxy = (int ***)malloc((size_t)N * sizeof(int **))) == NULL)
			allocfailed("CYxy[]");
		for(i=0; i<N; i++) {
			if((CYxy[i] = (int **)malloc((size_t)S * sizeof(int *))) == NULL)
				allocfailed("CYxy[][]");
			for(j=0; j<S; j++)
				if((CYxy[i][j] = (int *)malloc((size_t)S * sizeof(int))) == NULL)
					allocfailed("CYxy[][][]");
		}
		if((PXx = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
			allocfailed("PXx[]");
		for(i=0; i<N; i++)
			if((PXx[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("PXx[][]");
		if((PYy = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
			allocfailed("PYy[]");
		for(i=0; i<N; i++)
			if((PYy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("PYy[][]");
		if((PXxy = (double ***)malloc((size_t)N * sizeof(double **))) == NULL)
			allocfailed("PXxy[]");
		for(i=0; i<N; i++) {
			if((PXxy[i] = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
				allocfailed("PXxy[][]");
			for(j=0; j<S; j++)
				if((PXxy[i][j] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
					allocfailed("PXxy[][][]");
		}
		if((PYxy = (double ***)malloc((size_t)N * sizeof(double **))) == NULL)
			allocfailed("PYxy[]");
		for(i=0; i<N; i++) {
			if((PYxy[i] = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
				allocfailed("PYxy[][]");
			for(j=0; j<S; j++)
				if((PYxy[i][j] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
					allocfailed("PYxy[][][]");
		}
	}

	// initialize all counts to zero
	for(i=0; i<S; i++) {
		Cx[i] = 0;
		Cy[i] = 0;
	}
	for(i=0; i<S; i++)
		for(j=0; j<S; j++)
			Cxy[i][j] = 0;
	for(i=0; i<N; i++)
		for(j=0; j<S; j++) {
			CXx[i][j] = 0;
			CYy[i][j] = 0;
	}
	for(i=0; i<N; i++)
		for(j=0; j<S; j++)
			for(k=0; k<S; k++) {
				CXxy[i][j][k] = 0;
				CYxy[i][j][k] = 0;
			}

	// integrate symbol counts
	for(t=0; t<T; t++) {
		x = SymPhsX[t];
		xx = SymAdvX[t];
		y = SymPhsY[t];
		yy = SymAdvY[t];
		Cx[x]++;
		Cy[y]++;
		Cxy[x][y]++;
		CXx[xx][x]++;
		CYy[yy][y]++;
		CXxy[xx][x][y]++;
		CYxy[yy][x][y]++;
	}

	// compute Px
	for(i=0, sum=0.; i<S; i++)
		sum += (double)Cx[i];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<S; i++)
			Px[i] = suminv * (double)Cx[i];
	} else {
		for(i=0; i<S; i++)
			Px[i] = 0.;
	}

	// compute Py
	for(i=0, sum=0.; i<S; i++)
		sum += (double)Cy[i];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<S; i++)
			Py[i] = suminv * (double)Cy[i];
	} else {
		for(i=0; i<S; i++)
			Py[i] = 0.;
	}

	// compute Pxy
	for(i=0, sum=0.; i<S; i++)
		for(j=0; j<S; j++)
			sum += (double)Cxy[i][j];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<S; i++)
			for(j=0; j<S; j++)
				Pxy[i][j] = suminv * (double)Cxy[i][j];
	} else {
		for(i=0; i<S; i++)
			for(j=0; j<S; j++)
				Pxy[i][j] = 0.;
	}

	// compute PXx
	for(i=0, sum=0.; i<N; i++)
		for(j=0; j<S; j++)
			sum += (double)CXx[i][j];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				PXx[i][j] = suminv * (double)CXx[i][j];
	} else {
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				PXx[i][j] = 0.;
	}

	// compute PYy
	for(i=0, sum=0.; i<N; i++)
		for(j=0; j<S; j++)
			sum += (double)CYy[i][j];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				PYy[i][j] = suminv * (double)CYy[i][j];
	} else {
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				PYy[i][j] = 0.;
	}

	// compute PXxy
	for(i=0, sum=0.; i<N; i++)
		for(j=0; j<S; j++)
			for(k=0; k<S; k++)
				sum += (double)CXxy[i][j][k];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					PXxy[i][j][k] = suminv * (double)CXxy[i][j][k];
	} else {
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					PXxy[i][j][k] = 0.;
	}

	// compute PYxy
	for(i=0, sum=0.; i<N; i++)
		for(j=0; j<S; j++)
			for(k=0; k<S; k++)
				sum += (double)CYxy[i][j][k];
	if(sum > 0.) {
		suminv = 1. / sum;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					PYxy[i][j][k] = suminv * (double)CYxy[i][j][k];
	} else {
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					PYxy[i][j][k] = 0.;
	}

	// directional entropy
	switch(Direction) {

		case FWD:			// compute forward transfer entropy for each sample (centered on t)
			for(t=0; t<T; t++) {
				for(w=n=0, tt=t-W/2, sum=0.; w<W; w++, tt++) {
					if(tt > 0 && tt < T-1) {

						// get current symbols
						x = SymPhsX[tt];
						y = SymPhsY[tt];
						yy = SymAdvY[tt];

						// instantaneous Hx2y
						num = PYxy[yy][x][y] * Py[y];
						den = Pxy[x][y] * PYy[yy][y];
						if(num > 0. && den > 0.)
							if((tmp = num / den) >= 1.)
								sum += window[w] * PYxy[yy][x][y] * log2(num / den);
						n++;
					}
				}
				H[t] = sum * rate / (double)n;
			}
			break;

		case REV:		// compute reverse transfer entropy for each sample (centered on t)
			for(t=0; t<T; t++) {
				for(w=n=0, tt=t-W/2, sum=0.; w<W; w++, tt++) {
					if(tt > 0 && tt < T-1) {

						// get current symbols
						x = SymPhsX[tt];
						y = SymPhsY[tt];
						xx = SymAdvX[tt];

						// instantaneous Hy2x
						num = PXxy[xx][x][y] * Px[x];
						den = Pxy[x][y] * PXx[xx][x];
						if(num > 0. && den > 0.)
							if((tmp = num / den) >= 1.)
								sum += window[w] * PXxy[xx][x][y] * log2(num / den);
						n++;
					}
				}
				H[t] = sum * rate / (double)n;
			}
			break;

		default:
			cleanup("bad direction!");
			break;
	}
}
