// Sym2STEl() -- computes the time-dependent forward and reverse
//	symbolic transfer entropy between two time series using the
//	"leaky integrator" method. This subroutine assumes symbols were
//	generated by Aug2SymbolArray(). Observe caution for the value
//	of T. It is less than that for the length of the measurement
//	time-series. The upper case letters X & Y in the counts and
//	probabilities refer to the advance symbols, while lower case
//	x & y refer to the phase-space symbols.
//
//	This version allows for augmented advance symbols as suggested
//	by D. Kugiumtzis, "Improvement of Symbolic Transfer Entropy".
//	Changing the advance changes the number of symbolic states for
//	the advance. Hence, memory must be freed when re-using subroutine.
//
//	arguments to subroutine:
//		SymPhsX		-- X phase-space symbol time-series (input)
//		SymAdvX		-- X advance symbol time-series (input)
//		SymPhsY		-- Y phase-space symbol time-series (input)
//		SymAdvY		-- Y advance symbol time-series (input)
//		tau			-- integrator time constant - seconds (input)
//		rate		-- sample rate - Hz (input)
//		H			-- H[T] -- forward or reverse transfer entropy time-series (output)
//		Direction	-- forward or reverse transfer entropy (input)
//		E			-- phase-space dimension (input)
//		N			-- number of advance symbolic states (input)
//		T			-- number of symbols in X & Y (input)
//
//	Author:
//		Stephen E. Robinson, PhD (inspired by A. Mandell)
//		MEG Core Group
//		NIMH
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <lfulib.h> 				// for allocfailed()
#include <rvelib.h>


void	Sym2STEl(
	short	*SymPhsX,				// X phase-space symbol time-series
	short	*SymAdvX,				// X advance symbolic state time-series
	short	*SymPhsY,				// Y phase-space symbol time-series
	short	*SymAdvY,				// Y advance symbolic state time-series
	double	tau,					// leaky integration constant
	double	rate,					// sample rate
	float	*H,						// H[T] -- forward or reverse transfer entropy
	int		Direction,				// select forward or reverse transfer entropy
	int		E,						// phase space dimension
	int		T						// actual number of samples in X & Y symbol time series
) {
	static double	*Cx;			// Cx[S] -- x symbolic state counts
	static double	*Cy;			// Cy[S] -- y symbolic state counts
	static double	**Cxy;			// Cxy[S][S] -- joint counts of x,y states
	static double	**CXx;			// CXx[N][S] -- joint counts of X,x states
	static double	**CYy;			// CYy[N][S] -- joint counts of Y,y states
	static double	***CXxy;		// CXxy[N][S][S] -- joint counts of Y,x,y states
	static double	***CYxy;		// CYxy[N][S][S] -- joint counts of Y,x,y states
	static double	*Px;			// Px[S] -- x-state probability
	static double	*Py;			// Py[S] -- y-state probability
	static double	**Pxy;			// Pxy[S][S] -- joint probability of x,y states
	static double	**PXx;			// PXx[N][S] -- joint probability of X,x states
	static double	**PYy;			// PYy[N][S] -- joint probability of Y,y states
	static double	***PXxy;		// PXxy[N][S][S] -- joint probability of X,x,y
	static double	***PYxy;		// PYxy[N][S][S] -- joint probability of Y,x,y
	static double	alpha;			// leaky integrator multiplier per sample
	double			sum;			// integrator
	double			num;			// numerator
	double			den;			// denominator
	static int		N;				// number of advance symbols
	static int		S = 0;			// number of phase-space symbols
	int				i;				// i-index
	int				j;				// j-index
	int				k;				// k-index
	int				x;				// x-symbol value
	int				y;				// y-symbol value
	int				xx;				// x-advance symbol value
	int				yy;				// y-advance symbol value
	int				t;				// sample index

	// initialize constants & memory
	if(S == 0) {
		switch(E) {
			case 3: S = 6; N = 4; break;
			case 4: S = 24; N = 5; break;
			case 5: S = 120; N = 6; break;
			case 6: S = 720; N = 7; break;
			default: cleanup("phase space dimension out of range"); break;
		}
		if(tau > 0.)
			alpha = exp(-1. / (tau * rate));	// compute decay constant for 1/e time of tau
		else
			alpha = 1.;							// no decay time constant if tau is zero

		// allocate memory dependent upon S -- unchanged between calls
		if((Cx = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Cx[]");
		if((Cy = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Cy[]");
		if((Cxy = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
			allocfailed("Cxy[]");
		for(i=0; i<S; i++)
			if((Cxy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("Cxy[][]");
		if((Py = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Py[]");
		if((Pxy = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
			allocfailed("Pxy[]");
		for(i=0; i<S; i++)
			if((Pxy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("Pxy[][]");
		if((CXx = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
			allocfailed("CXx[]");
		for(i=0; i<N; i++)
			if((CXx[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("CXx[][]");
		if((CYy = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
			allocfailed("CYy[]");
		for(i=0; i<N; i++)
			if((CYy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("CYy[][]");
		if((CXxy = (double ***)malloc((size_t)N * sizeof(double **))) == NULL)
			allocfailed("CXxy[]");
		for(i=0; i<N; i++) {
			if((CXxy[i] = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
				allocfailed("CXxy[][]");
			for(j=0; j<S; j++)
				if((CXxy[i][j] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
					allocfailed("CXxy[][][]");
		}
		if((CYxy = (double ***)malloc((size_t)N * sizeof(double **))) == NULL)
			allocfailed("CYxy[]");
		for(i=0; i<N; i++) {
			if((CYxy[i] = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
				allocfailed("CYxy[][]");
			for(j=0; j<S; j++)
				if((CYxy[i][j] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
					allocfailed("CYxy[][][]");
		}
		if((PXx = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
			allocfailed("PXx[]");
		for(i=0; i<N; i++)
			if((PXx[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("PXx[][]");
		if((PYy = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
			allocfailed("PYy[]");
		for(i=0; i<N; i++)
			if((PYy[i] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
				allocfailed("PYy[][]");
		if((PXxy = (double ***)malloc((size_t)N * sizeof(double **))) == NULL)
			allocfailed("PXxy[]");
		for(i=0; i<N; i++) {
			if((PXxy[i] = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
				allocfailed("PXxy[][]");
			for(j=0; j<S; j++)
				if((PXxy[i][j] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
					allocfailed("PXxy[][][]");
		}
		if((PYxy = (double ***)malloc((size_t)N * sizeof(double **))) == NULL)
			allocfailed("PYxy[]");
		for(i=0; i<N; i++) {
			if((PYxy[i] = (double **)malloc((size_t)S * sizeof(double *))) == NULL)
				allocfailed("PYxy[][]");
			for(j=0; j<S; j++)
				if((PYxy[i][j] = (double *)malloc((size_t)S * sizeof(double))) == NULL)
					allocfailed("PYxy[][][]");
		}
	}

	// initialize counts
	for(i=0; i<S; i++) {
		Cx[i] = 1.;
		Cy[i] = 1.;
	}
	for(i=0; i<S; i++)
		for(j=0; j<S; j++)
			Cxy[i][j] = 1.;
	for(i=0; i<N; i++)
		for(j=0; j<S; j++) {
			CXx[i][j] = 1.;
			CYy[i][j] = 1.;
	}
	for(i=0; i<N; i++)
		for(j=0; j<S; j++)
			for(k=0; k<S; k++) {
				CXxy[i][j][k] = 1.;
				CYxy[i][j][k] = 1.;
			}

	// estimate forward transfer entropy as a function of time...
	for(t=0; t<T; t++) {

		// (1) allow all past state counts to decay
		for(i=0; i<S; i++) {
			Cx[i] *= alpha;
			Cy[i] *= alpha;
		}
		for(i=0; i<S; i++)
			for(j=0; j<S; j++)
				Cxy[i][j] *= alpha;
		for(i=0; i<N; i++)
			for(j=0; j<S; j++) {
				CXx[i][j] *= alpha;
				CYy[i][j] *= alpha;
			}
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++) {
					CXxy[i][j][k] *= alpha;
					CYxy[i][j][k] *= alpha;
				}

		// (2) fetch current symbol indices
		x = SymPhsX[t];
		y = SymPhsY[t];
		xx = SymAdvX[t];
		yy = SymAdvY[t];

		// (3) update C(x) & recompute marginal probability P(x)
		Cx[x] += 1.;
		for(i=0, sum=0.; i<S; i++)
			sum += Cx[i];
		for(i=0; i<S; i++)
			Px[i] = Cx[i] / sum;

		// (4) update C(y) & recompute marginal probability P(y)
		Cy[y] += 1.;
		for(i=0, sum=0.; i<S; i++)
			sum += Cy[i];
		for(i=0; i<S; i++)
			Py[i] = Cy[i] / sum;

		// (5) update C(xy) & recompute 2D joint probability p(xy)
		Cxy[x][y] += 1.;
		for(i=0, sum=0.; i<S; i++)
			for(j=0; j<S; j++)
				sum += Cxy[i][j];
		for(i=0; i<S; i++)
			for(j=0; j<S; j++)
				Pxy[i][j] = Cxy[i][j] / sum;


		// (6) update C(Xx) & recompute 2D joint probability P(Xx)
		CXx[xx][x] += 1.;
		for(i=0, sum=0.; i<N; i++)
			for(j=0; j<S; j++)
				sum += CXx[i][j];
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				PXx[i][j] = CXx[i][j] / sum;

		// (7) update C(Yy) & recompute 2D joint probability P(Yy)
		CYy[yy][y] += 1.;
		for(i=0, sum=0.; i<N; i++)
			for(j=0; j<S; j++)
				sum += CYy[i][j];
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				PYy[i][j] = CYy[i][j] / sum;

		// (8) update C(Xxy) & recompute 3D joint probability P(Xxy)
		CXxy[xx][x][y] += 1.;
		for(i=0, sum=0.; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					sum += CXxy[i][j][k];
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					PXxy[i][j][k] = CXxy[i][j][k] / sum;

		// (9) update C(Yxy) & recompute 3D joint probability P(Yxy)
		CYxy[yy][x][y] += 1.;
		for(i=0, sum=0.; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					sum += CYxy[i][j][k];
		for(i=0; i<N; i++)
			for(j=0; j<S; j++)
				for(k=0; k<S; k++)
					PYxy[i][j][k] = CYxy[i][j][k] / sum;

		// (10) directional entropy
		switch(Direction) {

			case FWD:		// compute current forward transfer entropy

				// forward transfer entropy
				for(i=0, sum=0.; i<S; i++) {	// integral over xn
					for(j=0; j<S; j++) {		// integral over yn
						for(k=0; k<N; k++) {	// integral over yn+advance
							num = PYxy[k][i][j] * Py[j];
							den = Pxy[i][j] * PYy[k][j];
							if(num > 0. && den > 0.)
								sum += PYxy[k][i][j] * log2(num / den);
						}
					}
				}
				H[t] = sum;
				break;

			case REV:		// compute current reverse transfer entropy

				// reverse transfer entropy
				for(i=0, sum=0.; i<S; i++) {	// integral over xn
					for(j=0; j<S; j++) {		// integral over yn
						for(k=0; k<N; k++) {	// integral over xn+advance
							num = PXxy[k][i][j] * Px[i];
							den = Pxy[i][j] * PXx[k][i];
							if(num > 0. && den > 0.)
								sum += PXxy[k][i][j] * log2(num / den);
						}
					}
				}
				H[t] = sum;
				break;

			default:
				cleanup("bad direction!");
				break;
		}
	}
}
