// rvestatic() -- computes rank vector entropy
//	currently set for rank vector length of 5
//
//	arguments to subroutine:
//		x			-- data time-series (input)
//		TS			-- starting sample in x (input)
//		TE			-- ending sample in x (input)
//		lowpass		-- data lowpass frequency (input)
//		rate		-- sample rate (input)
//
//	Author:
//		Stephen E. Robinson, PhD (based upon philosophy of A. Mandell)
//		MEG Core Group
//		NIMH
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <rvelib.h>


double	rvestatic(
	double	*x,						// x[T] -- input time-series
	int		TS,						// starting sample in x[T]
	int		TE,						// ending sample in x[T]
	int		E,						// phase space dimension
	double	LPFreq,					// lowpass frequency (Hz)
	double	rate					// sample rate (Hz)
) {
	static double	norm;			// entropy normalization factor
	static double	*Px;			// Px[S] -- probability histogram of symbols
	static double	*vector;		// vector[E] -- vector of lagged measurement values
	double			h;				// entropy
	double			sx;				// sum of state counts
	static int		*Cx;			// Cx[S] -- symbol counts
	static int		D;				// stride (lags)
	static int		S = 0;			// number of symbolic states
	int				s;				// states index
	int				w;				// word index
	int				index;			// symbol index
	int				t;				// time index
	int				tt;				// alternate time index

	// one-time initialization of constants
	if(S == 0) {
		switch(E) {
			case 3: S = 6; break;
			case 4: S = 24; break;
			case 5: S = 120; break;
			case 6: S = 720; break;
			default:
				cleanup("embedding space dimension out of range");
				break;
		}
		D = (int)floor(rate / (2. * LPFreq));	// the lags depend upon the sample rate & upper frequency limit
		norm = log2((double)S);					// maximum entropy for S states
		if((Px = (double *)malloc((size_t)S * sizeof(double))) == NULL)
			allocfailed("Px[]");
		if((Cx = (int *)malloc((size_t)S * sizeof(int))) == NULL)
			allocfailed("Cx[]");
		if((vector = (double *)malloc((size_t)E * sizeof(double))) == NULL)
			allocfailed("vector[]");
	}

	// initialize histogram counts to zero
	for(s=0; s<S; s++)
		Cx[s] = 0;

	// for the specified segment of data x(t)...
	for(t=TS; t<=(TE-E*D); t++) {

		// extract measurement vector from x,  convert to symbol, & update counts
		for(w=0; w<E; w++) {
			tt = t + w * D;
			vector[w] = x[tt];
		}
		index = (int)Vec2Symbol(vector, E);
		Cx[index]++;
	}

	// compute sum over all states (sum of probabilities must be 1.0)
	for(s=0, sx=0.; s<S; s++)
		sx += (double)Cx[s];
	for(s=0; s<S; s++)
		Px[s] = (sx > 0.)? (double)Cx[s] / sx: 0.;

	// calculate entropy
	for(s=0, h=0.; s<S; s++)
		h -= (Px[s] > 1.0e-12)? Px[s] * log2(Px[s]): 0.;

	// normalize to theoretical maximum entropy as upper bound.
	return(h / norm);
}
