// DIPLeadField2D() -- compute lead field matrix L for 2 tangential dipoles
//	for an array of sensors.
//
//	Note: origin/position is given in _headframe_coordinate_ &
//	not local sphere coordinates.
//
//	Author:	Stephen E. Robinson (after Nolte, Sarvas, & Hamalainen)
//			MEG Core Group
//			NIMH
//

#include <stdio.h>
#include <stdlib.h>
#include <samlib.h>
#include <math.h>
#include <cme.h>
#include <model.h>


int	DIPLeadField2D(
	double			CMEOrg[3],		// expansion origin (headframe coordinates)
	double			**L,			// L[2][M] - 2 x M lead field matrix (external array allocation)
	HeaderInfo		*Header,		// header information, including primary & reference sensor indices
	ChannelInfo		*Channel,		// Channel[M] -- channel structure with primary and reference sensors
	int				Order			// integration order: 1, 2, 3, 4, 5, & 7 are implemented
) {
	FIELD			Q6;				// test dipole
	static double	Sp[3];			// local sphere origin
	double			Vx[3];			// vector in the x' tangential direction
	double			Vy[3];			// vector in the y' tangential direction
	double			Vr[3];			// vector in the radial direction
	double			r;				// magnitude of r
	double			r2;				// r^2
	double			diff;			// difference between sphere origin and CMEOrg
	int				m;				// sensor index
	int				v;				// vector index
	static int		M = 0;			// number of sensors is initialized as zero

	//one-time determination of local sphere origin
	if(M == 0) {
		M = Header->NumPri;
		m = Header->PsIndex[0];				// get 1st sensor to read local sphere origin
		for(v=X_; v<=Z_; v++)
			Sp[v] = Channel[m].Geom.MEGSensor.LocalSphere[v];
	}
	
	// find normalized position vector
	for(v=X_, r2=0.; v<=Z_; v++) {
		diff = CMEOrg[v] - Sp[v];
		r2 += diff * diff;
	}
	r = sqrt(r2);

	// derive radial vector, & solve tangential plane
	for(v=X_; v<=Z_; v++)
		Vr[v] = (CMEOrg[v] - Sp[v]) / r;
	OrthoPlane(Vr, Vx, Vy);

	// for dipole moments, set origin to CMEOrg
	Q6.p[X_] = CMEOrg[X_];	Q6.p[Y_] = CMEOrg[Y_];	Q6.p[Z_] = CMEOrg[Z_];

	// (1) compute Qx'
	Q6.v[X_] = (double)Vx[X_];	Q6.v[Y_] = (double)Vx[Y_];	Q6.v[Z_] = (double)Vx[Z_];
	if(FwdVec(&Q6, L[QU_], Header, Channel, Order) == -1)
		return -1;

	// (2) compute Qy'
	Q6.v[X_] = (double)Vy[X_];	Q6.v[Y_] = (double)Vy[Y_];	Q6.v[Z_] = (double)Vy[Z_];
	if(FwdVec(&Q6, L[QV_], Header, Channel, Order) == -1)
		return -1;

	// normal return
	return 0;
}
