// DIPSolve2D() - solve for beamformer optimal forward solution
//  for current dipole in single local sphere model.
//
//  Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <cme.h>
#include <model.h>
#include <voxel.h>
#include <samlib.h>
#include <siglib.h>

#define QV          2               // 2 dipole components

void DIPSolve2D(
    VOXELINFO       *Voxel,         // dipole position, orientation, & flags
    double          *Bt,            // Bt[M] -- optimal target forward solution
    double          **Cinv,         // Cinv[M][M] -- inverse covariance matrix
    HeaderInfo      *Header,        // header information with sensor list
    ChannelInfo     *Channel,       // sensor information
    int             Order,          // field integration order
    double          *span           // eigenvalue span of 2x2 solution (return value)
) {
    double          L1[QV][QV];     // source power matrix
    double          L2[QV][QV];     // noise power matrix
    double          Z2[QV][QV];     // eigenvectors of SNR solution
    double          Val[QV];        // eigenvalues of this problem
    static double   **Cinv2;        // Cinv2[M][M] -- inverse covariance matrix squared
    static double   **A;            // A[M][QV] -- intermediate matrix
    static double   **Bp;           // Bp[QV][M] -- primary sensor lead field
    static double   **Cp = NULL;    // copy of most recent Cinv
    double          tmp;            // accumulator
    int             i;              // i-index
    int             j;              // j-index
    int             k;              // k-index
    int             M;              // number of sensors
    static int      rank = 0;       // primary sensor rank

    M = Header->NumPri;

    // if req'd, allocate memory & invert covariance
    if(rank != M) {

        // if req'd, free prior allocation
        if(rank != 0) {
            for(i=0; i<rank; i++)
                free((void *)Cinv[i]);
            free((void *)Cinv);
            for(i=0; i<rank; i++)
                free((void *)Cinv2[i]);
            free((void *)Cinv2);
            for(i=0; i<rank; i++)
                free((void *)A[i]);
            free((void *)A);
            for(i=0; i<QV; i++)
                free((void *)Bp[i]);
            free((void *)Bp);
        }

        // allocate matrices & arrays
        if((Cinv2 = (double **)malloc((size_t)(M * sizeof(double *)))) == NULL)
            allocfailed("Cinv2[]");
        for(i=0; i<M; i++)
            if((Cinv2[i] = (double *)malloc((size_t)(M * sizeof(double)))) == NULL)
                allocfailed("Cinv2[][]");
        if((A = (double **)malloc((size_t)(M * sizeof(double *)))) == NULL)
            allocfailed("A[]");
        for(i=0; i<M; i++)
            if((A[i] = (double *)malloc((size_t)(QV * sizeof(double)))) == NULL)
                allocfailed("A[][]");
        if((Bp = (double **)malloc((size_t)(QV * sizeof(double *)))) == NULL)
            allocfailed("Bp[]");
        for(i=0; i<QV; i++)
            if((Bp[i] = (double *)malloc((size_t)(M * sizeof(double)))) == NULL)
                allocfailed("Bp[][]");
        rank = M;
    }

    if (Cp != Cinv) {
        Cp = Cinv;
        // Cinv changed, recompute Cinv2
        for(i=0; i<M; i++)
            for(j=0; j<M; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv[i][k] * Cinv[k][j];
                Cinv2[i][j] = tmp;
            }
    }

    // compute lead field matrix - compute forward solution matrix for unit dipole
    if(DIPLeadField2D(Voxel->p, Bp, Header, Channel, Order) == -1)
        cleanup("'DipLeadField2D()' failed");

    // linear solution of dipole moment vector
    if(Voxel->Solve == TRUE) {

        // compute L1 = L' C^(-1) L
        for(i=0; i<M; i++)                      // A = C^(-1) * L
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv[i][k] * Bp[j][k];
                A[i][j] = tmp;
            }
        for(i=0; i<QV; i++)                     // L1 = L' A
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Bp[i][k] * A[k][j];
                L1[i][j] = tmp;
            }

        // compute L2 = L' C^(-1) L
        for(i=0; i<M; i++)                      // A = C^(-2) * L
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv2[i][k] * Bp[j][k];
                A[i][j] = tmp;
            }
        for(i=0; i<QV; i++)                     // L2 = L' A
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Bp[i][k] * A[k][j];
                L2[i][j] = tmp;
            }

        // solve generalized eigensystem problem for SNR
        geig22(L1, L2, Z2, Val);
        for(i=0; i<QV; i++)
            Voxel->v[i] = Z2[0][i];     // eigenvectors are row vectors in the 'geig22()' solution
        Voxel->v[2] = 0.;
        *span = Val[0] / Val[1];
    } else {
        *span = 0.;
    }

    // compute dipole forward solution
    for(i=0; i<M; i++)
        for(j=0, Bt[i]=0.; j<QV; j++)
            Bt[i] += Voxel->v[j] * Bp[j][i];
}
