// DIPSolve3D() - solve for beamformer optimal forward solution
//  for current dipole in multiple local spheres model.
//
//  Author: Stephen E. Robinson
//      MEG Core Facility
//      NIMH
//

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <cme.h>
#include <model.h>
#include <geoms.h>
#include <voxel.h>
#include <samlib.h>
#include <siglib.h>

#define QV      3                   // 3 dipole moment solution

void DIPSolve3D(
    VOXELINFO   *Voxel,             // dipole position, orientation, & flags
    double      *Bt,                // Bt[M] -- optimal target forward solution
    double      **Cinv,             // Cinv[M][M] -- inverse covariance matrix
    HeaderInfo  *Header,            // header information with sensor list
    ChannelInfo *Channel,           // sensor information
    int         Order,              // field integration order
    double      *span               // eigenvalue span of 3x3 solution (return value)
) {
    static double   **L1;           // L1[QV][QV] -- numerator matrix
    static double   **L2;           // L2[QV][QV] -- denominator matrix
    static double   *Vec;           // solution eigenvector
    static double   *Val;           // eigenvalues of this problem
    double          m;              // eigenvector normalization constant
    double          m2;             // m^2
    static double   **Cinv2;        // Cinv2[M][M] -- inverse covariance matrix squared
    static double   **A;            // A[M][QV] -- intermediate matrix
    static double   **Bp;           // Bp[QV][M] -- primary sensor lead field
    static double   **Cp = NULL;    // copy of most recent Cinv
    double          tmp;            // accumulator
    int             M;              // number of channels
    int             i;              // i-index
    int             j;              // j-index
    int             k;              // k-index
    static int      rank = 0;       // primary sensor rank

    M = Header->NumPri;

    // if req'd, allocate memory
    if(rank != M) {
        // if req'd, free prior allocation
        if(rank != 0) {
            for(i=0; i<rank; i++)
                free((void *)Cinv2[i]);
            free((void *)Cinv2);
            for(i=0; i<rank; i++)
                free((void *)A[i]);
            free((void *)A);
            for(i=0; i<QV; i++)
                free((void *)Bp[i]);
            free((void *)Bp);
        }

        // allocate matrices & arrays
        if((Cinv2 = (double **)malloc((size_t)(M * sizeof(double *)))) == NULL)
            allocfailed("Cinv2[]");
        for(i=0; i<M; i++)
            if((Cinv2[i] = (double *)malloc((size_t)(M * sizeof(double)))) == NULL)
                allocfailed("Cinv2[][]");
        if((L1 = (double **)malloc((size_t)QV * sizeof(double *))) == NULL)
            allocfailed("L1[]");
        for(i=0; i<QV; i++)
            if((L1[i] = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
                allocfailed("L1[][]");
        if((L2 = (double **)malloc((size_t)QV * sizeof(double *))) == NULL)
            allocfailed("L2[]");
        for(i=0; i<QV; i++)
            if((L2[i] = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
                allocfailed("L2[][]");
        if((Vec = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
            allocfailed("Vec[]");
        if((Val = (double *)malloc((size_t)QV * sizeof(double))) == NULL)
            allocfailed("Val[]");
        if((A = (double **)malloc((size_t)(M * sizeof(double *)))) == NULL)
            allocfailed("A[]");
        for(i=0; i<M; i++)
            if((A[i] = (double *)malloc((size_t)(QV * sizeof(double)))) == NULL)
                allocfailed("A[][]");
        if((Bp = (double **)malloc((size_t)(QV * sizeof(double *)))) == NULL)
            allocfailed("Bp[]");
        for(i=0; i<QV; i++)
            if((Bp[i] = (double *)malloc((size_t)(M * sizeof(double)))) == NULL)
                allocfailed("Bp[][]");
        rank = M;
    }

    if (Cp != Cinv) {
        Cp = Cinv;
        // Cinv changed, recompute Cinv2
        for(i=0; i<M; i++)
            for(j=0; j<M; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv[i][k] * Cinv[k][j];
                Cinv2[i][j] = tmp;
            }
    }

    // compute DIP lead field matrix - compute forward solution matrix for unit dipole terms
    if(DIPLeadField3D(Voxel->p, Bp, Header, Channel, Order) == -1)
        cleanup("'DipLeadField3D()' failed");

    // linear solution of dipole moment vector
    if(Voxel->Solve == TRUE) {

        // compute L1 = L' C^(-1) L
        for(i=0; i<M; i++)                          // A = C^(-1) * L
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv[i][k] * Bp[j][k];
                A[i][j] = tmp;
            }
        for(i=0; i<QV; i++)                         // L1 = L' A
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Bp[i][k] * A[k][j];
                L1[i][j] = tmp;
            }

        // compute L2 = L' C^(-1) L
        for(i=0; i<M; i++)                          // A = C^(-2) * L
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Cinv2[i][k] * Bp[j][k];
                A[i][j] = tmp;
            }
        for(i=0; i<QV; i++)                         // L2 = L' A
            for(j=0; j<QV; j++) {
                for(k=0, tmp=0.; k<M; k++)
                    tmp += Bp[i][k] * A[k][j];
                L2[i][j] = tmp;
            }

        // solve generalized eigensystem using "TRUE"
        // TRUE solves for eigenvector of maximum eigenvalue
        // FALSE for minimum eigenvalue
        if(SolveEigen(L1, L2, Vec, Val, QV, TRUE) == 0) {

            // normalize moment to unit magnitude
            for(i=0, m2=0.; i<QV; i++)
            m2 += Vec[i] * Vec[i];
            m = sqrt(m2);
            for(i=0; i<QV; i++)     // output moment vector
                Voxel->v[i] = Vec[i] / m;
            *span = Val[0] / Val[QV-1];
        } else {                    // SolveEigen failed -- set Vec & span to zero
            for(i=0; i<QV; i++)
                Voxel->v[i] = 0.;
            *span = 0.;
        }
    } else {
        *span = 0.;
    }

    // compute dipole forward solution
    for(i=0; i<M; i++)
        for(j=0, Bt[i]=0.; j<QV; j++)
            Bt[i] += Voxel->v[j] * Bp[j][i];
}
