// Fwd2() -- compute field sensed by arbitrary gradiometer due to current dipole
//	This version features programmable integration order from 1st to 7th
//
// Author:	Stephen E. Robinson (after Roth & Sato)
//			Neuromagnetism Laboratory
//			Dept. of Neurology
//			Henry Ford Hospital
//			Detroit, MI
//			Copyright (c) 2009
//

#include <stdio.h>
#include <math.h>
#include <geoms.h>
#include <samlib.h>

double	Fwd2(
	FIELD			*Q,			// location & orientation of current dipole
	SENSOR			*Sensor,	// sensor channel (magnetometer, gradiometer) structure
	int				Order		// integration order: 1, 2, 3, 4, 5, & 7 are implemented
)

{
	double			R[3];		// vector from Sp to B
	double			R0[3];		// vector from Sp to Q
	double			A[3];		// vector from B to Q
	double			V[3];
	double			DELF[3];
	double			r2;
	double			a2;
	double			r;
	double			a;
	double			r0dr;
	double			adr;
	double			vdr;
	double			f;
	double			k0;
	double			k1;
	double			k2;
	double			k3;
	double			bc;			// magnetic flux sensed by coil
	double			bs;			// total field sensed by all coils
	double			Vx[3];		// unit x-vector
	double			Vy[3];		// unit y-vector
	double			m;			// |m|
	double			m2;			// m^2
	double			Radius;		// radius of integration point
	double			XDisp;		// x position in plane of coil
	double			YDisp;		// y position in plane of coil
	double			angle;		// angle of integration point
	static double	mu0_4pi = 1.0e-7;
	register int	c;			// coil index
	register int	i;			// integration index
	register int	v;			// vector index

	// if required, set number of integration points & coefficients
	if(Sensor->Coil[0].Evaluated == FALSE) {		// if the first coil hasn't been evaluated, the rest certainly are not!
		switch(Order) {
			case 1:
				Sensor->Coil[0].NumInt = 1;
				Sensor->Coil[0].w[0] = 1.;
				break;
			case 2:
				Sensor->Coil[0].NumInt = 3;
				Sensor->Coil[0].w[0] = 0.333333;
				Sensor->Coil[0].w[1] = 0.333333;
				Sensor->Coil[0].w[2] = 0.333333;
				break;
			case 3:
				Sensor->Coil[0].NumInt = 4;
				Sensor->Coil[0].w[0] = 0.25;
				Sensor->Coil[0].w[1] = 0.25;
				Sensor->Coil[0].w[2] = 0.25;
				Sensor->Coil[0].w[3] = 0.25;
				break;
			case 4:
				Sensor->Coil[0].NumInt = 6;
				Sensor->Coil[0].w[0] = 0.25;
				Sensor->Coil[0].w[1] = 0.25;
				Sensor->Coil[0].w[2] = 0.25;
				Sensor->Coil[0].w[3] = 0.25;
				Sensor->Coil[0].w[4] = 0.25;
				Sensor->Coil[0].w[5] = 0.15;
				break;
			case 5:
				Sensor->Coil[0].NumInt = 7;
				Sensor->Coil[0].w[0] = 0.125;
				Sensor->Coil[0].w[1] = 0.125;
				Sensor->Coil[0].w[2] = 0.125;
				Sensor->Coil[0].w[3] = 0.125;
				Sensor->Coil[0].w[4] = 0.125;
				Sensor->Coil[0].w[5] = 0.125;
				Sensor->Coil[0].w[6] = 0.25;
				break;
			case 7:		// 7th-order integration -- set 12-point integration weights
				Sensor->Coil[0].NumInt = 12;
				Sensor->Coil[0].w[0] = 0.12321;
				Sensor->Coil[0].w[1] = 0.12321;
				Sensor->Coil[0].w[2] = 0.12321;
				Sensor->Coil[0].w[3] = 0.12321;
				Sensor->Coil[0].w[4] = 0.074074;
				Sensor->Coil[0].w[5] = 0.074074;
				Sensor->Coil[0].w[6] = 0.074074;
				Sensor->Coil[0].w[7] = 0.074074;
				Sensor->Coil[0].w[8] = 0.052715;
				Sensor->Coil[0].w[9] = 0.052715;
				Sensor->Coil[0].w[10] = 0.052715;
				Sensor->Coil[0].w[11] = 0.052715;
				break;
			default:
				return 0.;
				break;
		}

		// evaluate all coil integration points
		for(c=0; c<Sensor->NumCoils; c++) {

			// compulsively make certain that the coil vector is unit
			for(v=X_, m2=0.; v<=Z_; v++)
				m2 += Sensor->Coil[c].origin.v[v] * Sensor->Coil[c].origin.v[v];
			m = sqrt(m2);
			for(v=X_; v<=Z_; v++)
				Sensor->Coil[c].origin.v[v] /= m;	// there now... I feel better already!

			// if req'd, solve for coil plane from coil vector
			if(Order > 1)
				OrthoPlane(Sensor->Coil[c].origin.v, Vx, Vy);

			// test for valid integration order & set up integration points
			switch(Order) {
				case 1:		// single point
					for(v=X_; v<=Z_; v++)
							Sensor->Coil[c].B[0].p[v] = Sensor->Coil[c].origin.p[v];
					break;

				case 2:		// three integration points
					Radius = M_SQRT1_2 * Sensor->Coil[c].radius;
					for(i=0; i<3; i++) {
						angle = (double)i * 2. * M_PI / 3.;
						XDisp = Radius * cos(angle);
						YDisp = Radius * sin(angle);
						for(v=X_; v<=Z_; v++)
							Sensor->Coil[c].B[i].p[v] = Sensor->Coil[c].origin.p[v] + Vx[v] * XDisp + Vy[v] * YDisp;
					}
					break;

				case 3:		// four integration points
					Radius = M_SQRT1_2 * Sensor->Coil[c].radius;
					for(i=0; i<4; i++) {
						angle = (double)i * M_PI / 2.;
						XDisp = Radius * cos(angle);
						YDisp = Radius * sin(angle);
						for(v=X_; v<=Z_; v++)
							Sensor->Coil[c].B[i].p[v] = Sensor->Coil[c].origin.p[v] + Vx[v] * XDisp + Vy[v] * YDisp;
					}
					break;

				case 4:		// six integration points
					Radius = 0.81650 * Sensor->Coil[c].radius;	// 1st five integration points
					for(i=0; i<5; i++) {
						angle = (double)i * 2. * M_PI / 5.;
						XDisp = Radius * cos(angle);
						YDisp = Radius * sin(angle);
						for(v=X_; v<=Z_; v++)
							Sensor->Coil[c].B[i].p[v] = Sensor->Coil[c].origin.p[v] + Vx[v] * XDisp + Vy[v] * YDisp;
					}
					for(v=X_; v<=Z_; v++)				// 6th integration point
						Sensor->Coil[c].B[i].p[v] = Sensor->Coil[c].origin.p[v];
					break;

				case 5:		// seven integration points
					Radius = 0.81650 * Sensor->Coil[c].radius;	// 1st six integration points
					for(i=0; i<6; i++) {
						angle = (double)i * M_PI / 3.;
						XDisp = Radius * cos(angle);
						YDisp = Radius * sin(angle);
						for(v=X_; v<=Z_; v++)
							Sensor->Coil[c].B[i].p[v] = Sensor->Coil[c].origin.p[v] + Vx[v] * XDisp + Vy[v] * YDisp;
					}
					for(v=X_; v<=Z_; v++)				// 7th integration point
						Sensor->Coil[c].B[6].p[v] = Sensor->Coil[c].origin.p[v];
					break;

				case 7:		// twelve integration points
					Radius = 0.45667 * Sensor->Coil[c].radius;	// 1st four integration points
					for(i=0; i<4; i++) {
						angle = ((double)i - 0.5) * M_PI / 2.;
						XDisp = Radius * cos(angle);
						YDisp = Radius * sin(angle);
						for(v=X_; v<=Z_; v++)
							Sensor->Coil[c].B[i].p[v] = Sensor->Coil[c].origin.p[v] + Vx[v] * XDisp + Vy[v] * YDisp;
					}
					Radius = 0.86603 * Sensor->Coil[c].radius;	// 2nd four integration points
					for(i=4; i<8; i++) {
						angle =(double)i * M_PI / 2.;
						XDisp = Radius * cos(angle);
						YDisp = Radius * sin(angle);
						for(v=X_; v<=Z_; v++)
							Sensor->Coil[c].B[i].p[v] = Sensor->Coil[c].origin.p[v] + Vx[v] * XDisp + Vy[v] * YDisp;
					}
					Radius = 0.91100 * Sensor->Coil[c].radius;	// 3rd four integration points
					for(i=8; i<12; i++) {
						angle = ((double)i - 0.5) * M_PI / 2.;
						XDisp = Radius * cos(angle);
						YDisp = Radius * sin(angle);
						for(v=X_; v<=Z_; v++)
							Sensor->Coil[c].B[i].p[v] = Sensor->Coil[c].origin.p[v] + Vx[v] * XDisp + Vy[v] * YDisp;
					}
					break;

				default:
					return 0.;
					break;
			}
			Sensor->Coil[c].Evaluated = TRUE;
			if(c > 0) {
				Sensor->Coil[c].NumInt = Sensor->Coil[0].NumInt;
				for(i=0; i<Sensor->Coil[c].NumInt; i++)
					Sensor->Coil[c].w[i] = Sensor->Coil[0].w[i];
			}
		}
	}

	// sum field detected by all coils
	for(c=0, bs=0.; c<Sensor->NumCoils; c++) {

		// integrate coil computed field at each integration point
		for(i=0, bc=0.; i<Sensor->Coil[c].NumInt; i++) {

			// compute vectors A, R0, & R, & scalars R0<dot>R, A<dot>R, |R|^2, & |A|^2
			for(v=X_, r2=a2=r0dr=adr=0.; v<=Z_; v++) {
				R[v] = Sensor->Coil[c].B[i].p[v] - Sensor->LocalSphere[v];
				R0[v] = Q->p[v] - Sensor->LocalSphere[v];
				A[v] = R[v] - R0[v];
				r0dr += R0[v] * R[v];
				adr += A[v] * R[v];
				r2 += R[v] * R[v];
				a2 += A[v] * A[v];
			}

			// compute V = Q<cross>R0
			V[X_] = Q->v[Y_] * R0[Z_] - Q->v[Z_] * R0[Y_];
			V[Y_] = Q->v[Z_] * R0[X_] - Q->v[X_] * R0[Z_];
			V[Z_] = Q->v[X_] * R0[Y_] - Q->v[Y_] * R0[X_];

			// compute |R|, |A|, f, & coefficients for <del>f
			r = sqrt(r2);
			a = sqrt(a2);
			f = a * (r * a + r2 - r0dr);
			k0 = adr / a + 2. * r;
			k1 = a2 / r + 2. * a + k0;
			k2 = a + k0;
			k3 = mu0_4pi / (f * f);

			// compute V<dot>R & <del>f
			for(v=X_, vdr=0.; v<=Z_; v++) {
				vdr += V[v] * R[v];
				DELF[v] = k1 * R[v] - k2 * R0[v];
			}

			// compute B
			for(v=X_; v<=Z_; v++)
				Sensor->Coil[c].B[i].v[v] = k3 * (f * V[v] - vdr * DELF[v]);
			for(v=X_; v<=Z_; v++)
				bc += Sensor->Coil[c].w[i] * Sensor->Coil[c].origin.v[v] * Sensor->Coil[c].B[i].v[v];
		}
		bs += bc * (double)Sensor->Coil[c].SenseTurns;
	}
	return bs / (double)abs(Sensor->Coil[0].SenseTurns);
}
