// GJ() -- Gauss/Jordan inverse solution
//  (Yes, this was shamelessly lifted from "Numerical Recipes in C")
//
//  Author: Stephen E. Robinson
//          Neuromagnetism Laboratory
//          Dept.of Neurology
//          Henry Ford Hospital
//          Detroit, MI
//          Copyright (c) 2007
//

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <samlib.h>

int GJ(
    double          **a,        // a[N][N] -- input (& inverse) matrix
    int             N,          // order of matrix a
    double          *b          // b[N] -- input (& solution) matrix
) {
    double          big;
    double          tmp;
    double          pivot;
    int             col = 0;    // current column -- set to silence
    int             row = 0;    // current row    -- warnings
    register int    i;
    register int    j;
    register int    k;
    register int    l;
    register int    m;
    static int      *colIndex;  // colIndex[N]
    static int      *rowIndex;  // rowIndex[N]
    static int      *pivotIndex;    // pivotIndex[N]
    static int      mem = 0;

    // allocate workspace, if needed
    if(mem != N) {
        if(mem != 0) {
            free((void *)pivotIndex);
            free((void *)rowIndex);
            free((void *)colIndex);
        }
        if((colIndex = (int *)malloc((size_t)(N * sizeof(int)))) == NULL)
            return -1;
        if((rowIndex = (int *)malloc((size_t)(N * sizeof(int)))) == NULL)
            return -1;
        if((pivotIndex = (int *)malloc((size_t)(N * sizeof(int)))) == NULL)
            return -1;
        mem = N;
    }

    // initalize pivot indices to null value
    for(j=0; j<N; j++)
        pivotIndex[j] = 0;

    // the main loop for reducing columns
    for(i=0; i<N; i++) {

        // search for pivot element
        for(j=0, big=0.; j<N; j++) {
            if(pivotIndex[j] != 1)
                for(k=0; k<N; k++) {
                    if(pivotIndex[k] == 0) {
                        if(fabs(a[j][k]) >= big) {
                            big = fabs(a[j][k]);
                            row = j;
                            col = k;
                        }
                    } else
                        if(pivotIndex[k] > 1)
                            return -1;
                }
        }
        pivotIndex[col]++;
        if(row != col) {
            for(l=0; l<N; l++)
                DSwap(&a[row][l], &a[col][l]);
            DSwap(&b[row], &b[col]);
        }
        rowIndex[i] = row;
        colIndex[i] = col;
        if(a[col][col] == 0.)
            return -1;
        pivot = a[col][col];
        a[col][col] = 1.;
        for(l=0; l<N; l++)
            a[col][l] /= pivot;
        b[col] /= pivot;
        for(m=0; m<N; m++)
            if(m != col) {
                tmp = a[m][col];
                a[m][col] = 0.;
                for(l=0; l<N; l++)
                    a[m][l] -= a[col][l] * tmp;
                b[m] -= b[col] * tmp;
            }
    }
    for(l=(N-1); l>=0; l--) {
        if(rowIndex[l] != colIndex[l]) {
            row = rowIndex[l];
            col = colIndex[l];
            for(k=0; k<N; k++)
                DSwap(&a[k][row], &a[k][col]);
        }
    }
    return 0;
}
