// GetVoxelsFS() -- populate an array of voxels using a
//  FreeSurfer atlas (from FSnormals.py & reduce_surface.py)
//
//      Author: Stephen E. Robinson
//              MEG Core Facility
//              NIMH
//

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <geoms.h>
#include <voxel.h>
#include <lfulib.h>
#include <samlib.h>

#define MAX_RAD     0.12        // 120 mm maximum voxel distance from local sphere origin
#define MIN_RAD     0.005       // 5 mm minimum voxel distance from local sphere origin
#define ETA         1.0e-12     // a small fraction of a step
#define GRID        250         // assume a maximum 250x250x250 grid

void GetVoxelsFS(
    char        *AtlasName,
    VOXELINFO   **OutVoxel,
    int         *NumVoxels,
    double      min[3],
    double      max[3],
    double      *step)
{
    static VOXELINFO    *Voxel;         // array of voxels
    double              Point[3][GRID]; // x, y, z coordinate list
    double              size[3];        // step size
    double              Vc[3];          // position vector
    double              Vn[3];          // normal vector
    double              r;              // normal length
    double              r2;             // r^2
    double              tmp;            // you know what this means, don't you?
    static int          V;              // number of voxels
    int                 N;              // number of vertices
    int                 steps[3];       // number of steps in x, y, z directions
    int                 i;              // array index
    int                 j;
    int                 k;
    int                 n;              // voxel index
    int                 v;              // vector index
    int                 dupe;           // duplicate flag
    char                fpath[256];     // path name
    FILE                *fp;            // file pointer for headshape

    // initialize Point
    for(i=0; i<GRID; i++)
        for(v=X_; v<=Z_; v++)
            Point[v][i] = 10.;      // 10 metres is well outside the head!

    // read Atlas file
    sprintf(fpath, "%s", AtlasName);
    if((fp = fopen(fpath, "r")) == NULL)
        cleanup("atlas file not found");
    if(fscanf(fp, "%d", &N) != 1)
        cleanup("error reading number of vertices");
    if(N < 1 || N > 1000000)
        cleanup("number of vertices out of range");
    for(n=0; n<N; n++) {

        // read one voxel
        if(fscanf(fp, "%lf%lf%lf%lf%lf%lf", &Vc[X_], &Vc[Y_], &Vc[Z_], &Vn[X_], &Vn[Y_], &Vn[Z_]) != 6)
            cleanup("error reading vertex");

        // parse unique coordinates
        for(v=X_; v<=Z_; v++) {
            for(i=0, dupe=FALSE; i<GRID; i++)
                if(Vc[v] == Point[v][i])     // a duplicate!
                    dupe = TRUE;
            if(dupe == FALSE)
                for(i=0; i<GRID; i++)       // unique coordinate -- add to array
                    if(Point[v][i] > 1.) {
                        Point[v][i] = Vc[v];
                        break;
                    }
        }
    }
#if DEBUG
    printf("\nread %d voxels from '%s'\n", N, AtlasName);
#endif

    // determine start & end coordinates
    for(v=X_; v<=Z_; v++)
        for(i=0, min[v]=HUGE, max[v]=(-HUGE); i<GRID; i++)
            if(Point[v][i] < 1.) {
                if(Point[v][i] < min[v])
                    min[v] = Point[v][i];
                if(Point[v][i] > max[v])
                    max[v] = Point[v][i];
            }
#if DEBUG
    printf("x: start=%f. end=%f (cm)\n", 100.*min[X_], 100.*max[X_]);
    printf("y: start=%f. end=%f (cm)\n", 100.*min[Y_], 100.*max[Y_]);
    printf("z: start=%f. end=%f (cm)\n", 100.*min[Z_], 100.*max[Z_]);
#endif
    // determine number of steps in x, y, z directions
    for(v=X_; v<=Z_; v++)
        for(i=0, steps[v]=0; i<GRID; i++)
            if(Point[v][i] < 1.)
                steps[v]++;
            else
                break;
#if DEBUG
    printf("x-steps: %d\n", steps[X_]);
    printf("y-steps: %d\n", steps[Y_]);
    printf("z-steps: %d\n", steps[Z_]);
#endif
    // determine step size
    for(v=X_, *step=0.; v<=Z_; v++) {
        size[v] = (max[v] - min[v]) / (double)(steps[v] - 1);
        *step += size[v];
    }
    *step /= 3.;    // mean step size
#if DEBUG
    printf("x stepsize: %f (cm)\n", 100.*size[X_]);
    printf("y stepssize: %f (cm)\n", 100.*size[Y_]);
    printf("z stepsize: %f (cm)\n", 100.*size[Z_]);
#endif
    // generate a rectangular array of voxels
    V = steps[X_] * steps[Y_] * steps[Z_];
#if DEBUG
    printf("allocating %d voxel array\n", V);
#endif
    if((Voxel = (VOXELINFO *)malloc((size_t)V * sizeof(VOXELINFO))) == NULL)
        allocfailed("Voxel[]");

    // populate linear voxel array with 3D coordinates
    for(i=n=0, Vc[X_]=min[X_]; i<steps[X_]; i++, Vc[X_]+=size[X_]) {
        for(j=0, Vc[Y_]=min[Y_]; j<steps[Y_]; j++, Vc[Y_]+=size[Y_]) {
            for(k=0, Vc[Z_]=min[Z_]; k<steps[Z_]; k++, Vc[Z_]+=size[Z_]) {
                for(v=X_; v<=Z_; v++) {
                    Voxel[n].p[v] = Vc[v];
                    Voxel[n].v[v] = 0.;
                }
                Voxel[n].Solve = FALSE;
                Voxel[n].ROI = FALSE;
                n++;
            }
        }
    }

    // re-read atlas list & put vertex normals in correct indices
    rewind(fp);
    if(fscanf(fp, "%d", &N) != 1)
        cleanup("error reading number of vertices");

    // for each voxel...
    for(n=0; n<N; n++) {

        // read one voxel
        if(fscanf(fp, "%lf%lf%lf%lf%lf%lf", &Vc[X_], &Vc[Y_], &Vc[Z_], &Vn[X_], &Vn[Y_], &Vn[Z_]) != 6)
            cleanup("error reading vertex");

        // determine index for this coordinate
        for(i=0; i<V; i++) {

            // enforce unit normal vector
            for(v=X_, r2=0.; v<=Z_; v++)
                r2 += Vn[v] * Vn[v];
            r = sqrt(r2);
            if(r > 0.) {
                for(v=X_; v<=Z_; v++)
                    Vn[v] = Vn[v] / r;
            }

            // compute distance from ith voxel to this coordinate
            for(v=X_, r2=0.; v<=Z_; v++) {
                tmp = Vc[v] - Voxel[i].p[v];
                r2 += tmp * tmp;
            }
            r = sqrt(r2);
#if DEBUG
    printf("distance of atlas voxel %d to array voxel %d = %f (cm)\n", n, i, 100.*r);
#endif
            // is this our voxel?
            if(r < 0.001) {             // 1 mm error permitted
                for(v=X_; v<=Z_; v++)
                    Voxel[i].v[v] = Vn[v];
                Voxel[i].Solve = FALSE;
                Voxel[i].ROI = TRUE;
#if DEBUG
    printf("found index %d\n", i);
#endif
                break;
            }
        }
    }
    fclose(fp);
#if DEBUG
    for(i=n=0; i<V; i++)
        if(Voxel[i].ROI == TRUE)
            n++;
    printf("inserted %d voxels into ROI array\n", n);
#endif
    // output
    *OutVoxel = Voxel;
    *NumVoxels = V;
}
