// SAMsolve2() -- all-in-one SAM solution
//  This version uses Cholesky decomposition of covariance for inverse
//  solution.
//
//  Author: Stephen E. Robinson (after Sarvas & Hamalainen)
//      MEG Core Facility
//      NIMH
//

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "samutil.h"    // for allocfailed() and cleanup()

void    SAMsolve2(
    double  **C,        // C[M][M] -- unaugmented covariance matrix (input)
    double  *Bp,        // Bp[M] -- forward solution vector (input)
    double  *w,         // w[M] -- SAM weights (output)
    double  noise,      // mean sensor noise (input)
    int     rank,       // rank (M) (input)
    double  *s2,        // SAM source power (output)
    double  *n2         // SAM noise power (output)
) {
    static double   **A;    // A[M][M] -- C^(-1)
    static double   *p;     // p[M]
    static double   **Cp = NULL;
    register double sum;
    register int    i;
    register int    j;
    register int    k;
    static int      M = 0;

    // if req'd, allocate Cinv unless M changes
    if(M == 0) {
        M = rank;
        if((A = (double **)malloc((size_t)(M * sizeof(double *)))) == NULL)
            allocfailed("A[]");
        for(i=0; i<M; i++)
            if((A[i] = (double *)malloc((size_t)(M * sizeof(double)))) == NULL)
                allocfailed("A[][]");
        if((p = (double *)malloc((size_t)(M * sizeof(double)))) == NULL)
            allocfailed("p[]");
    }

    // if req'd, invert
    if(Cp != C) {

        // copy C to A
        for(i=0; i<M; i++)
            for(j=0; j<M; j++)
                A[i][j] = C[i][j];

        // Cholesky decomposition of A = L L'
        for(i=0; i<M; i++) {
            for(j=i; j<M; j++) {
                for(k=i-1, sum=A[i][j]; k>=0; k--)
                    sum -= A[i][k] * A[j][k];
                if(i == j) {
                    if(sum <= 0.)
                        cleanup("Cholesky decomposition failed");
                    p[i] = sqrt(sum);
                } else
                    A[j][i] = sum / p[i];
            }
        }

        // update pointer
        Cp = C;
    }

    // solve A x = b (i.e., C w = Bp)
    for(i=0; i<M; i++)          // zero w
        w[i] = 0.;
    for(i=0; i<M; i++) {
        for(k=(i-1), sum=Bp[i]; k>=0; k--)
            sum -= A[i][k] * w[k];
        w[i] = sum / p[i];
    }
    for(i=(M-1); i>=0; i--) {
        for(k=(i+1), sum=w[i]; k<M; k++)
            sum -= A[k][i] * w[k];
        w[i] = sum / p[i];
    }

    // solve source power & noise power
    for(i=0, sum=0.; i<M; i++)
        sum += Bp[i] * w[i];
    for(i=0, *n2=0.; i<M; i++) {        // noise power
        w[i] /= sum;
        *n2 += noise * w[i] * w[i];
    }
    if(C == NULL) {
        *s2 = 1. / sum;
    } else {
        for(i=0, *s2=0.; i<M; i++)
            for(j=0; j<M; j++)
                *s2 += w[i] * w[j] * C[i][j];
    }
}
