// pinv() -- Moore-Penrose Pseudo Inverse by Singular Value Decomposition
//
//      MEG Core Facility
//      NIMH
//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <geoms.h>
#include "samlib.h"
#include "samutil.h"

extern void svd(SVD *);

static double *d;

static int cmp(const void *a, const void *b)
{
    if (d[*(int *)a] < d[*(int *)b]) {
        return -1;
    }
    return 1;
}

static int *argsort(double *a, int n)
{
    int i;
    int *idx;

    idx = new_arrayE(int, n, "argsort array");
    for (i = 0; i < n; i++) {
        idx[i] = i;
    }

    d = a;
    qsort(idx, n, sizeof(int), cmp);

    return idx;
}

/* Set tol to -1 for automatic. */

void pinv(
    double **c,
    int m,
    int ndim
) {
    int i, j, k;
    int *idx;
    double tol;
    SVD *s;

    s = newE(SVD, "SVD structure");

    s->U = new_matrixE(m, m, "input matrix");
    s->W = new_arrayE(double, m, "singular values");
    s->V = new_matrixE(m, m, "orthogonal matrix");
    s->M = m;
    s->N = m;

    /* Input goes in U. */

    for (i = 0; i < m; i++) {
        for (j = 0; j < m; j++) {
            s->U[i][j] = c[i][j];
        }
    }

    /* SVD. */

    svd(s);

    /* Find the n'th smallest singular value. */

    idx = argsort(s->W, m);
    tol = s->W[idx[ndim - 1]];
    msg("[pinv] truncating at %g\n", tol);

    /* The pseudoinverse is now V W^-1 U^T. */

    for (i = 0; i < m; i++) {
        if (s->W[i] > tol) {
            s->W[i] = 1. / s->W[i];
        } else {
            s->W[i] = 0.;
        }
    }
    free(idx);

    /* (i x j) = (i x k)(k x j)^T */

    for (i = 0; i < m; i++) {
        for (j = 0; j < m; j++) {
            c[i][j] = 0.;
            for (k = 0; k < m; k++) {
                c[i][j] += s->V[i][k] * s->W[k] * s->U[j][k];
            }
        }
    }

    free_matrix(s->V);
    free(s->W);
    free_matrix(s->U);
    free(s);
}
