// svd() -- singular value decomposition
//
//	Author:	Stephen E. Robinson (after Sarvas & Hamalainen)
//		MEG Core Facility
//		NIMH
//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <geoms.h>
#include <lfulib.h>

#define MAXITS	1000

static double at, bt, ct;
#define PYTHAG(a,b) ((at=fabs(a)) > (bt=fabs(b)) ? (ct=bt/at,at*sqrt(1.0+ct*ct)) : (bt ? (ct=at/bt,bt*sqrt(1.0+ct*ct)): 0.0));

static double maxarg1,maxarg2;
#define MAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ? (maxarg1) : (maxarg2))

#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

void	svd(
	SVD	*sp
) {
	static double	*rv1;
	double			norm;
	double			c;
	double			f;
	double			g;
	double			h;
	double			s;
	double			scale;
	double			x;
	double			y;
	double			z;
	double			max;
	double			min;
	int				l;
	int				nm;
	int				M;
	int				N;
	int				flag;
	register int	i;
	register int	its;
	register int	j;
	register int	k;
	register int	jj;
	static int		mem = 0;

	// create local row/column sizes & test
	M = sp->M;
	N = sp->N;
	if(M < N)	// rows, M must be equal or greater than columns, N
		cleanup("M must be >= N for SVD");

	// allocate memory for temporary array
	if(mem < N) {
		if(mem != 0)
			free((void *)rv1);
		if((rv1 = (double *)malloc((size_t)N * sizeof(double))) == NULL)
			allocfailed("rv1[]");
		mem = N;
	}

	// always initialize [V] & [W]
	for(i=0; i<N; i++) {
		for(j=0; j<N; j++)
			sp->V[i][j] = 0.;
		sp->W[i] = 0.;
	}

// (1) Householder reduction to bi-diagonal form
	for(i=0, g=scale=norm=0.; i<N; i++) {
		l = i + 1;
		rv1[i] = scale * g;
		g = s = scale = 0.;
		if(i < M) {
			for(k=i; k<M; k++)
				scale += fabs(sp->U[k][i]);
			if(scale) {
				for(k=i; k<M; k++) {
					sp->U[k][i] /= scale;
					s += sp->U[k][i] * sp->U[k][i];
				}
				f = sp->U[i][i];
				g = -SIGN(sqrt(s), f);
				h = f * g - s;
				sp->U[i][i] = f - g;
				if(i != (N-1))
					for(j=l; j<N; j++) {
						for(k=i, s=0.; k<M; k++)
							s += sp->U[k][i] * sp->U[k][j];
						f = s / h;
						for(k=i; k<M; k++)
							sp->U[k][j] += f * sp->U[k][i];
					}
				for(k=i; k<M; k++)
					sp->U[k][i] *= scale;
			}
		}
		sp->W[i] = scale * g;
		g = s = scale = 0.;
		if(i < M && i != (N-1)) {
			for(k=l; k<N; k++)
				scale += fabs(sp->U[i][k]);
			if(scale) {
				for(k=l; k<N; k++) {
					sp->U[i][k] /= scale;
					s += sp->U[i][k] * sp->U[i][k];
				}
				f = sp->U[i][l];
				g = -SIGN(sqrt(s), f);
				h = f * g - s;
				sp->U[i][l] = f - g;
				for(k=l; k<N; k++)
					rv1[k] = sp->U[i][k] / h;
				if(i != (M-1)) {
					for(j=l; j<M; j++) {
						for(k=l, s=0.; k<N; k++)
							s += sp->U[j][k] * sp->U[i][k];
						for(k=l; k<N; k++)
							sp->U[j][k] += s * rv1[k];
					}
				}
				for(k=l; k<N; k++)
					sp->U[i][k] *= scale;
			}
		}
		norm = MAX(norm, (fabs(sp->W[i])+fabs(rv1[i])));
	}

// (2) accumulation of right-hand transformations
	for(i=(N-1); i>=0; i--) {
		if(i < (N-1)) {
			if(g) {
				for(j=l; j<N; j++)	// double division to avoid possible underflow
					sp->V[j][i] = (sp->U[i][j] / sp->U[i][l]) / g;
				for(j=l; j<N; j++) {
					for(k=l, s=0.; k<N; k++)
						s += sp->U[i][k] * sp->V[k][j];
					for(k=l; k<N; k++)
						sp->V[k][j] += s * sp->V[k][i];
				}
			}
			for(j=l; j<N; j++)
				sp->V[i][j] = sp->V[j][i] = 0.;
		}
		sp->V[i][i] = 1.;
		g = rv1[i];
		l = i;
	}

// (3) accumulation of left-hand transformations
	for(i=(N-1); i>=0; i--) {
		l = i + 1;
		g = sp->W[i];
		if(i < (N-1))
			for(j=l; j<N; j++)
				sp->U[i][j] = 0.;
		if(g) {
			g = 1. / g;
			if(i != (N-1))
				for(j=l; j<N; j++) {
					for(k=l, s=0.; k<M; k++)
						s += sp->U[k][i] * sp->U[k][j];
					f = (s / sp->U[i][i]) * g;
					for(k=i; k<M; k++)
						sp->U[k][j] += f * sp->U[k][i];
				}
			for(j=i; j<M; j++)
				sp->U[j][i] *= g;
		} else
			for(j=i; j<M; j++)
				sp->U[j][i] = 0.;
		sp->U[i][i] += 1.;
	}

// (4) diagonalization of the bidiagonal form
	for(k=(N-1); k>=0; k--) {			// loop over singular values
		for(its=0; its<MAXITS; its++) {		// loop over allowed iterations
			for(l=k, flag=1; l>=0; l--) {	// test for splitting
				nm = l - 1;		// note that rv1[0] always zero
				if((fabs(rv1[l]) + norm) == norm) {
					flag = 0;
					break;
				}
				if(nm >= 0)
			//	if(nm > 0)
					if((fabs(sp->W[nm]) + norm) == norm)
						break;
			}
			if(flag == 1 && nm >= 0) {
				for(i=l, c=0., s=1.; i<=k; i++) {	// cancellation of rv1[l], if l>1
					f = s * rv1[i];
					if((fabs(f) + norm) != norm) {
						g = sp->W[i];
						h = PYTHAG(f, g);
						sp->W[i] = h;
						h = 1. / h;
						c = g * h;
						s = (-f * h);
						for(j=0; j<M; j++) {
							y = sp->U[j][nm];
							z = sp->U[j][i];
							sp->U[j][nm] = y * c + z * s;
							sp->U[j][i] = z * c - y * s;
						}
					}
				}
			}
			z = sp->W[k];
			if(l == k) {			// convergence
				if(z < 0.) {		// singular value is made nonnegative
					sp->W[k] = -z;
					for(j=0; j<N; j++)
						sp->V[j][k] = (-sp->V[j][k]);
				}
				break;
			}
			if(its == (MAXITS-1))
				cleanup("maximum iterations exceeded in SVD");
			x = sp->W[l];			// shift from bottom 2-by-2 minor
			nm = k - 1;
			y = sp->W[nm];
			g = rv1[nm];
			h = rv1[k];
			f = ((y - z) * (y + z) + (g - h) * (g + h)) / (2. * h * y);
			g = PYTHAG(f, 1.);
			f = ((x - z) * (x + z) + h * ((y / (f + SIGN(g, f))) - h)) / x;

// (5) next QR transformation
			for(j=l, c=s=1.; j<=nm; j++) {
				i = j + 1;
				g = rv1[i];
				y = sp->W[i];
				h = s * g;
				g *= c;
				rv1[j] = z = PYTHAG(f, h);
				c = f / z;
				s = h / z;
				f = x * c + g * s;
				g = g * c - x * s;
				h = y * s;
				y *= c;
				for(jj=0; jj<N; jj++) {
					x = sp->V[jj][j];
					z = sp->V[jj][i];
					sp->V[jj][j] = x * c + z * s;
					sp->V[jj][i] = z * c - x * s;
				}
				sp->W[j] = z = PYTHAG(f, h);	// rotation can be arbitrary if z=0
				if(z) {
					z = 1. / z;
					c = f * z;
					s = h * z;
				}
				f = (c * g) + (s * y);
				x = (c * y) - (s * g);
				for(jj=0; jj<M; jj++) {
					y = sp->U[jj][j];
					z = sp->U[jj][i];
					sp->U[jj][j] = y * c + z * s;
					sp->U[jj][i] = z * c - y * s;
				}
			}
			rv1[l] = 0.;
			rv1[k] = f;
			sp->W[k] = x;
		}
	}

	// find minimum & maximum singular values
	for(i=0, max=0., min=1.0e+100; i<N; i++) {
		if(sp->W[i] > max)
			max = sp->W[i];
		if(sp->W[i] < min)
			min = sp->W[i];
	}
	sp->span = max / min;
	sp->noise = min;
}
