// SolveEigen() -- solve the generalized eigensystem
//  of two N x N positive-definite matrices, using
//  Cholesky decomposition
//
//  Cholesky decomposition "borrowed" from "Numerical Recipes in C"
//
//  Author: Stephen E. Robinson
//      MEG Core Facility
//      NIMH
//

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <siglib.h>
#include <lfulib.h>

int SolveEigen(
    double  **A,                // A[N][N] -- positive-definite matrix
    double  **B,                // B[N][N] -- positive-definite matrix
    double  *Vec,               // Vec[N] -- eigenvector for largest/smallest eigenvalue
    double  *Val,               // Val[N] -- eigenvalues
    int     N,                  // rank of generalized eigenproblem
    int     MinMax              // TRUE=maximum, FALSE=minimum
) {
    static double   **Linv;     // Linv[N][N] -- L^-1
    static double   **C;        // C[N][N] -- regularized matrix C = L' A L
    static double   **D;        // D[N][N] -- intermediate matrix
    static double   *p;         // p[N] -- diagonal vector
    double          sum;
    double          tmp;
    int             i;
    int             j;
    int             k;
    int             index;      // index of largest/smallest eigenvalue
    static int      rank = 0;   // rank of solution

    // if req'd, allocate memory
    if(rank != N) {
        if(rank != 0) {
            free((void *)p);
            for(i=0; i<rank; i++)
                free((void *)D[i]);
            free((void *)D);
            for(i=0; i<rank; i++)
                free((void *)C[i]);
            free((void *)C);
            for(i=0; i<rank; i++)
                free((void *)Linv[i]);
            free((void *)Linv);
        }
        rank = N;
        if((Linv = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
            allocfailed("Linv[]");
        for(i=0; i<N; i++)
            if((Linv[i] = (double *)malloc((size_t)N * sizeof(double))) == NULL)
                allocfailed("Linv[][]");
        if((C = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
            allocfailed("C[]");
        for(i=0; i<N; i++)
            if((C[i] = (double *)malloc((size_t)N * sizeof(double))) == NULL)
                allocfailed("C[][]");
        if((D = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
            allocfailed("D[]");
        for(i=0; i<N; i++)
            if((D[i] = (double *)malloc((size_t)N * sizeof(double))) == NULL)
                allocfailed("D[][]");
        if((p = (double *)malloc((size_t)N * sizeof(double))) == NULL)
            allocfailed("p[]");
    }

    // compute L for Cholesky decomposition of matrix B (B contains L on completion)
    for(i=0; i<N; i++) {
        for(j=i; j<N; j++) {
            for(k=i-1, sum=B[i][j]; k>=0; k--)
                sum -= B[i][k] * B[j][k];
            if(i == j) {
                if(sum <= 0.)
                    return -1;      // Cholesky decomposition failed
                p[i] = sqrt(sum);
            } else {
                B[j][i] = sum / p[i];
            }
        }
    }

    // compute L^-1 -- on completion, lower triangle of B has L^-1
    for(i=0; i<N; i++) {
        B[i][i] = 1. / p[i];        // diagonal inverse elements
        for(j=i+1; j<N; j++) {
            for(k=i, sum=0.; k<j; k++)
                sum -= B[j][k] * B[k][i];
            B[j][i] = sum / p[j];   // off diagonal inverse elements
        }
    }

    // save L^-1 with zero upper triangle
    for(i=0; i<N; i++)
        for(j=0; j<N; j++)
            if(j >= i)
                Linv[j][i] = B[j][i];
            else
                Linv[j][i] = 0.;

    // compute C = L^-1 A L-1'
    for(i=0; i<N; i++)              // compute D = L^-1 A
        for(j=0; j<N; j++) {
            for(k=0, sum=0.; k<N; k++)
                sum += Linv[i][k] * A[k][j];
            D[i][j] = sum;
        }
    for(i=0; i<N; i++)              // compute C = D * L^-1'
        for(j=0; j<N; j++) {
            for(k=0, sum=0.; k<N; k++)
                sum += D[i][k] * Linv[j][k];
            C[i][j] = sum;
        }

    // solve eigendecomposition of C (on output, C holds the eigenvectors of C, & Val are the eigenvalues)
    eigen(C, Val, N);

    // solve eigenvectors -- D = L' C
    for(i=0; i<N; i++)
        for(j=0; j<N; j++) {
            for(k=0, tmp=0.; k<N; k++)
                tmp += Linv[k][i] * C[k][j];
            D[i][j] = tmp;
        }

    // normalize eigenvectors
    for(i=0; i<N; i++) {
        for(j=0, sum=0.; j<N; j++)
            sum += D[j][i] * D[j][i];
        tmp = sqrt(sum);
        for(j=0; j<N; j++)
            D[j][i] /= tmp;
    }

    // find the index of the largest/smallest eigenvalue
    index = -1;
    if(MinMax == TRUE) {    // find index of largest eigenvalue
        for(i=0, tmp=0.; i<N; i++) {
            if(Val[i] > tmp) {
                tmp = Val[i];
                index = i;
            }
        }
    } else {                // find index of smallest eigenvalue
        for(i=0, tmp=HUGE; i<N; i++) {
            if(Val[i] < tmp) {
                tmp = Val[i];
                index = i;
            }
        }
    }

    // output indexed eigenvector
    for(i=0; i<N; i++)
        Vec[i] = D[i][index];

    return 0;       // normal return value
}
