// eigen3() -- eigendecomposition of real, symmetrical  3x3 matrix
//	(shamelessly kludged together from Numerical Recipes
//	routines: 'tred2()' & 'tqli()')
//
//	Author:	Stephen E. Robinson
//			Neuromagnetism Laboratory
//			Dept. of Neurology
//			Henry Ford Hospital
//			Detroit, MI
//			Copyright (c) 2009
//

#include <stdlib.h>
#include <math.h>

#define N	3
#define SIGN(a,b) ((b)<0. ? -fabs(a): fabs(a))

int	eigen3(
	double		A[3][3],	// input matrix & output eigenvectors
	double		B[3]		// B[N] -- eigenvalues
)

{
	double		C[3];		// off-diagonal elements
	double		b;
	double		c;
	double		dd;
	double		f;
	double		g;
	double		h;
	double		hh;
	double		p;
	double		r;
	double		s;
	double		scale;
	int			i;
	int			iter;
	int			j;
	int			k;
	int			l;
	int			m;

	// begin 'tred2()'
	for(i=(N-1); i>=1; i--) {
		l = i - 1;
		h = scale = 0.;
		if(l > 0) {
			for(k=0; k<=l; k++)
				scale += fabs(A[i][k]);
			if(scale == 0.)						// skip transformation
				C[i] = A[i][l];
			else {
				for(k=0; k<=l; k++) {
					A[i][k] /= scale;			// use scaled A's for transformation
					h += A[i][k] * A[i][k];		// form sigma in h
				}
				f = A[i][l];
				g = (f > 0.) ? -sqrt(h): sqrt(h);
				C[i] = scale * g;
				h -= f * g;						// now h is equation (11.2.4)
				A[i][l] = f - g;				// store u in the ith row of A
				for(j=0, f=0.; j<=l; j++) {
					A[j][i] = A[i][j] / h;		// store u/H in ith column of A
					for(k=0, g=0.; k<=j; k++)
						g += A[j][k] * A[i][k];
					for(k=j+1; k<=l; k++)
						g += A[k][j] * A[i][k];
					C[j] = g / h;				// form element of p in temporary unused element of E
					f += C[j] * A[i][j];
				}
				hh = f / (h + h);				// form kappa, equation (11.2.11)
				for(j=0; j<=l; j++) {			// form q & store in E overwriting p
					f = A[i][j];				// note that C[l] = C[i-1] survives
					C[j] = g = C[j] - hh * f;
					for(k=0; k<=j; k++)			// reduce A, equation (11.2.13)
						A[j][k] -= (f * C[k] + g * A[i][k]);
				}
			}
		} else
			C[i] = A[i][l];
		B[i] = h;
	}
	for(i=0, B[0]=C[0]=0.; i<N; i++) {			// begin accumulation of transformation matrix
		l = i - 1;
		if(B[i]) {								// this block skipped when i=0
			for(j=0; j<=l; j++) {
				for(k=0, g=0.; k<=l; k++)		// use u & u/H stored in A to for P.Q
					g += A[i][k] * A[k][j];
				for(k=0; k<=l; k++)
					A[k][j] -= g * A[k][i];
			}
		}
		B[i] = A[i][i];
		A[i][i] = 1.;							// reset row & column of A to identity matrix for next iteration
		for(j=0; j<=l; j++)
			A[j][i] = A[i][j] = 0.;
	}

	// begin 'tqli()'
	for(i=1; i<N; i++)
		C[i-1] = C[i];							// convenient to renumber elements of E
	for(l=0, C[N-1]=0.; l<N; l++) {
		iter = 0;
		do {
			for(m=l; m<(N-1); m++) {			// look for single small subdiagonal element to split matrix
				dd = fabs(B[m]) + fabs(B[m+1]);
				if(fabs(C[m])+dd == dd)
					break;
			}
			if(m != l) {
				if(iter++ == 500)
					return -1;
				g = (B[l+1] - B[l]) / (2. * C[l]);			// form shift
				r = sqrt(g * g + 1.);
				g = B[m] - B[l] + C[l] / (g + SIGN(r, g));	// this dm - ks
				for(i=(m-1), s=c=1., p=0.; i>=l; i--) {		// a plane rotation as in the original QL, followed by 
					f = s * C[i];							// Givens rotations to restore tridiagonal form
					b = c * C[i];
					if(fabs(f) >= fabs(g)) {
						c = g / f;
						r = sqrt(c * c + 1.);
						C[i+1] = f * r;
						c *= (s = 1. / r);
					} else {
						s = f / g;
						r = sqrt(s * s + 1.);
						C[i+1] = g * r;
						s *= (c = 1. / r);
					}
					g = B[i+1] - p;
					r = (B[i] - g) * s + 2. * c * b;
					p = s * r;
					B[i+1] = g + p;
					g = c * r - b;
					for(k=0; k<N; k++) {					// form eigenvectors
						f = A[k][i+1];
						A[k][i+1] = s * A[k][i] + c * f;
						A[k][i] = c * A[k][i] - s * f;
					}
				}
				B[l] = B[l] - p;
				C[l] = g;
				C[m] = 0.;
			}
		} while(m != l);
	}
	return 0;
}
