// equilizer() -- equalizes power/amplitude at all frequencies in specified bandpass
//
//  (1) compute FFT of signal
//  (2) compute the spectrum sqrt(re^2 + im^2) -- signal, not power
//  (3) solve for mean signal between highpass & lowpass frequencies
//  (4) normalize spectrum at all frequency by the mean & multiply by filter
//          to prevent amplification of signals outside the bandpass
//  (5) compute inverse FFT & output real part of result
//
// Author:  Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fftw3.h>
#include <siglib.h>


void    equalizer(
    double      *data,      // data[N] -- input data
    double      *result,    // result[N] -- output data (equilized)
    FFTSPEC     *filter,    // FFT filter structure
    double      fhi,        // highpass frequency
    double      flo,        // lowpass frequency
    double      srate,      // sample rate
    int         len         // number of data samples
)
{
    static fftw_plan    p1;             // forward fft plan
    static fftw_plan    p2;             // reverse fft plan
    static fftw_complex *x;             // real (only real part used)
    static fftw_complex *X;             // complex
    static double       *gain;          // equilizer gain vs frequency
    double              *p;             // pointer
    double              mean;           // mean power
    int                 i;              // i-index
    int                 j;              // j-index
    static int          bhi;            // highpass bin number
    static int          blo;            // lowpass bin number
    static int          planlen = 0;    // saved plan length for initialization
    char *Wisfile = NULL;               // wisdom file name
    FILE                *wisdom;        // wisdom file pointer

    // Keep the arrays and plans around from last time, since this
    // is a very common case. Reallocate them if they change.
    if (len != planlen && planlen > 0) {
        fftw_destroy_plan(p1);
        fftw_destroy_plan(p2);
        fftw_free(x);
        fftw_free(X);
        fftw_free(gain);
        planlen = 0;
    }
    if (planlen == 0) {
        planlen = len;
        x = fftw_malloc(sizeof(fftw_complex) * len);
        X = fftw_malloc(sizeof(fftw_complex) * len);
        gain = fftw_malloc(sizeof(double) * (size_t)len);

        // Get any accumulated wisdom.
        set_wisfile();
        wisdom = fopen(Wisfile, "r");
        if(wisdom) {
            fftw_import_wisdom_from_file(wisdom);
            fclose(wisdom);
        }

        // Set up the fftw plans.
        p1 = fftw_plan_dft_1d(len, x, X, FFTW_FORWARD, FFTW_MEASURE);
        p2 = fftw_plan_dft_1d(len, X, x, FFTW_BACKWARD, FFTW_MEASURE);

        // Save the wisdom.
        wisdom = fopen(Wisfile, "w");
        if(wisdom) {
            fftw_export_wisdom_to_file(wisdom);
            fclose(wisdom);
        }

        // get bin number limits for computing mean spectrum
        bhi = (int)floor(fhi * len / srate + .5);
        blo = (int)floor(flo * len / srate + .5);
    }

    // Convert the input to complex.
    memset(x, 0, sizeof(fftw_complex) * len);
    for(i=0; i<len; i++)
        x[i][0] = data[i];

    // FFT.
    fftw_execute(p1);   // x -> X

    // compute amplitude vs frequency over the entire signal and its mean within bandpass
    mean = 0.;
    for(i=0, j=len-1; i<len/2; i++, j--) {
        gain[i] = sqrt(X[i][0] * X[i][0] + X[i][1] * X[i][1]);
        gain[j] = sqrt(X[j][0] * X[j][0] + X[j][1] * X[j][1]);
        if(i >= bhi && i <= blo)
            mean += gain[i];
    }
    mean /= (double)(len/2);            // mean signal amplitude between fhi & flo

    // multiply spectral coefficients by equilizer gain -- careful to include positive & negative frequencies!
    for(i=0, j=len-1; i<len/2; i++, j--) {
        X[i][0] *= filter->Gain[i] * gain[i] / mean;        // positive frequency real
        X[i][1] *= filter->Gain[i] * gain[i] / mean;        // positive frequency imaginary
        X[j][0] *= filter->Gain[j] * gain[j] / mean;        // negative frequency real
        X[j][1] *= filter->Gain[j] * gain[j] / mean;        // negative frequency imaginary
    }

    // Inverse FFT.
    fftw_execute(p2);   // X -> x

    // Fill in the rows of the result.
    for(i=0, p=result; i<len; i++)
        *p++ = x[i][0] / (double)len;
}
