// whiten() -- use eigendecomposition of covariance to whiten data
//
//  Author: Stephen E. Robinson
//      MEG Core Facility
//      NIMH
//

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <samlib.h>
#include <siglib.h>
#include <lfulib.h>

void whiten(
    double  **x,            // x[N][T] -- input data
    double  **y,            // y[N][T] -- output (whitened) data
    int     N,              // matrix rank
    int     *K,             // effective data rank
    int     T               // number of samples
)
{
    static double   **C;        // C[N][N] -- covariance matrix (and eigenvectors)
    static double   **S;        // S[N][N] -- sphering matrix
    static double   *L;         // L[N] -- eigenvalues
    static double   *SortL;     // SortL[N] -- sorted eigenvalues
    double          delta;
    double          min;
    long double     nt;         // number of samples
    long double     lsum;       // integrator
    double          tmp;        // what else could this be?
    int             i;
    int             j;
    int             k;
    int             t;          // sample index
    static int      rank = 0;   // remember matrix rank

    // allocate diagonal vector
    if(rank == 0) {
        if((C = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
            allocfailed("C[]");
        for(i=0; i<N; i++)
            if((C[i] = (double *)malloc((size_t)N * sizeof(double))) == NULL)
                allocfailed("C[][]");
        if((S = (double **)malloc((size_t)N * sizeof(double *))) == NULL)
            allocfailed("S[]");
        for(i=0; i<N; i++)
            if((S[i] = (double *)malloc((size_t)N * sizeof(double))) == NULL)
                allocfailed("S[][]");
        if((L = (double *)malloc((size_t)N * sizeof(double))) == NULL)
            allocfailed("L[]");
        if((SortL = (double *)malloc((size_t)N * sizeof(double))) == NULL)
            allocfailed("SortL[]");
        rank = N;
    }
    nt = (long double)T;

    // compute covariance of x
    for(i=0; i<N; i++)
        for(j=0; j<N; j++) {
            for(t=0, lsum=0.; t<T; t++)
                lsum += (long double)(x[i][t] * x[j][t]);
            C[i][j] = (double)(lsum / nt);
        }

    // compute eigendecomposition of covariance
    eigenl(C, L, N);            // notes: using a long double eigendecomposition; the eigenvectors placed in C are column vectors!

    // sort the eigenvalues in descending order
    for(i=0; i<N; i++)
        SortL[i] = L[i];
    Sort(SortL, N);

    // look at derivatives - find K for minimum derivative
    for(k=j=N-3, min=HUGE; k>=2; k--) {
        delta = SortL[k - 2] - SortL[k + 2];
        if(delta < min) {
            min = delta;
            j = k;
        }
    }
    *K = j;     // output effective rank

    // compute sphering matrix
    for(i=0; i<N; i++) {
        tmp = sqrt(L[i]);
        for(j=0; j<N; j++)
            S[i][j] = (double)C[j][i] / tmp;
    }

    // compute y(t) = L^(-1) x(t)
    for(t=0; t<T; t++) {
        for(i=0; i<N; i++) {
            for(j=0, lsum=0.; j<N; j++)
                lsum += (long double)(S[i][j] * x[j][t]);
            y[i][t] = (double)lsum;
// printf("x[%d][%d]=%e, y[%d][%d]=%e\n", i, t, x[i][t], i, t, y[i][t]);
        }
    }
}
