/* fastica -- independent components analysis, using algorithms developed at
 *	Helsinki University of Technology, by Hyvarinen
 *
 *	command line flags:
 *		-r <run name>	-- MEG run name
 *		-n <rank>	-- number of independent components
 *		-f <hp, lp>	-- high & lowpass frequencies
 *		-g <function>	-- nonlinear function used for ICA
 *		-p <precision>	-- number of digits for convergence
 *		-b <order>	-- baseline removal order
 *		-d		-- delete backup meg4 file (optional)
 *		-v		-- verbose
 *
 * Author: Stephen E. Robinson
 * Copyright (c) CTF Systems Inc., 2001.  All Rights Reserved.
 * Commercially Confidential Information
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <float.h>
#include <byteswap.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <model.h>
#include <samlib.h>
#include <siglib.h>
#include <SAMfiles.h>
#include <DataFiles.h>
#include <samUtils.h>
#include <CTF_Revision.h>

#define PRG_REV		"-1.7"
#define TRUE		1
#define FALSE		0
#define IIRORDER	3
#define MAXITS		250			// maximum number of iterations
#define MAXRETRIES	20			// maximum number of retries before terminating search
#define MAXGUESS	20			// maximum number of successive guesses with declining dot product tolerance
#define DIGITS		10			// default number of digits of precision
#define SCALE		100000.			// scale factor for saving time-series as 32-bit integer data for DataEditor
#define BASELINE	1			// remove 1st-order mean (trend)
#define NOISE		1.0e-28			// 10 fT/rt-Hz sensor noise

main(
	int	argc,
	char	**argv
)

{
	IIRSPEC			HPFilter;		// highpass IIR section
	IIRSPEC			LPFilter;		// lowpass IIR section
	HeaderInfo		Header;			// header information data
	ChannelInfo		*Channel;		// Channel[M] -- channel information data
	EpochInfo		*Epoch;			// Epoch[E] -- epoch information data
	long			*xint;			// xint[T] -- single channel, one-trial time-series data
	double			**x;			// x[K][E*T] -- sphered data time-series data
	double			**y;			// y[K][T] -- ICA output time-series data
	long double		**C;			// C[M][M] -- active-state covariance matrix
	double			**S;			// S[M][M] -- sphering matrix
	double			**A;			// A[M][M] -- resultant of BB'
	double			**B;			// B[M][M] -- mixing matrix
	double			*u;				// u[M] -- updated weight vector
	double			*v;				// v[M] -- intermediate weight vector
	double			*w;				// w[M] -- weight vector
	double			*hypTan;		// hypTan[TT] - hyperbolic tangent time-series
	double			*In;			// In[T] -- single-channel time-series
	double			*Out;			// Out[T] -- single-channel time-series
	long double		*L;				// L[M] -- eigenvalues
	double			sum;			// summation of tanh() transformed data
	double			r;
	double			r2;
	double			rold;
	long double		nt;				// number of samples (long double)
	double			ntt;			// total samples (double)
	long double		ne;				// number of epochs (long double)
	double			Gauss;
	double			z;
	double			z2;
	double			ex;
	double			dot;
	double			tolerance;		// convergence tolerance
	double			SensorNoise;	// noise variance of one sensor
	static double	a1 = 1.;
	static double	a2 = 1.;
	static double	HighPass = 0.;	// highpass frequency
	static double	LowPass = 0.;	// lowpass frequency
	register double	tmp;			// the famous temporary variable of hacker mythology
	register long double lsum;		// an attempt to promote the aforementioned variable to god-like fame
	int				*ChanIndex;		// ChanIndex[M] -- index of primary channels [used]
	int				*sample;		// sample[T] -- index of sample indices
	int				E;				// number of good epochs
	int				M;				// number of good channels
	int				T;				// samples per epoch per channel
	int				TT;				// total samples
	int				e;				// epoch index
	int				m;				// sensor index
	int				n;				// 'n' stands for 'nuther sensor index
	int				t;				// time index
	int				tt;				// dataset time index (multiple epochs)
	int				i;				// a very popular index
	int				j;				// an oft-times useful index
	int				k;				// one could not multiply matrices without this index
	int				c;				// a character
	int				its;			// iterations
	int				guess;			// number of bad guesses
	int				retry;			// number of retries
	int				fdin;			// input file descriptor
	int				fdout;			// output file descriptor
	int				dbytes;			// data bytes
	unsigned char	**Bad;			// Bad[E][T] -- flags for bad samples
	static int		digits = 10;	// default number of digits for convergence (10)
	static int		function = 1;	// default nonlinear function is kurtosis (1)
	static int		mem = 0;		// initializer memory
	static int		K = 0;			// rank
	static int		baseline = BASELINE;	// default trend removal
	static int		dflg = FALSE;	// delete backup file flag
	static int		rflg = FALSE;	// dataset name flag
	static int		vflg = FALSE;	// verbose mode flag
	long			Ltmp;			// temporary long integer variable
	extern char		*optarg;
	extern int		optind;
	extern int		opterr;
	char			DataPath[256];	// data set name full path
	char			fpath[256];		// general path name
	char			DSName[256];	// MEG dataset name
	char			OldName[256];	// old .meg4 file name
	char			NewName[256];	// new .meg4 file name
	char			FileID[9];		// file ID string
	char			CoeffID[] =		// SAM image file ID
		"SAMCOEFF";
	static char		MEG4Hdr[] =		// MEG 4.1 header string
		"MEG41CP";
	FILE			*fp;			// file pointer for log
	off_t			offset;			// number of bytes offset
	void			Cleanup(char *);

	// parse command line parameters
	opterr = 1;		// enable error reporting
	while((c = getopt(argc, argv, "r:f:g:p:b:dv")) != EOF) {
		switch(c) {

			case 'r':	// data set directory name (w/o '.ds')
				sprintf(DSName, "%s", optarg);
				rflg = TRUE;
				break;

			case 'f':	// output high & low pass filter frequencies
				if(sscanf(optarg, "%lf%lf", &HighPass, &LowPass) != 2)
					Cleanup("'-f' flag requires high and lowpass frequencies, in quotes");
				if(HighPass >= LowPass)
					Cleanup("highpass frequency must be less than lowpass frequency");
				break;

			case 'g':	// nonlinear function
				function = atoi(optarg);
				if(function < 1 || function > 3)
					Cleanup("nonlinear function must be in range 1 to 3");
				break;

			case 'p':	// precision
				digits = atoi(optarg);
				if(digits < 2 || digits > 10)
					Cleanup("number of significant digits for convergence must be in range 2 to 10");
				tolerance = pow((double)10., (double)(-digits));
				break;

			case 'b':	// baseline removal order
				baseline = atoi(optarg);
				if(baseline < 0 || baseline > 4)
					Cleanup("baseline removal order must be in range from 0 to 4");
				break;

			case 'd':	// delete backup meg4 file
				dflg = TRUE;
				break;

			case 'v':	// verbose operation
				vflg = TRUE;
				break;

			default:
				Cleanup("unknown option flag");
				break;
		}
	}
	if(rflg == FALSE) {
		fprintf(stderr, "usage: fastica\t-r <dataset name>\t%s%s, %s\n", REVISION, PRG_REV, __DATE__);
		fprintf(stderr, "\t\t-t <start, end (s) for sphering matrix>\n");
		fprintf(stderr, "\t\t-f <highpass, lowpass (Hz)>\n");
		fprintf(stderr, "\t\t-g <nonlinear function value>\n");
		fprintf(stderr, "\t\t\t1 -- kurtosis\n");
		fprintf(stderr, "\t\t\t2 -- hyperbolic tangent\n");
		fprintf(stderr, "\t\t\t3 -- gaussian\n");
		fprintf(stderr, "\t\t-p <number of digits for convergence tolerance>\n");
		fprintf(stderr, "\t\t-b <baseline removal order>\n");
		fprintf(stderr, "\t\t-d -- delete backup meg4 file\n");
		fprintf(stderr, "\t\t-v -- verbose mode\n");
		exit(-1);
	}

	// announce & get data with sensor structures
	if(vflg == TRUE) {
		printf("F A S T   I N D E P E N D E N T   C O M P O N E N T S   A N A L Y S I S\t%s%s\n", REVISION, PRG_REV);
		printf("\tusing ");
		switch(function) {
			case 1: printf("kurtosis"); break;
			case 2: printf("hyperbolic tangent"); break;
			case 3: printf("gaussian"); break;
			default: Cleanup("no such nonlinear function"); break;
		}
		printf(" nonlinearity\n");
		printf("\ttolerance for convergence will be %e (%d digits)\n", (double)tolerance, digits);
		printf("opening data file");
		fflush(stdout);
	}

	// get data with sensor structures
	GetDsInfo(DSName, &Header, &Channel, &Epoch, &Bad, TRUE);
	sprintf(DSpath, "%s/%s.ds", Header.DsPath, Header.SetName);
//	sprintf(SAMpath, "%s/SAM", DSpath);

	// count number of good channels & epochs
	for(m=M=0; m<Header.NumChannels; m++)
		if(Channel[m].Flag == TRUE && Channel[m].ChannelType == TYPE_MEG) {
			M++;
			Channel[m].Flag = TRUE;
		} else {
			Channel[m].Flag = FALSE;
		}
	for(e=E=0; e<Header.NumEpochs; e++)
		if(Epoch[e].Flag == TRUE)
			E++;
	T = Header.MaxSamples;
	TT = T * E;
	nt = (double)T;
	ne = (double)E;
	ntt = (double)TT;

	// test filter limits, &, if req'd, make filter
	if(LowPass != 0. || HighPass != 0.) {
		if(vflg == TRUE) {
			printf(" - done\n");
			printf("making %4.1f to %5.1f Hz filter", HighPass, LowPass);
			fflush(stdout);
		}
		if(HighPass < 0. || LowPass > Header.SampleRate/2.)
			Cleanup("filter bandpass out of range");
		if(LowPass > 0 && HighPass >= LowPass)
			Cleanup("illegal bandpass parameters");
		if(HighPass > 0.) {
			HPFilter.type = HIGHPASS;
			HPFilter.fh = Header.SampleRate / 2.;
			HPFilter.fl = HighPass;
			HPFilter.rate = Header.SampleRate;
			HPFilter.order = IIRORDER;
			HPFilter.enable = TRUE;
			if(mkiir(&HPFilter) == -1)
				Cleanup("'mkiir()' failed");
		} else
			HPFilter.enable = FALSE;
		if(LowPass > 0.) {
			LPFilter.type = LOWPASS;
			LPFilter.fh = LowPass;
			LPFilter.fl = 0.;
			LPFilter.rate = Header.SampleRate;
			LPFilter.order = IIRORDER;
			LPFilter.enable = TRUE;
			if(mkiir(&LPFilter) == -1)
				Cleanup("'mkiir()' failed");
		} else
			LPFilter.enable = FALSE;
		SensorNoise = NOISE * (LowPass - HighPass);
	} else {
		SensorNoise = NOISE * Header.SampleRate / 3.;
	}

	// open ica logfile
	sprintf(fpath, "%s-fastica.log", DSName);
	if((fp = fopen(fpath, "w")) == NULL)
		Cleanup("can't open logfile");
	fprintf(fp, "Fastica Log File for dataset: %s.ds\n", SetName);
	fprintf(fp, "Number of Channels: %d\n", M);
	fprintf(fp, "Number of Epochs: %d\n", E);
	fprintf(fp, "Number of Samples/Epoch: %d\n", T);
	fprintf(fp, "Sample Rate: %7.2f Hz\n", Header.SampleRate);
	if(LowPass > 0. || HighPass > 0.) {
		fprintf(fp, "Bandpass: %4.1f to %5.1f Hz\n", HighPass, LowPass);	
	} else {
		fprintf(fp, "(No Bandpass Filters)\n");
	}
	fprintf(fp, "Using ");
	switch(function) {
		case 1: fprintf(fp, "Kurtosis"); break;
		case 2: fprintf(fp, "Hyperbolic Tangent"); break;
		case 3: fprintf(fp, "Gaussian"); break;
		default: Cleanup("no such nonlinear function"); break;
	}
	fprintf(fp, " Nonlinearity\n");
	fprintf(fp, "Tolerance for Convergence: %e (%d digits)\n", (double)tolerance, digits);
	fflush(fp);

	// allocate memory for arrays
	if(vflg == TRUE) {
		printf(" - done\n");
		printf("allocating memory");
		fflush(stdout);
	}
	if((xint = (long *)malloc((size_t)T * sizeof(long))) == NULL)
		allocfailed("xint[]");
	if((x = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
		allocfailed("x[]");
	for(i=0; i<M; i++)
		if((x[i] = (double *)malloc((size_t)TT * sizeof(double))) == NULL)
			allocfailed("x[][]");
	if((y = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
		allocfailed("y[]");
	for(i=0; i<M; i++)
		if((y[i] = (double *)malloc((size_t)T * sizeof(double))) == NULL)
			allocfailed("y[][]");
	if((C = (long double **)malloc((size_t)M * sizeof(long double *))) == NULL)
		allocfailed("C[]");
	for(i=0; i<M; i++)
		if((C[i] = (long double *)malloc((size_t)M * sizeof(long double))) == NULL)
			allocfailed("C[][]");
	if((S = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
		allocfailed("S[]");
	for(i=0; i<M; i++)
		if((S[i] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
			allocfailed("S[][]");
	if((A = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
		allocfailed("A[]");
	for(i=0; i<M; i++)
		if((A[i] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
			allocfailed("A[][]");
	if((B = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
		allocfailed("B[]");
	for(i=0; i<M; i++)
		if((B[i] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
			allocfailed("B[][]");
	if((u = (double *)malloc((size_t)M * sizeof(double))) == NULL)
		allocfailed("u[]");
	if((v = (double *)malloc((size_t)M * sizeof(double))) == NULL)
		allocfailed("v[]");
	if((w = (double *)malloc((size_t)M * sizeof(double))) == NULL)
		allocfailed("w[]");
	if((L = (long double *)malloc((size_t)M * sizeof(long double))) == NULL)
		allocfailed("L[]");
	if((hypTan = (double *)malloc((size_t)TT * sizeof(double))) == NULL)
		allocfailed("hypTan[]");
	if((In = (double *)malloc((size_t)T * sizeof(double))) == NULL)
		allocfailed("In[]");
	if((Out = (double *)malloc((size_t)T * sizeof(double))) == NULL)
		allocfailed("Out[]");
	if((ChanIndex = (int *)malloc((size_t)M * sizeof(int))) == NULL)
		allocfailed("ChanIndex[]");
	if((sample = (int *)malloc((size_t)T * sizeof(int))) == NULL)
		allocfailed("sample[]");

	// make index of all good primary channels
	for(m=i=0; m<Header.NumChannels; m++)
		if(Channel[m].Flag == TRUE)
			ChanIndex[i++] = m;

	// open input & output '.meg4' files
	if(vflg == TRUE) {
		printf(" - done\n");
		printf("opening data files");
		fflush(stdout);
	}
	sprintf(fpath, "%s/%s.meg4", DSPath, DSName);
	if((fdin = open(fpath, O_RDONLY)) == -1)
		Cleanup("can't open '.meg4' file for read");
	if(read(fdin, FileID, 8) != 8)
		Cleanup("can't read header from '.meg4' file");
	if(strncmp(FileID, "MEG4", 4))
		Cleanup("unknown header ID in '.meg4' file");
	dbytes = T * sizeof(long);
	sprintf(fpath, "%s/%s.meg4.NEW", DSPath, DSName);
	if((fdout = open(fpath, O_WRONLY | O_CREAT | O_TRUNC, 0666)) == -1)
		Cleanup("can't open '.meg4' file for write");
	if(write(fdout, MEG4Hdr, 8) != 8)
		Cleanup("can't write header to '.meg4' file");

	// for each epoch
	if(vflg == TRUE) {
		printf(" - done\n");
		printf("reading data & computing covariance");
		fflush(stdout);
	}
	for(i=0; i<M; i++)		// zero covariance matrix
		for(j=0; j<M; j++)
			C[i][j] = 0.;
	for(e=0; e<E; e++) {

		// get data
		for(i=0; i<M; i++) {
			m = ChanIndex[i];
			offset = (off_t)dbytes * (off_t)(m + e * Header.NumChannels) + (off_t)8;
			if(lseek(fdin, offset, SEEK_SET) != offset)
				Cleanup("'lseek()' to channel failed");
			if(read(fdin, (char *)xint, dbytes) != dbytes)
				Cleanup("can't read data from 'meg4' file");
			for(t=0; t<T; t++) {
				Ltmp = bswap_32(xint[t]);
				In[t] = (double)Ltmp * Channel[m].Scale;
			}

			// apply lowpass & highpass filter sections, & remove mean
			if(bdiir(In, Out, T, &LPFilter) == -1)
				Cleanup("'bdiir()' failed");
			if(bdiir(Out, y[i], T, &HPFilter) == -1)
				Cleanup("'bdiir()' failed");
			if(debase(y[i], T, baseline) == -1)
				Cleanup("'debase()' failed");
		}

		// compute active & control-state covariance matrices
		for(i=0; i<M; i++)	// compute upper triangle
			for(j=i; j<M; j++) {
				for(t=0, lsum=0.; t<T; t++)
					lsum += y[i][t] * y[j][t];
				C[i][j] += lsum / nt;
			}

		// heartbeat
		if(vflg == TRUE) {
			printf(".");
			fflush(stdout);
		}
	}

	// compute total covariance matrix
	for(i=0; i<M; i++)		// normalize upper triangle by number of epochs
		for(j=i; j<M; j++)
			C[i][j] /= ne;
	for(i=1; i<M; i++)	// copy to lower triangle
		for(j=0; j<i; j++)
			C[i][j] = C[j][i];

	// prewhiten the measurement data by inverse square-root of covariance
	if(vflg == TRUE) {
		printf(" - done\n");
		printf("computing eigendecomposition of covariance");
		fflush(stdout);
	}
	if(eigen(C, L, M) == -1)	// note that the eigenvectors placed in C are column vectors!
		Cleanup("'eigen()' failed");

	// determine rank of difference covariance (note the reverse order of eigenvalues
	tmp = fabs((double)L[M-1]);	// copy the value of the most-significant eigenvalue
	for(i=(M-1), K=1; i>=0; i--, K++)
		if(fabs(tmp / (double)L[i]) > 1.0e+06 || fabs((double)L[i]) < SensorNoise)
			break;
	for(i=0; i<M; i++) {
		n = M - i - 1;		// remember, the eigenvectors are in reverse order
		tmp = sqrt(fabs((double)L[i]));
		for(j=0; j<M; j++)
			S[n][j] = (double)C[j][i] / tmp;
	}

	// log sphering matrix
	if(vflg == TRUE) {
		printf(" - done\n");
		printf("solution rank: %d\n", K);
		fflush(stdout);
	}
	fprintf(fp, "Rank of Solution: %d\n", K);
	fprintf(fp, "Sphering Matrix:\n");
	for(i=0; i<K; i++)
		for(j=0; j<M; j++) {
			fprintf(fp, "S[%d][%d]=%e\n", i, j, S[i][j]);
			fflush(fp);
		}

	// project data onto inverse square root covariance -- x = S data
	if(vflg == TRUE) {
		printf("sphering data");
		fflush(stdout);
	}
	for(e=0; e<E; e++) {

		// re-read data
		for(i=0; i<M; i++) {
			m = ChanIndex[i];
			offset = (off_t)dbytes * (off_t)(m + e * Header.NumChannels) + (off_t)8;
			if(lseek(fdin, offset, SEEK_SET) != offset)
				Cleanup("'lseek()' to channel failed");
			if(read(fdin, (char *)xint, dbytes) != dbytes)
				Cleanup("can't read data from 'meg4' file");
			for(t=0; t<T; t++) {
				Ltmp = bswap_32(xint[t]);
				In[t] = (double)Ltmp * Channel[m].Scale;
			}

			// apply lowpass & highpass filter sections, & remove mean
			bdiir(In, Out, T, &LPFilter);
			bdiir(Out, y[i], T, &HPFilter);
			if(debase(y[i], T, baseline) == -1)
				Cleanup("'debase()' failed");
		}

		// sphering data
		for(i=0; i<K; i++)		// 'i' is eigenvector index
			for(t=0; t<T; t++) {
				tt = e * T + t;
				for(j=0, tmp=0.; j<M; j++)
					tmp += S[i][j] * y[j][t];
				x[i][tt] = tmp;
			}

		// heartbeat
		if(vflg == TRUE) {
			printf(".");
			fflush(stdout);
		}
	}

	// on first call, initialize seed using system time
	if(mem == 0) {
		srand48((long)time((time_t *)0));
		mem++;
	}

	// remove trend from sphered data
	for(i=0; i<K; i++)
		if(debase(x[i], TT, baseline) == -1)
			Cleanup("'debase()' failed");

	// initialize mixing matrix
	for(i=0; i<K; i++)
		for(j=0; j<K; j++)
			B[i][j] = 0.;

	// recursively find each component
	if(vflg == TRUE) {
		printf(" - done\n");
		printf("finding independent components:\n");
		fflush(stdout);
	}
	for(m=0; m<K; m++) {	// assume finding K source components
		if(vflg == TRUE) {
			printf("component %d:\n", m+1);
			fflush(stdout);
		}
		fprintf(fp, "\nIndependent Component %d:\n", m+1);
		fflush(fp);
		retry = 0;
restart:
		// compute random weight vector
		for(i=0; i<K; i++)
			w[i] = 2. * drand48() - 1.;

		// compute A = I - B B', where B is the current mixing matrix
		for(i=0; i<K; i++)
			for(j=0; j<K; j++) {
				for(k=0, tmp=0.; k<K; k++)
					tmp += B[i][k] * B[j][k];
				if(i == j)
					A[i][j] = 1. - tmp;
				else
					A[i][j] = 0. - tmp;
			}
		
		// remove projections of previous independent components
		for(i=0, r2=0.; i<K; i++) {
			for(j=0, tmp=0.; j<K; j++)
				tmp += w[j] * A[j][i];
			u[i] = tmp;
			r2 += tmp * tmp;
		}

		// make w a unit vector
		r = sqrt(r2);
		for(i=0; i<K; i++)
			w[i] = u[i] / r;

		// log seed weights
		fprintf(fp, "\tTry %d, Seed Weight Vector:\n", retry);
		for(i=0; i<K; i++) {
			fprintf(fp, "W[%d]=%e\n", i, w[i]);
			fflush(fp);
		}

		// iterate weight vector
		rold = 0.;
		its = guess = 0;
		do {

			// select nonlinearity
			switch(function) {

				case 1:		// compute kurtosis nonlinearity
					for(i=0; i<K; i++) {
						for(t=0, u[i]=0.; t<TT; t++) {
							for(j=0, tmp=0.; j<K; j++)
								tmp += w[j] * x[j][t];
							u[i] += tmp * tmp * tmp * x[i][t];
						}
						u[i] = u[i] / ntt - 3. * w[i];
					}
					break;

				case 2:		// compute hyperbolic tangent nonlinearity
					for(t=0, sum=0.; t<TT; t++) {
						for(i=0, tmp=0.; i<K; i++)
							tmp += x[i][t] * w[i];
						hypTan[t] = tanh(a1 * tmp);
						sum += 1. - hypTan[t] * hypTan[t];
					}

					// compute w = (x * hypTan - a1 * sum(1 - hypTan^2)' * w) / ntt
					for(i=0; i<K; i++) {
						for(t=0, tmp=0.; t<TT; t++)
							tmp += x[i][t] * hypTan[t];
						dot = tmp;
						u[i] = (dot - a1 * sum * w[i]) / ntt;
					}
					break;

				case 3:		// compute Gaussian nonlinearity
					for(i=0; i<K; i++)
						u[i] = 0.;
					for(t=0, sum=0.; t<TT; t++) {
						for(i=0, tmp=0.; i<K; i++)
							tmp += x[i][t] * w[i];
						z = tmp;
						z2 = tmp * tmp;
						ex = exp(-a2 * z2 / 2.);
						Gauss = z * ex;			// Gauss[t]
						sum += (1. - a2 * z2) * ex;	// dGauss[t]
						for(i=0; i<K; i++)
							u[i] += x[i][t] * Gauss;
					}
					for(i=0; i<K; i++) {
						u[i] -= sum * w[i];
						u[i] /= ntt;
					}
					break;

				default:
					Cleanup("no such nonlinear function");
					break;
			}

			// remove projections of previous independent components
			for(i=0, r2=0.; i<K; i++) {
				for(j=0, tmp=0.; j<K; j++)
					tmp += u[j] * A[j][i];
				v[i] = tmp;
				r2 += tmp * tmp;
			}

			// normalize u, compute dot product with w, & update w
			r = sqrt(r2);
			for(i=0, tmp=0.; i<K; i++) {
				u[i] = v[i] / r;
				tmp += u[i] * w[i];
				w[i] = u[i];
			}

			r = sqrt(fabs(tmp));
			its++;
			if(vflg == TRUE) {
				printf("\tpass %d: norm=%13.10f\n", its, r);
				fflush(stdout);
			}

			// check for bad guess
			if(r <= rold)
				guess++;
			else
				guess = 0;
			if(guess > MAXGUESS || (its > 50 && r < 0.99) || (its > 100 && r < 0.9999) || (its > 200 && r < 0.999999)) {
				if(retry > MAXRETRIES) {
					K = m - 1;
					printf("\t\t(terminating search at %d components)\n", K);
					fflush(stdout);
					fprintf(fp, "search terminated at %d components\n", K);
					fflush(fp);
					break;
				}
				if(vflg == TRUE) {
					printf("\t\t(restart component %d after %d retries)\n", m+1, retry);
					fflush(stdout);
				}
				retry++;
				goto restart;
			} else
				rold = r;

		} while((1. - r) > tolerance && its < MAXITS);

		// update mixing matrix B
		for(i=0; i<K; i++)
			B[i][m] = w[i];

		// log final weights
		fprintf(fp, "\tComponent %d ICA Weight Vector:\n", m+1);
		for(i=0; i<K; i++) {
			fprintf(fp, "W[%d]=%e\n", i, w[i]);
			fflush(fp);
		}
	}	// end K components loop

	// project sphered data through rotation matrix
	if(vflg == TRUE) {
		printf(" - done\n");
		printf("projecting sphered data through unmixing matrix");
		fflush(stdout);
	}
	for(e=0; e<E; e++) {
		for(i=0; i<K; i++) {
			for(t=0; t<T; t++) {
				tt = e * T + t;
				for(j=0, tmp=0.; j<K; j++)
					tmp += B[j][i] * x[j][tt];
				y[i][t] = tmp;
			}
			for(t=0; t<T; t++)
				y[i][t] *= SCALE;
		}

		// write one epoch of data
		for(m=i=0; m<Header.NumChannels; m++) {
			if(m == ChanIndex[i]) {
				if(i < K) {
					for(t=0; t<T; t++)
						xint[t] = (long)(y[i][t]);
					if(write(fdout, (char *)xint, dbytes) != dbytes)
						Cleanup("can't write data to '.meg4' file");
					i++;
				} else {
					offset = (off_t)dbytes * (off_t)(m + e * Header.NumChannels) + (off_t)8;
					if(lseek(fdin, offset, SEEK_SET) != offset)
						Cleanup("'lseek()' to channel failed");
					if(read(fdin, (char *)xint, dbytes) != dbytes)
						Cleanup("can't read data from 'meg4' file");
					if(write(fdout, (char *)xint, dbytes) != dbytes)
						Cleanup("can't write data to '.meg4' file");
				}

			} else {		// channel isn't in primary channels list -- just copy it...
				offset = (off_t)dbytes * (off_t)(m + e * Header.NumChannels) + (off_t)8;
				if(lseek(fdin, offset, SEEK_SET) != offset)
					Cleanup("'lseek()' to channel failed");
				if(read(fdin, (char *)xint, dbytes) != dbytes)
					Cleanup("can't read data from 'meg4' file");
				if(write(fdout, (char *)xint, dbytes) != dbytes)
					Cleanup("can't write data to '.meg4' file");
			}
		}

		// hearbeat
		if(vflg == TRUE) {
			printf(".");
			fflush(stdout);
		}
	}
	close(fdin);
	close(fdout);
	fclose(fp);

	// rename original '.meg4' file
	if(vflg == TRUE) {
		printf(" - done\n");
		printf("moving & renaming '.meg4' file");
		fflush(stdout);
	}
	sprintf(OldName, "%s/%s.ds/%s.meg4", getenv("MEG4_DIR"), SetName, SetName);
	sprintf(NewName, "%s/%s.ds/%s.meg4.bak", getenv("MEG4_DIR"), SetName, SetName);
	if(rename(OldName, NewName) == -1)
		Cleanup("backup of original '.meg4' file failed");
	sprintf(OldName, "%s/%s.ds/%s.meg4.NEW", getenv("MEG4_DIR"), SetName, SetName);
	sprintf(NewName, "%s/%s.ds/%s.meg4", getenv("MEG4_DIR"), SetName, SetName);
	if(rename(OldName, NewName) == -1)
		Cleanup("rename of filtered '.meg4' file failed");
	if(dflg == TRUE) {
		if(vflg == TRUE) {
			printf(" - done\n");
			printf("deleting backup (original) '.meg4' file");
			fflush(stdout);
		}
		sprintf(OldName, "%s/%s.ds/%s.meg4.bak", getenv("MEG4_DIR"), SetName, SetName);
		if(unlink(OldName) == -1)
			Cleanup("can't delete backup data file");
	}

	if(vflg == TRUE) {
		printf(" - done\n");
		printf("fastica: done\n");
		fflush(stdout);
	}
	exit(0);
}


void	Cleanup(
char	*msg		// error message string
)

{
	fprintf(stderr, "\n\07fastica: %s\n", msg);
	exit(-1);
}
