// NIFTIPeak -- find the peaks in all NIFTI volumetric image files
//  in the SAM subdirectory of the named dataset
//
//  flags:
//  -r <run name>   -- MEG run name
//  -i <image name> -- MEG SAM volume file name
//  -t <threshold>  -- threshold value (usually z-deviate)
//  -s <size>       -- radius of maximum search
//  -n <max>        -- maximum number of maxima
//  -p              -- list power
//  -v              -- verbose mode
//
//  Author: Stephen E. Robinson
//      MEG Core Facility
//      NIMH
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <dirent.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <byteswap.h>
#include <model.h>
#include <samlib.h>
#include <DataFiles.h>
#include <SAMfiles.h>
#include <nifti1.h>
#include <version.h>
#include "samutil.h"
#include "sam_parse.h"

#define IVI     3                   // image value index
#define MINOR_REV   1

int main(
    int     argc,
    char    **argv
) {
    HeaderInfo      Header;         // MEG data header
    ChannelInfo     *Channel;       // MEG channel info
    EpochInfo       *Epoch;         // MEG epoch info
    PARMINFO        Params;         // analysis parameters
    PARM            *p;             // pointer into active parameter list
    nifti_1_header  NiiHdr;         // .nii header
    struct dirent   *dp;            // directory entry structure
    DIR             *dirp;          // directory pointer
    float           **Image;        // Image[D][V] -- image voxels
    float           *Voxel[3];      // Voxel[3][V] -- SAM image voxel coordinates
    float           *Max[4];        // Max[4][M] -- array of maxima
    float           *SortMax[4];    // SortMax[4][M] -- sorted array of maxima
    float           Vc[3];          // cartesian position
    float           Vs[3];          // starting coordinates
    float           Ve[3];          // ending coordinates
    float           Vstep[3];       // step size for each axis
    float           rms;            // rms non-zero voxel values
    float           sum;            // sum of non-zero voxel values^2
    float           Radius;         // search radius for maximum
    float           r2;             // test radius^2
    float           R2;             // maximum radius^2
    float           max;            // maximum voxel value
    float           voxel;          // voxel value;
    float           tmp;            // this is the most famous temporary variable
    float           ImgThresh;      // threshold for current image
    static float    size = 0.;      // size of search region
    static float    threshold = 0.; // image threshold
    int             V;              // number of maxima voxels
    int             c;              // input character & virtual channel
    int             cc;             // secondary voxel index
    int             v;              // a vector index
    int             i;              // matrix index
    int             j;              // matrix index
    int             k;              // and another
    int             D;              // number of 3d images
    int             I;              // number of i-axis voxels
    int             J;              // number of j-axis voxels
    int             K;              // number of k-axis voxels
    int             M;              // number of maxima
    int             S;              // number of sensors
    int             eflg;           // command-line error flag
    int             mflg;           // maxima flag
    static int      rflg = FALSE;   // dataset name
    static int      pflg = FALSE;   // power list flag
    static int      vflg = FALSE;   // verbose mode flag
    static int      N = 10;         // maximum number of maxima
    char            fpath[256];     // general path name
    char            *DSName = NULL; // MEG run name
    char            DSpath[256];    // dataset path
    char            SAMpath[256];   // SAM subdirectory pathname
    char            NiiPath[256];   // path for finding NIFTI files
    char            ImgPath[256];   // output path for writing maxima
    char            NiiName[256];   // SAM volume image name
    unsigned char   **Bad;          // Bad[E][T] -- flags for bad samples
    FILE            *fp;            // maximum list file

    Progname = argv[0];
    reg_usage(MINOR_REV, __DATE__);

    // Register the parameters used by this program.

    reg_std_parm();

    reg_parm("ImageDirectory");
    reg_parm("MRIDirectory");
    reg_parm("MRIPattern");
    reg_parm("PrefixLength");

    reg_parm("Threshold");
    reg_parm("SearchRadius");
    reg_parm("NumMaxima");
    reg_parm("ShowPower");

    // Parse the arguments, environment variables, and parameter files.

    do_parse_args(argc, argv);

    // Extract specified parameters.

    eflg = get_params(&Params);

    p = get_parm("DataSet");
    if (p->set) {
        rflg = TRUE;
        DSName = Params.DataSetName;
    }
#if BTI
    p = get_parm("PDFName");
    if (p->set) {
        dflg = TRUE;
        PDFName = (char *)p->ptr;
    }
#endif
    p = get_parm("verbose");
    vflg = p->set;
    p = get_parm("param");
    mflg = p->set;

#if BTI
    if(!rflg || !dflg || !mflg || eflg) {
#else
    if(!rflg || !mflg || eflg) {
#endif
        msg("dataset (-r) and parameter file (-m) are required\n");
        do_help();  // doesn't return
    }

    p = get_parm("Threshold");
    if (p->set) {
        threshold = (float)(*(double *)p->ptr);
    }
    p = get_parm("SearchRadius");
    if (p->set) {
        size = (float)(*(double *)p->ptr);
        size *= 0.01;
        if (size <= 0.)
            cleanup("search radius must be positive");
    }
    p = get_parm("NumMaxima");
    if (p->set) {
        N = *(int *)p->ptr;
    }
    p = get_parm("ShowPower");
    pflg = p->set;

    // announce
    if(vflg == TRUE) {
        printf("N I F T I   V O L U M E   F I L E   I M A G E   P E A K   F I N D E R\t%s\n", PRG_REV);
        fflush(stdout);
    }

    // get data with sensor structures
    GetDsInfo(DSName, &Header, &Channel, &Epoch, &Bad, TRUE);
    sprintf(DSpath, "%s/%s.ds", Header.DsPath, Header.SetName);
    sprintf(SAMpath, "%s/SAM", DSpath);

    // NIFTI files are found either in the dataset SAM subdirectory or in the image directory
    if(strncmp(Params.DirName, "NULL", 4)) {
        sprintf(NiiPath, "%s", Params.DirName);
    } else {
        sprintf(NiiPath, "%s", SAMpath);
    }

    // maxima files are written to either the dataset SAM subdirectory or to the MRI subdirectory
    if(strncmp(Params.MRIdirectory, "NULL", 4)) {
        GetMRIPath(ImgPath, sizeof(ImgPath), &Params, "", FALSE);
    } else {
        strcpy(ImgPath, SAMpath);
    }

    // loop through directory entries
    if(vflg == TRUE) {
        printf(" - done\n");
        printf(" reading NIFTI file(s):\n");
        fflush(stdout);
    }
    dirp = opendir(NiiPath);
    if (dirp == NULL) {
        cleanup("can't open input directory '%s'", NiiPath);
    }
    while((dp = readdir(dirp)) != NULL) {

        // determine if this is a NIFTI file
        strcpy(NiiName, dp->d_name);            // copy the full file name
        i = strlen(NiiName);
        if(!strcmp(&NiiName[i-4], ".nii")) {    // Yes! This is a NIFTI file!
            NiiName[i-4] = '\0';                // strip off .nii

            // open NIFTI file
            sprintf(fpath, "%s/%s.nii", NiiPath, NiiName);
            if(vflg == TRUE) {
                printf("\nopening NIFTI image file: '%s.nii'", NiiName);
                fflush(stdout);
            }
            GetNIFTI(fpath, &D, &Image, &NiiHdr);
            if(D != 1)
                Cleanup("3d+time images not supported");

            // set up arrays for voxel coordinates & values (convert from mm)
            S = Header.NumPri;
            Vs[X_] = 0.001 * NiiHdr.srow_y[3];
            Vs[Y_] = -0.001 * NiiHdr.srow_x[3];
            Vs[Z_] = 0.001 * NiiHdr.srow_z[3];
            Vstep[X_] = 0.001 * NiiHdr.srow_y[2];   // i-axis SAM step size
            Vstep[Y_] = -0.001 * NiiHdr.srow_x[1];  // j-axis "
            Vstep[Z_] = 0.001 * NiiHdr.srow_z[0];   // k-axis "
            for(i=0; i<4; i++) {
                if((Max[i] = (float *)malloc((size_t)(S * sizeof(float)))) == NULL)
                    allocfailed("Max[][]'");
                if((SortMax[i] = (float *)malloc((size_t)(S * sizeof(float)))) == NULL)
                    allocfailed("SortMax[][]");
            }
            I = NiiHdr.dim[3];              // slowest changing index (x)
            J = NiiHdr.dim[2];              // next (y)
            K = NiiHdr.dim[1];              // most rapidily changing index (z)
            V = I * J * K;
// printf("\n\nI=%d, J=%d, K=%d, Vx=%f, Vy=%f, Vz=%f\n", I, J, K, 100.*Vs[X_], 100.*Vs[Y_], 100.*Vs[Z_]);
// printf("stepX=%f, stepY=%f, stepZ=%f\n", 100.*Vstep[X_], 100.*Vstep[Y_], 100.*Vstep[Z_]);
            for(v=X_; v<=Z_; v++)
                if((Voxel[v] = (float *)malloc((size_t)V * sizeof(float))) == NULL)
                    allocfailed("Voxel[][]");

            Ve[0] = Ve[1] = Ve[2] = 0.; // shut the compiler up when using -Wmaybe-uninitialized
            k = 0;
            for(i=c=0, Vc[X_]=Vs[X_]; i<I; i++, Vc[X_]+=Vstep[X_]) {
                Ve[X_] = Vc[X_];
                for(j=0, Vc[Y_]=Vs[Y_]; j<J; j++, Vc[Y_]+=Vstep[Y_]) {
                    Ve[Y_] = Vc[Y_];
                    for(k=0, Vc[Z_]=Vs[Z_]; k<K; k++, Vc[Z_]+=Vstep[Z_], c++) {
                        Ve[Z_] = Vc[Z_];
                        for(v=X_; v<=Z_; v++)
                            Voxel[v][c] = Vc[v];
                    }
                }
            }

            // determine rms non-zero voxel absolute value
            if(vflg == TRUE) {
                printf(" - done\n");
                printf("finding rms value of all non-zero voxels");
                fflush(stdout);
            }
            for(c=cc=0, sum=0.; c<V; c++) {
                voxel = fabsf(Image[0][c]);
                if(voxel > 0. && voxel < 1.0e+12) {
                    sum += voxel * voxel;
                    cc++;
                }
            }
            if(cc > 1) {
                rms = sqrtf(sum / (float)cc);
                if(threshold == 0.)
                    ImgThresh = rms;
                else
                    ImgThresh = threshold;
            } else {
                ImgThresh = rms = sqrtf(sum / (float)V);
            }

            // search image for absolute maxima above threshold
            if(vflg == TRUE) {
                printf(" - done\n");
                printf("rms = %7.3f rms\n", rms);
                printf("searching for maxima using threshold = %7.3f", ImgThresh);
                fflush(stdout);
            }
            if(size == 0.)
                Radius = 2.000001 * Vstep[X_];  // all steps are identical
            else
                Radius = size;
            R2 = Radius * Radius;
            for(c=M=0; c<V; c++) {
                voxel = fabsf(Image[0][c]);
                if(voxel > ImgThresh) {
                    for(cc=0, mflg=TRUE; cc<V; cc++)
                        if(cc != c) {
                            for(v=X_, r2=0.; v<=Z_; v++) {
                                tmp = Voxel[v][c] - Voxel[v][cc];
                                r2 += tmp * tmp;
                            }
                            if(r2 <= R2 && fabsf(Image[0][cc]) > voxel)
                                mflg = FALSE;
                        }
                    if(mflg == TRUE) {

                        // check to see if the maximum is on the boundary
                        if(Voxel[X_][c] > Vs[X_] && Voxel[X_][c] < Ve[X_]
                        && Voxel[Y_][c] > Vs[Y_] && Voxel[Y_][c] < Ve[Y_]
                        && Voxel[Z_][c] > Vs[Z_] && Voxel[Z_][c] < Ve[Z_]) {
                            for(v=X_; v<=Z_; v++)
                                Max[v][M] = Voxel[v][c];
                            Max[IVI][M] = voxel;
                            if(M < S)
                                M++;
                            else
                                cleanup("more maxima than sensor channels -- increase threshold");
                            if(vflg == TRUE) {
                                printf(".");
                                fflush(stdout);
                            }
                        }
                    }
                }
            }

            // sort the maxima in descending order of voxel magnitude
            if(M > 0) {
                if(vflg == TRUE) {
                    printf(" - done\n");
                    printf("%d local maxima found above threshold\n", M);
                    fflush(stdout);
                }

                // open file to list maxima
                sprintf(fpath, "%s/%s.max", ImgPath, NiiName);
                if((fp = fopen(fpath, "w")) == NULL)
                    cleanup("'%s' open for write failed", fpath);

                // sort maxima
                for(i=0; i<M; i++) {
                    for(j=0, max=0.; j<M; j++)
                        if(Max[IVI][j] > max) {
                            max = Max[IVI][j];
                            k = j;
                        }
                    for(v=X_; v<=Z_; v++)
                        SortMax[v][i] = Max[v][k];
                    SortMax[IVI][i] = Max[IVI][k];
                    Max[IVI][k] = 0.;
                }
                if(M < N)
                    N = M;
                if(vflg == TRUE) {
                    printf("saving %d out of %d maxima", N, M);
                    fflush(stdout);
                }
                for(i=0; i<N; i++) {
                    fprintf(fp, "%7.3f\t%7.3f\t%7.3f", 100.*SortMax[X_][i], 100.*SortMax[Y_][i], 100.*SortMax[Z_][i]);
                    if(pflg == FALSE)
                        fprintf(fp, "\n");
                    else
                        fprintf(fp, "\t%7.3f\n", SortMax[IVI][i]);
                    fflush(fp);
                }
                fclose(fp);
            } else {
                if(vflg == TRUE) {
                    printf(" - done\n");
                    printf("no maxima found above threshold\n");
                    fflush(stdout);
                }
            }

            // free allocated space
            for(i=0; i<4; i++) {
                free((void *)Max[i]);
                free((void *)SortMax[i]);
            }
        }
    }

    if(vflg) {
        printf(" - done\n'%s' done\n", Progname);
    }
    exit(0);
}
