// sam_4d -- synthetic aperture magnetometry imaging for source event related
//  time-locked signals. sam_4d reads all SAM weights for volume image, & computes
//  the source waveform for each voxel. Given specifications for active & control
//  time segments, SAM4d computes the mean power & variance time-series relative
//  to the specified event windows. The result is represented stored in a NIFTI image file
//  '.nii' file.
//
//  flags:
//      -r <dataset name>       -- MEG dataset name
//      -d <data file name>     -- MEG data file name (4D, only)
//      -m <parameter name>     -- name of parameter file
//      -v                      -- verbose mode
//
//  Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

// include...
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <geoms.h>
#include <samlib.h>
#include <siglib.h>
#include <rvelib.h>
#include <DataFiles.h>
#include <SAMfiles.h>
#include <filters.h>
#include <TimeSegs.h>
#include <nifti1.h>
#include <version.h>
#include "samutil.h"
#include "sam_parse.h"

// internal constants
#define MINOR_REV   5
#define SIG_ORDER   4

typedef struct {
    int     Epoch;              // epoch/trial
    int     TS;                 // starting sample
    double  *sample;            // segment source moment
} SAM_SEG;

typedef struct {
    char    Name[64];           // marker file name
    double  **Wgt;              // Wgt[V][M] -- SAM weights
    double  Noise;              // noise power
    SAM_HDR WgtHdr;             // SAM weight file header
    SAM_SEG *Segment;           // Segment[NumSegments] -- time-segment specs
    int     NumSegments;        // number of segments for this named marker
} IMAGE_STAT;

// INCOMPLETE: Need to work on DataSegments, integration, & time steps

int main(
    int     argc,
    char    **argv
) {
    IIRSPEC         DataIIR;        // IIR filter structure
    FFTSPEC         DataFFT;        // FFT data filter structure
    FFTSPEC         NotchFFT;       // notch FFT filter structure
    IMAGE_STAT      *Stats;         // Stats[N] -- structure for markers
    PARMINFO        Params;         // analysis parameters
    PARM            *p;             // pointer into active parameter list
    SAM_MARKS       *Marker;        // Marker[N] -- array of time-ordered markers
    HeaderInfo      Header;         // MEG data header
    ChannelInfo     *Channel;       // MEG channel info
    EpochInfo       *Epoch;         // MEG epoch info
    COV_HDR         CovHdr;         // noise covariance header
    nifti_1_header  NiiHdr;         // NIFTI header
    unsigned char   *ExtHdr;        // NiiHdr[N] -- extended NIFTI header
    double          ***x;           // x[E][T][M] -- filtered multisensor data
    double          **Cn;           // Cn[M][M] -- noise covariance matrix
    double          *In;            // In[T] -- input time-series
    double          *Out;           // Out[T] -- output time-series
    double          *Avg;           // Avg[TSEG] -- average
    double          *Var;           // Var[TSEG] -- variance
    double          time;           // time (sec)
    double          mark;           // time mark
    double          offset;         // offset from segment start to average window start
    double          BWFreq;         // effective bandwidth
    double          BWratio;        // ratio of covariance bandwidth to image bandwidth (nominal, only)
    double          Noise;          // effective noise
    double          percent;        // percent done
    double          a1;             // sum of metric
    double          a2;             // sum of metric squared
    double          mean;           // mean baseline
    float           **ImgM;         // ImgM[D][V] -- image mean
    float           **ImgV;         // ImgV[D][V] -- image variance
    double          tmp;            // what else?
    double          tmp2;
    double          tmp3;
    int             *ChanIndex;     // ChanIndex[M] -- index of channels used
    int             E;              // number of good epochs
    int             M;              // number of MEG or reference channels used
    int             N;              // just a number to take care of extended header read
    int             T;              // number of samples
    int             TS;             // data segment start
    int             TE;             // data segment end
    int             BS;             // baseline start sample
    int             BE;             // baseline end sample
    int             TB = 0;         // number of baseline samples
    int             TSEG;           // number of samples in data segment
    int             V;              // number of voxels weight file
    int             c;              // input character & virtual channel
    int             pct;            // integer percentage
    int             m;              // channel index
    int             d;              // image step index
    int             e;              // epoch index
    int             f;              // 1st primary channel index
    int             t;              // sample index
    int             tt;             // alternate sample index
    int             i;              // matrix index
    int             j;              // ...and another
    int             k;              // this is the last one -- for sure...
    int             n;              // condition index
    int             NumMarkers;     // number of markers in dataset -- e.g., 'MarkerFile.mrk'
    int             NumChans;       // number of channels in weight file
    int             NumTot;         // total number of named markers
    int             NumImg;         // number of SAM images
    int             NumBox;         // number of boxcar integrator samples
    TIMERANGE       *outrange;      // output only times in this range
    int             eflg = FALSE;   // command-line error flag
    int             bflg = FALSE;   // baseline removal flag
    int             mflg = FALSE;   // marker parameter file name flag
    int             nflg = FALSE;   // use noise covariance file
    int             rflg = FALSE;   // run name flag
    int             vflg = FALSE;   // verbose mode flag
    char            fpath[256];     // general path name
    char            *DSName = NULL; // MEG dataset name
    char            DSpath[256];    // MEG dataset path
    char            SAMpath[256];   // SAM subdirectory path
    char            WgtDir[256];    // weight directory path name
    char            *ParmName = NULL; // parameter file name
    char            Prefix[64];     // incomplete pathname
    char            Name1[256];     // partial image name
    char            Name2[256];     // partial image name
    char            FilterName[8];  // filter name -- IIR | FFT
    char            extension[4];   // extension
    FILE            *fp;            // data output file pointer

    // system-dependent variables
#if BTI
    static int      dflg;           // pdf file name flag
    char            *PDFName;       // name of data file
#else
    double          scale;          // bits-->Tesla conversion
    unsigned char   **Bad;          // Bad[E][T] -- flags for bad samples
#endif

    Progname = argv[0];
    reg_usage(MINOR_REV, __DATE__);

    // Register the parameters used by this program.

    reg_std_parm();

    reg_parm("NumMarkers");
    reg_parm("Marker");

    reg_parm("CovType");
    reg_parm("CovBand");
    reg_parm("ImageBand");
    reg_parm("NoiseBand");
    reg_parm("SmoothBand");
    reg_parm("FilterType");

    reg_parm("DataSegment");
    reg_parm("Baseline");
    reg_parm("TimeStep");
    reg_parm("BoxWidth");

    reg_parm("ImageMetric");
    set_parm_arg_help("ImageMetric", "Signal|Power|RankVectorEntropy");
    reg_parm("ImageDirectory");
    reg_parm("PrefixLength");
    reg_parm("ImageFormat");
    reg_parm("OutName");
    reg_parm("OutRange");

    reg_parm("Notch");
    reg_parm("Hz");

    // Parse the arguments, environment variables, and parameter files.

    do_parse_args(argc, argv);

    // Extract specified parameters.

    eflg = get_params(&Params);

    p = get_parm("DataSet");
    if (p->set) {
        rflg = TRUE;
        DSName = Params.DataSetName;
    }
#if BTI
    p = get_parm("PDFName");
    if (p->set) {
        dflg = TRUE;
        PDFName = (char *)p->ptr;
    }
#endif
    p = get_parm("verbose");
    vflg = p->set;
    p = get_parm("param");
    mflg = p->set;

#if BTI
    if (eflg || !rflg || !dflg || !mflg) {
#else
    if (eflg || !rflg || !mflg) {
#endif
        msg("dataset (-r) and parameter file (-m) are required\n");
        do_help();  // doesn't return
    }

    ParmName = Params.ParmName;

    // get dataset information structures
    if (vflg) {
        printf("4 D   S A M   I M A G I N G\t%s\n", PRG_REV);
        printf("opening data file");
        fflush(stdout);
    }

    // get data with sensor structures
#if BTI
    if (GetMagnesInfo(DSName, PDFName, &Header, &Channel, &Epoch) == -1)
        Cleanup("can't read Magnes information");
    sprintf(DSpath, "%s/%s", Header.DsPath, Header.SetName);
#else
    GetDsInfo(DSName, &Header, &Channel, &Epoch, &Bad, TRUE);
    sprintf(DSpath, "%s/%s.ds", Header.DsPath, Header.SetName);
#endif
    sprintf(SAMpath, "%s/SAM", DSpath);

    // count data dimensions
    // set constants
    M = Header.NumPri;
    E = Header.NumEpochs;
    T = Header.MaxSamples;
    f = Header.PsIndex[0];      // 1st MEG primary sensor channel index

    if (vflg) {
        printf(" - done\n");
        printf("'%s' uses %d epochs, %d channels, %d samples/channel/epoch\n", DSName, E, M, T);
    }

    // sanity check
    if (M == 0 || E == 0 || T == 0)
        Cleanup("no valid data");

    // announce metric type
    if (vflg) {
        printf("Imaging functions of ");
        switch (Params.ImageMetric) {
            case MOMENT: printf(" moment\n");break;
            case POWER: printf("power\n"); break;
            case RV_ENTROPY: printf("rank vector entropy\n"); break;
            default: Cleanup("selected ImageMetric not supported"); break;
        }
    }

    // check imaging parameters
    if (Params.NumPrefix < 0 || Params.NumPrefix > strlen(Header.SetName))
        Cleanup("'PrefixLength' must be between 0 and the length of the dataset name.");
    if (Params.BaseStart != -999. && Params.BaseEnd != -999.)
        bflg = TRUE;
    if (Params.NumMark < 1)
        Cleanup("'NumMarkers' must be >= 1");
    if (Params.ImageMetric == -1)
        Cleanup("parameters must specify 'ImageMetric'");
    if (Params.CovType == -1)
        Cleanup("'CovType' is mandatory");
    if (Params.CovHP == -999. || Params.CovLP == -999.)          // CovBand is always required!
        Cleanup("'CovBand' specifications required");
    if (Params.ImageHP == -999. || Params.ImageLP == -999.)      // ImageBand is always required!
        Cleanup("'ImageBand' specifications required");
    if (Params.ImageHP < Channel[f].AnalogHpFreq)
        Cleanup("ImageBand highpass cutoff frequency outside recorded range");
    if (Params.ImageLP > Channel[f].AnalogLpFreq)
        Cleanup("ImageBand lowpass cutoff frequency outside recorded range");
    if (Params.SmoothHP == -999. || Params.SmoothLP == -999.)
        Cleanup("SmoothBand specifications required");
    if (Params.TimeStep < 0.)
        Cleanup("must specify 'TimeStep'");
    if (Params.FilterType == IIR)
        strcpy(FilterName, "IIR");
    else
        strcpy(FilterName, "FFT");
    BWratio = (Params.ImageLP - Params.ImageHP) / (Params.CovLP - Params.CovHP);
    if (BWratio > 1.)
        Cleanup("covariance bandwidth must be >= image bandwidth");

    // data segment extends the start & end times for each trial -- allowing filtering and entropy computations on short time-series
    time = 1. / Params.SmoothLP;
    NumBox = (int)rint(time * Header.SampleRate);
    p = get_parm("BoxWidth");
    if (p->set) {
        n = *(int *)p->ptr;
        if (n < 14 || n >= T) {
            fatalerr("improper box width %d", n);
        }
        n /= 2;             // make sure it's even
        NumBox = n * 2;
    }
    if (Params.DataSegStart == -999. || Params.DataSegEnd == -999.) {   // if DataSegment not specified, pad out by 100ms
        Params.DataSegStart = Params.Marker[0].MarkStart - time;
        Params.DataSegEnd = Params.Marker[0].MarkEnd + time;
    }

    // set up analysis constants
    time = Params.Marker[0].MarkEnd - Params.Marker[0].MarkStart;   // duration of average window (for all markers!)
    NumImg = (int)rint(time / Params.TimeStep);                     // number of 3d images
    time = Params.DataSegEnd - Params.DataSegStart;                 // duration of segment analysis window
    TSEG = (int)rint(time * Header.SampleRate);                     // number of samples in segment analysis window
    offset = Params.Marker[0].MarkStart - Params.DataSegStart;
    if (offset < 0)
        Cleanup("response time window must begin after segment start");
    if (bflg) {
        TS = (int)rint(Params.DataSegStart * Header.SampleRate);    // start of segment
        TE = (int)rint(Params.DataSegEnd * Header.SampleRate);      // end of segment
        time = Params.BaseEnd - Params.BaseStart;                   // duration of baseline window (seconds)
        TB = (int)rint(time * Header.SampleRate);                   // number of samples in baseline window
        BS = (int)rint(Params.BaseStart * Header.SampleRate);       // baseline starting sample relative to trigger marker (can be negative!)
        BE = (int)rint(Params.BaseEnd * Header.SampleRate);         // baseline ending sample relative to trigger marker (can be negative!)
        if(BS < TS || BS > TE || BE < TS || BE > TE)
            Cleanup("baseline starting & ending samples out of range");
    }

    // set prefix to NumPrefix characters (default is 8-character dataset hashcode)
    memset(Prefix, 0, sizeof(Prefix));
#if BTI
    memcpy(Prefix, "BIOMAG4D", 8);              // 4D datasets just have serial numbers
#else
    memcpy(Prefix, DSName, Params.NumPrefix);   // default: copy eight-letter NIMH hash code in dataset name
#endif

    // if req'd, initialize NOISE
    if (Params.NoiseHP != -999. && Params.NoiseLP != -999.) {
        if(Params.NoiseHP < Channel[0].AnalogHpFreq)
            Cleanup("NoiseBand highpass cutoff frequency outside recorded range");
        if(Params.NoiseLP > Channel[0].AnalogLpFreq)
        if(vflg == TRUE) {
            printf("reading noise covariance file");
            fflush(stdout);
        }
        sprintf(fpath, "%s/%s,%-d-%-dHz/Noise.cov", SAMpath, ParmName, (int)Params.CovHP, (int)Params.CovLP);
        GetCov(fpath, &CovHdr, &ChanIndex, &Cn);
        nflg = TRUE;
    }

    // if req'd, summarize epoch dimensions
    if (vflg) {
        printf("dataset '%s' has %d epochs\n", DSName, E);
        printf("\tduration %6.3f seconds/epoch\n", (double)T / Header.SampleRate);
    }

    // if req'd, read markers
    if (Params.NumMark > 0) {
#if BTI
        if (vflg) {
            printf("parsing trigger markers");
            fflush(stdout);
        }
        if (GetMagnesMarkers(DSpath, PDFName, &Header, Channel, &Marker, &NumMarkers) == -1)
            Cleanup("cannot get trigger markers");
#else
        if (vflg) {
            printf("reading 'MarkerFile.mrk'");
            fflush(stdout);
        }
        sprintf(fpath, "%s/MarkerFile.mrk", DSpath);
        GetMarkers(fpath, &Marker, &NumMarkers);
#endif
        if (vflg) {
            printf(" - found %d markers", NumMarkers);
            printf(" - done\n");
        }
    }

    // populate segment array
    if (vflg) {
        printf("populating segments:\n");
    }

    // allocate 'Stats' structure
    if ((Stats = (IMAGE_STAT *)malloc((size_t)Params.NumMark * sizeof(IMAGE_STAT))) == NULL)
        allocfailed("Stats[]");

    // for each marker...
    for (n=0; n<Params.NumMark; n++) {

        // carry over marker name
        sprintf(Stats[n].Name, "%s", Params.Marker[n].MarkName);
        if (vflg) {
            printf("\t'%s'\n", Stats[n].Name);
        }

        // count the total markers for each named marker
        for (i=NumTot=0; i<NumMarkers; i++)
            if (!strcmp(Params.Marker[n].MarkName, Marker[i].Name))
                NumTot++;
        if (NumTot == 0)
            Cleanup("no markers found with specified names");

        // allocate segment array
        if ((Stats[n].Segment = (SAM_SEG *)malloc((size_t)NumTot * sizeof(SAM_SEG))) == NULL)
            allocfailed("Stats[].Segment[]");

        // allocate metric time-series array per segment
        for (i=0; i<NumTot; i++)
            if ((Stats[n].Segment[i].sample = (double *)malloc((size_t)TSEG * sizeof(double))) == NULL)
                allocfailed("Stats[].Segment[].sample[]");

        // populate Stats for each condition
        for (i=k=Stats[n].NumSegments=0, e=(-1); i<NumMarkers; i++) {
            if (Marker[i].Epoch != e)
                e = Marker[i].Epoch;
            if (Epoch[e].Flag == TRUE) {

                // populate segment windows
                if (!strcmp(Params.Marker[n].MarkName, Marker[i].Name)) {
                    mark = Marker[i].Latency;
                    TS = (int)rint((mark + Params.DataSegStart) * Header.SampleRate) + Header.PreTrigSamples;
                    TE = (int)rint((mark + Params.DataSegEnd) * Header.SampleRate) + Header.PreTrigSamples;
                    if (TS >= 0 && TE < T) {
                        Stats[n].Segment[k].Epoch = e;
                        Stats[n].Segment[k].TS = TS;
                        Stats[n].NumSegments++;
                        k++;
                    }
                }       // if (!strcmp...
            }           // if (Epoch[e].Flag...
        }               // for (i...
        if (Stats[n].NumSegments == 0)
            Cleanup("there were no valid data segments for specified markers & time windows");
    }                   // for(n...

    // read SAM coefficients
    if (vflg) {
        printf(" - done\n");
        printf("reading SAM coefficients");
        fflush(stdout);
    }
    sprintf(WgtDir, "%s/%s,%-d-%-dHz", SAMpath, ParmName, (int)Params.CovHP, (int)Params.CovLP);
    switch (Params.CovType) {

        case GLOBAL_:       // read Global weights
            GetNoise(WgtDir, "Global", &Stats[GLOBAL_].Noise);
            if (Params.ImageFormat == TLRC)
                sprintf(fpath, "%s/Global_at.nii", WgtDir);
            else
                sprintf(fpath, "%s/Global.nii", WgtDir);
            for (n=0; n<Params.NumMark; n++) {
                GetNIFTIWts(fpath, &Stats[n].Wgt, &NumChans, &V, &NiiHdr, extension, &ExtHdr);
                if (NumChans != M)
                    Cleanup("number of channels in dataset does not match number in weight file");
                Stats[n].Noise = Stats[GLOBAL_].Noise;
            }
            break;

        case SUM_:          // read Sum weights
            GetNoise(WgtDir, "Sum", &Stats[SUM_].Noise);
            if (Params.ImageFormat == TLRC)
                sprintf(fpath, "%s/Sum_at.nii", WgtDir);
            else
                sprintf(fpath, "%s/Sum.nii", WgtDir);
            for (n=0; n<Params.NumMark; n++) {
                GetNIFTIWts(fpath, &Stats[n].Wgt, &NumChans, &V, &NiiHdr, extension, &ExtHdr);
                if (NumChans != M)
                    Cleanup("number of channels in dataset does not match number in weight file");
                Stats[n].Noise = Stats[SUM_].Noise;
            }
            break;

        case ALL_:      // read individual (All) weights
            for (n=0; n<Params.NumMark; n++) {
                GetNoise(WgtDir, Stats[n].Name, &Stats[n].Noise);
                if (Params.ImageFormat == TLRC)
                    sprintf(fpath, "%s/%s_at.nii", WgtDir, Stats[n].Name);
                else
                    sprintf(fpath, "%s/%s.nii", WgtDir, Stats[n].Name);
                GetNIFTIWts(fpath, &Stats[n].Wgt, &NumChans, &V, &NiiHdr, extension, &ExtHdr);
                if (NumChans != M)
                    Cleanup("number of channels in dataset does not match number in weight file");
            }
            break;

        default:
            Cleanup("no such CovType");
            break;
    }

    // allocate memory
    if (vflg) {
        printf(" - done\n");
        printf("allocating memory");
        fflush(stdout);
    }
    if((ImgM = (float **)malloc((size_t)NumImg * sizeof(float *))) == NULL)
        allocfailed("ImgM[]");
    for(d=0; d<NumImg; d++)
        if((ImgM[d] = (float *)malloc((size_t)V * sizeof(float))) == NULL)
            allocfailed("ImgM[][]");
    if((ImgV = (float **)malloc((size_t)NumImg * sizeof(float *))) == NULL)
        allocfailed("ImgV[]");
    for(d=0; d<NumImg; d++)
        if((ImgV[d] = (float *)malloc((size_t)V * sizeof(float))) == NULL)
            allocfailed("ImgV[][]");
    if((x = (double ***)malloc((size_t)E * sizeof(double **))) == NULL)
        allocfailed("x[]");
    for(e=0; e<E; e++) {
        if((x[e] = (double **)malloc((size_t)T * sizeof(double *))) == NULL)
            allocfailed("x[][]");
        for(t=0; t<T; t++)
            if((x[e][t] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
                allocfailed("x[][][]");
    }
    if((In = (double *)malloc((size_t)T * sizeof(double))) == NULL)
        allocfailed("In[]");
    if((Out = (double *)malloc((size_t)T * sizeof(double))) == NULL)
        allocfailed("Out[]");
    if((Avg = (double *)malloc((size_t)TSEG * sizeof(double))) == NULL)
        allocfailed("Avg[]");
    if((Var = (double *)malloc((size_t)TSEG * sizeof(double))) == NULL)
        allocfailed("Var[]");

    // compute data filter
    if (vflg) {
        printf(" - done\n");
        printf("making %6.3f to %6.3f Hz %s filter for weight bandpass", Params.ImageHP, Params.ImageLP, FilterName);
        fflush(stdout);
    }
    if (Params.FilterType == IIR)
        mkfilter(&DataIIR, SIG_ORDER, Header.SampleRate, Params.ImageHP, Params.ImageLP, &BWFreq);
    else
        Butterworth(&DataFFT, T, Params.ImageHP, Params.ImageLP, Header.SampleRate, BANDPASS, MAXORDER, &BWFreq);

    // compute notch filter
    if (Params.Notch) {
        if (Params.FilterType == FFT) {
            if (vflg) {
                printf(" - done\n");
                printf("making FFT notch filter");
                fflush(stdout);
            }
            mknotch(&NotchFFT, Params.CovHP, Params.CovLP, Header.SampleRate, T, Params.Hz);
        } else {
            msg("\nreminder: when using IIR filters, data should be notch filtered for power mains prior to analysis\n");
        }
    }

    // summarize boxcar integrator & offset
    if (vflg) {
        printf(" - done\n");
        printf("boxcar integrator: %d samples (%f sec)\n", NumBox, (double)NumBox/Header.SampleRate);
    }

    // read & filter MEG data
    if (vflg) {
        printf("reading & filtering MEG data");
        fflush(stdout);
    }
    for (e=0; e<E; e++) {
        for (m=0; m<M; m++) {
#if BTI
            if (GetMagnesChan(DSpath, PDFName, &Header, Channel, e, m, In) == -1)
                Cleanup("'GetMagnesChan()' failed");
#else
            n = Header.PsIndex[m];
            scale = Channel[n].Scale;
            GetDsData(&Header, e, n, scale, In);
#endif
            demean(In, T);
            if (Params.FilterType == IIR) {
                bdiir(In, Out, T, &DataIIR);
                for (t=0; t<T; t++)
                    x[e][t][m] = Out[t];
            } else {
                FFTfilter(In, Out, &DataFFT, T);
                if (Params.Notch == TRUE) {
                    FFTfilter(Out, In, &NotchFFT, T);
                    for (t=0; t<T; t++)
                        x[e][t][m] = In[t];
                } else {
                    for (t=0; t<T; t++)
                        x[e][t][m] = Out[t];
                }
            }
        }
        if (vflg) {                 // if req'd - heartbeat
            printf(".");
            fflush(stdout);
        }
    }

    // for each named marker
    for (n=0; n<Params.NumMark; n++) {

        // announce condition
        if (vflg) {
            printf(" - done\n");
            printf("computing image voxels for '%s':\n", Stats[n].Name);
        }

        // loop on all coordinates in the ROI
        for (c=0, pct=(-1); c<V; c++) {

            // get voxel weight vector & determine if within head boundary
            for (m=0, tmp=0; m<M; m++)
                tmp += fabs(Stats[n].Wgt[c][m]);

            if (tmp > 0.) {     // compute image only for non-zero weights

                // compute rms noise
                if (!nflg) {
                    for (m=0, tmp=0.; m<M; m++)
                        tmp += Stats[n].Wgt[c][m] * Stats[n].Noise * Stats[n].Wgt[c][m];
                    tmp *= BWratio;
                } else {
                    for (i=0, tmp=0.; i<M; i++)
                        for (j=0; j<M; j++)
                            tmp += Stats[n].Wgt[c][i] * Cn[i][j] * Stats[n].Wgt[c][j];
                    tmp *= (BWFreq / CovHdr.BWFreq);    // adjust noise power to image bandwidth
                }
                Noise = sqrt(tmp);

                // compute sam time-series for each segment
                for (k=0; k<Stats[n].NumSegments; k++) {

                    // set up segment initialization
                    e = Stats[n].Segment[k].Epoch;
                    tt = Stats[n].Segment[k].TS;

                    // compute SAM time series for segment
                    for (t=0; t<TSEG; t++, tt++) {
                        for (m=0, tmp=0.; m<M; m++)
                            tmp += x[e][tt][m] * Stats[n].Wgt[c][m];
                        In[t] = tmp / Noise;
                    }

                    switch (Params.ImageMetric) {
                        case MOMENT:        // event-related moment
                            break;
                        case POWER:         // power estimate in the box
                            for (t=NumBox/2; t<(TSEG-NumBox/2); t++) {
                                TS = t - NumBox / 2;
                                TE = t + NumBox / 2;
                                Stats[n].Segment[k].sample[t] = power(In, TS, TE);
                            }
                            break;
                        case RV_ENTROPY:    // rank vector entropy
                            Data2RVE(In, Stats[n].Segment[k].sample, TSEG, Params.Dims, Params.ImageLP, Params.Tau, Header.SampleRate);
                            break;
                        default:
                            Cleanup("ImageMetric not implemented");
                            break;
                    }

                }   // for (k...

                // compute average & variance
                for (t=0; t<TSEG; t++) {         // for each sample in the data segments
                    for (k=0, a1=a2=0.; k<Stats[n].NumSegments; k++) {
                        a1 += Stats[n].Segment[k].sample[t];
                        a2 += Stats[n].Segment[k].sample[t] * Stats[n].Segment[k].sample[t];
                    }
                    Avg[t] = (Stats[n].NumSegments > 2)? a1 / (double)Stats[n].NumSegments: 0.;
                    Var[t] = (Stats[n].NumSegments > 2)? (a2 - (a1 * a1 / (double)Stats[n].NumSegments)) / (double)(Stats[n].NumSegments - 1.): 0.;
                }

                // if req'd, remove baseline from Avg
                if (bflg) {
                    for (t=0, tt=BS, tmp=0.; t<TB; t++)
                        tmp += Avg[t];
                    mean = tmp / (double)TB;
                    for (t=0; t<TSEG; t++)
                        Avg[t] -= mean;
                }

                // downsample average & variance
                for (d=0, time=offset; d<NumImg; d++, time+=Params.TimeStep) {
                    t = (int)rint(time * Header.SampleRate);
                    ImgM[d][c] = (float)Avg[t];
                    ImgV[d][c] = (float)Var[t];
                }

            } else {    // if (tmp...
                for (d=0; d<NumImg; d++) {
                    ImgM[d][c] = 0.;
                    ImgV[d][c] = 0.;
                }
            }

            // percent done
            if (vflg) {
                percent = 100. * (double)c / (double)V;
                if (rint(percent) > (double)pct) {
                    pct = (int)rint(percent);
                    printf("\r\t%d percent done", pct);
                    fflush(stdout);
                }
            }
        }           // for (c...

        // write mean & variance images
        if (vflg) {
            printf("\nwriting 'Mean'");
            fflush(stdout);
        }
        p = get_parm("OutName");
        if (p->set) {
            ParmName = (char *)p->ptr;
        }
        if (!strncmp(Params.DirName, "NULL", 4))
            sprintf(Name1, "%s/%s,%s,%s,", SAMpath, Prefix, ParmName, Stats[n].Name);
        else
            sprintf(Name1, "%s/%s,%s,%s,", Params.DirName, Prefix, ParmName, Stats[n].Name);
        switch (Params.ImageMetric) {
            case MOMENT:        sprintf(Name2, "%s4D_MOM,", Name1); break;
            case POWER:         sprintf(Name2, "%s4D_PWR,", Name1); break;
            case RV_ENTROPY:    sprintf(Name2, "%s4D_RVE,", Name1); break;
            default:            Cleanup("ImageMetric not implemented"); break;
        }
        sprintf(fpath, "%sMean.nii", Name2);
        if ((fp = fopen(fpath, "w")) == NULL)
            Cleanup("can't open .nii file for write");

        // Set output range (default all)
        TS = 0;
        TE = NumImg;
        tmp3 = Params.DataSegStart;
        p = get_parm("OutRange");
        if (p->set) {
            outrange = (TIMERANGE *)p->ptr;
            tmp = outrange->t0 - Params.DataSegStart;
            tmp2 = NumImg * Params.TimeStep;
            if (tmp < 0. || tmp > tmp2) {               // ignore it if it's out of range
                tmp = 0.;
            }
            TS = (int)floor(tmp / Params.TimeStep);
            tmp = outrange->t1 - Params.DataSegStart;
            if (tmp < 0. || tmp > tmp2) {
                tmp = tmp2;
            }
            TE = (int)floor(tmp / Params.TimeStep);

            tmp3 = Params.DataSegStart + TS * Params.TimeStep;

            if (vflg) {
                tmp = tmp3 + (TE - TS) * Params.TimeStep;
                printf(" -- writing from %g to %g", tmp3, tmp);
            }
        }

        // modify NIFTI header
        memset(extension, 0, 4);                                    // for now, ignore the fact that we may have an extended header
        NiiHdr.dim[0] = 4;                                          // 3d+time
        NiiHdr.dim[4] = TE - TS;                                    // number of 3d images
        NiiHdr.toffset = (float)tmp3;                               // time offset shift
        NiiHdr.slice_duration = (float)Params.TimeStep;             // time step size
        NiiHdr.pixdim[4] = (float)Params.TimeStep;                  // TR is really stored here
        NiiHdr.slice_code = NIFTI_SLICE_SEQ_INC;
        NiiHdr.xyzt_units = (NIFTI_UNITS_MM | NIFTI_UNITS_SEC);     // 3d+time units
        memset(NiiHdr.intent_name, 0, 16);
        memcpy(NiiHdr.intent_name, "Mean", 4);

        // write header & extension
        if (fwrite((void *)&NiiHdr, sizeof(nifti_1_header), 1, fp) != 1)
            Cleanup("can't write nifti header to .nii file");
        if (fwrite((void *)extension, 4, 1, fp) != 1)
            Cleanup("can't write 'extension' to .nii file");
        if ((int)NiiHdr.vox_offset != 352 && ExtHdr != NULL) {
            N = (int)NiiHdr.vox_offset - 352;
            if (fwrite((void *)ExtHdr, N, 1, fp) != 1)
                Cleanup("can't write extended header to .nii file");
        }
        for (d=TS; d<TE; d++)   // write time as BRIKs
            if (fwrite((void *)ImgM[d], sizeof(float), V, fp) != V)
                Cleanup("can't write data image file");
        fclose(fp);

        // write variance image
        if (vflg) {
            printf(" - done\n");
            printf("writing 'Variance'");
            fflush(stdout);
        }
        sprintf(fpath, "%sVariance.nii", Name2);
        if ((fp = fopen(fpath, "w")) == NULL)
            Cleanup("can't open .nii file for write");

        // modify NIFTI header
        memset(extension, 0, 4);                                    // for now, ignore the fact that we may have an extended header
        NiiHdr.dim[0] = 4;                                          // 3d+time
        NiiHdr.dim[4] = TE - TS;                                    // number of 3d images
        NiiHdr.toffset = (float)tmp3;                               // time offset shift
        NiiHdr.slice_duration = (float)Params.TimeStep;             // time step size
        NiiHdr.pixdim[4] = (float)Params.TimeStep;                  // TR is really stored here
        NiiHdr.slice_code = NIFTI_SLICE_SEQ_INC;
        NiiHdr.xyzt_units = (NIFTI_UNITS_MM | NIFTI_UNITS_SEC);     // 3d+time units
        memset(NiiHdr.intent_name, 0, 16);
        memcpy(NiiHdr.intent_name, "Variance", 8);

        // write header & extension
        if (fwrite((void *)&NiiHdr, sizeof(nifti_1_header), 1, fp) != 1)
            Cleanup("can't write nifti header to .nii file");
        if (fwrite((void *)extension, 4, 1, fp) != 1)
            Cleanup("can't write 'extension' to .nii file");
        if ((int)NiiHdr.vox_offset != 352 && ExtHdr != NULL) {
            N = (int)NiiHdr.vox_offset - 352;
            if(fwrite((void *)ExtHdr, N, 1, fp) != 1)
                Cleanup("can't write extended header to .nii file");
        }
        for (d=TS; d<TE; d++)    // write time as BRIKs
            if (fwrite((void *)ImgV[d], sizeof(float), V, fp) != V)
                Cleanup("can't write data image file");
        fclose(fp);
    }               // for (n...

    if (vflg) {
        printf(" - done\n'%s' done\n", Progname);
    }
    exit(0);
}
