// sam_Nwts -- linear synthetic aperture magnetometry coefficient solution.
//  Linear solution for the Cartesian moment vector. Optional 2nd covariance
//  matrix for determining the beamformer orientation vector.
//
//  flags:
//      -r <run name>   -- MEG dataset name
//      -m <parameters> -- parameter file name
//      -Z              -- noise normalization
//      -v              -- Verbose mode
//
//  Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <model.h>
#include <samlib.h>
#include <siglib.h>
#include <lfulib.h>
#include <DataFiles.h>
#include <SAMfiles.h>
#include <byteswap.h>
#include <version.h>
#include "samutil.h"
#include "sam_parse.h"

#define MINOR_REV   5

extern int CoeffFlag;                   // for initializing Nolte solution

int main(
    int             argc,
    char            **argv
) {
    PARMINFO        Params;             // analysis parameters
    PARM            *p;                 // pointer into active parameter list
    HEAD_MODEL      Model;              // head model structure for *.hdm file
    HULL            Hull;               // cortical hull model -- vertices & normals
    VOXELINFO       *Target;            // Target[N] -- structure for target parameters
    HeaderInfo      Header;             // MEG data header
    ChannelInfo     *Channel;           // MEG channel info
    EpochInfo       *Epoch;             // MEG epoch info
    COV_HDR         WgtCovHdr;          // weight covariance header
    COV_HDR         OriCovHdr;          // orientation covariance header
    SAM_HDR         SAMHeader;          // SAM file header
    double          **Co;               // Co[M][M] -- orientation covariance matrix
    double          **Cf;               // Cf[M][M] -- solution covariance matrix
    double          **L;                // L[M][2N/3N] -- n-target lead field matrix
    double          **Bin;              // Bin[M][N] -- n-target forward solution matrix
    double          **Bout;             // Bout[M][N] -- n-target forward solution matrix (orthogonalized)
    double          **W;                // W[M][N] -- n-target weights
    double          HPFreq = 0.;        // highpass frequency
    double          LPFreq = 0.;        // lowpass frequency
    double          Noise = 0.;         // effective rank-limited noise
    double          Mu_f;               // weight solution regularization
    double          Mu_o;               // orientation regularization
    double          Dtmp;
    register double tmp;                // what else?
    int             *ChanIndex;         // ChanIndex[M] -- index of channels used for covariance
    int             M;                  // number of MEG or reference channels used
    int             N;                  // number of targets
    int             R;                  // number of physical reference SQUID channels
    int             S;                  // number of physical SQUID channels
    int             d;                  // dipole index
    int             v;                  // vector index
    int             i;                  // matrix index
    int             j;                  // matrix index
    int             m;                  // sensor index
    int             length;             // string length
    int             found;              // string found flag
    int             vectors = 0;        // number of vectors (2 or 3)
    static int      eflg = FALSE;       // command-line error flag
    static int      pflg = FALSE;       // parameter file flag
    static int      rflg = FALSE;       // dataset name flag
    static int      tflg = FALSE;       // target file flag
    static int      vflg = FALSE;       // verbose mode flag (default: silent)
    static int      zflg = FALSE;       // noise normalization flag
    char            fpath[256];         // general path name
    char            *DSName = NULL;     // MEG dataset name
    char            DSpath[256];        // MEG dataset path
    char            SAMpath[256];       // SAM subdirectory path
    char            HullPath[256];      // hull.shape path
    char            Prefix[100];        // dataset prefix for finding MRI
    char            CovDirName[256];    // covariance subdirectory name
    char            *ParmName = NULL;   // name of parameter file
    char            CoeffID[] =         // SAM image file ID
        "SAMCOEFF";
    FILE            *sp;                // SAM data pointer

    // system-dependent variables
#if BTI
    static int      dflg = FALSE;       // pdf file name flag
    char            *PDFName;           // name of data file
#endif

    Progname = argv[0];
    reg_usage(MINOR_REV, __DATE__);

    // Register the parameters used by this program.

    reg_std_parm();

    reg_parm("CovBand");
    reg_parm("ImageBand");
    reg_parm("Mu");

    reg_parm("Model");
    reg_parm("TargetName");
    reg_parm("MRIDirectory");
    reg_parm("MRIPattern");
    reg_parm("PrefixLength");
    reg_parm("HullName");
    reg_parm("Normalize");

    // Parse the arguments, environment variables, and parameter files.

    do_parse_args(argc, argv);

    // Extract specified parameters.

    eflg = get_params(&Params);

    p = get_parm("DataSet");
    if (p->set) {
        rflg = TRUE;
        DSName = Params.DataSetName;
    }
#if BTI
    p = get_parm("PDFName");
    if (p->set) {
        dflg = TRUE;
        PDFName = (char *)p->ptr;
    }
#endif
    p = get_parm("verbose");
    vflg = p->set;
    p = get_parm("param");
    pflg = p->set;
    if (pflg) ParmName = Params.ParmName;

    p = get_parm("Normalize");      // -Z
    zflg = p->set;
    p = get_parm("TargetName");     // -t
    tflg = p->set;

#if BTI
    if(eflg || !rflg || !dflg || !pflg || !tflg) {
#else
    if(eflg || !rflg || !pflg || !tflg) {
#endif
        msg("dataset (-r), parameter file (-m), and target file (-t) are required\n");
        do_help();  // doesn't return
    }

    // announce
    if(vflg) {
        printf("N - S O U R C E   S A M   C O E F F I C I E N T   G E N E R A T O R\t%s\n", PRG_REV);
        printf("opening data file");
        fflush(stdout);
    }

    // get data with sensor structures
#if BTI
    if(GetMagnesInfo(DSName, PDFName, &Header, &Channel, &Epoch) == -1)
        Cleanup("can't read Magnes information");
    sprintf(DSpath, "%s/%s", Header.DsPath, Header.SetName);
#else
    GetDsInfo(DSName, &Header, &Channel, &Epoch, NULL, TRUE);
    sprintf(DSpath, "%s/%s.ds", Header.DsPath, Header.SetName);
#endif
    sprintf(SAMpath, "%s/SAM", DSpath);

    // set channel & epoch count constants
    M = Header.NumPri;
    R = Header.NumRef;
    S = Header.NumPri + Header.NumRef;

    // define forward solution model
    if(Params.Model == -1)
        Cleanup("'Model' specification missing from parameter file");
    if(vflg) {
        printf(" - done\n");
        printf("modeling forward solution using ");
        fflush(stdout);
    }

    switch(Params.Model) {

        // single sphere origin
        case SSPHERE:
            if(vflg) {
                printf("single local sphere origin: { %6.3f, %6.3f, %6.3f } (cm)", 100. * Params.Sp[X_], 100. * Params.Sp[Y_], 100. * Params.Sp[Z_]);
                fflush(stdout);
            }
            model = SPHERE;
            for(i=0; i<M; i++) {
                j = Header.PsIndex[i];
                for(v=X_; v<=Z_; v++)
                    Channel[j].Geom.MEGSensor.LocalSphere[v] = Params.Sp[v];
            }
            for(i=0; i<R; i++) {
                j = Header.RsIndex[i];
                for(v=X_; v<=Z_; v++)
                    Channel[j].Geom.MEGSensor.LocalSphere[v] = Params.Sp[v];
            }
            Hull.nv = 1;
            if((Hull.vertex = (FIELD *)malloc((size_t)2 * sizeof(FIELD))) == NULL)
                allocfailed("Hull.vertex[1]");
            for(v=X_; v<=Z_; v++)
                Hull.Vo[v] = Params.Sp[v];
            vectors = 2;                // only two moment vectors
            break;

        // multiple local spheres
        case MSPHERE:

            if(vflg) {
                printf("multiple local spheres");
                fflush(stdout);
            }

            // read .hdm file into 'Model'
            sprintf(fpath, "%s/default.hdm", DSpath);
            GetHDM(fpath, &Model);

            // test if the model matches the number of sensors
            if(Model.NumSensors != S)
                Cleanup("number of sensors in '.hdm' file does not match that of the dataset");

            // set multiple sphere origins
            for(i=0; i<S; i++) {
                length = (int)strlen(Model.MultiSphere[i].ChanName) - 1;
                for(j=0, found=FALSE; j<Header.NumChannels; j++)
                    if(!strncmp(Model.MultiSphere[i].ChanName, Channel[j].ChannelName, length)) {
                        for(v=X_; v<=Z_; v++)
                            Channel[j].Geom.MEGSensor.LocalSphere[v] = Model.MultiSphere[i].Origin[v];
                        found = TRUE;
                    }
                if(found == FALSE)
                    Cleanup("could not find matching channel name in '.HDM' file");
            }
            vectors = 3;                    // 3 moment vectors
            break;

        // Nolte realistic head model
        case NOLTE:

            if(vflg) {
                printf("'hull.shape'");
                fflush(stdout);
            }
            memset(Prefix, 0, sizeof(Prefix));
            memcpy(Prefix, DSName, Params.NumPrefix);
            p = get_parm("HullName");
            if (p->set) {
                GetMRIPath(HullPath, sizeof(HullPath), &Params, (char *)p->ptr, TRUE);
            } else {
                GetMRIPath(HullPath, sizeof(HullPath), &Params, "hull.shape", TRUE);
            }
            GetHull(HullPath, &Hull);       // read hull model from MRI directory
            CoeffFlag = TRUE;
            vectors = 3;                    // 3 moment vectors
            break;

        default:
            break;
    }

    // first try to use CovBand frequencies -- use ImageBand only if CovBand has not been specified!
    if(Params.CovHP >= 0. && Params.CovLP >= 0.) {
        HPFreq = Params.CovHP;
        LPFreq = Params.CovLP;
    } else if(Params.ImageHP >= 0. && Params.ImageLP >= 0.) {
        HPFreq = Params.ImageHP;
        LPFreq = Params.ImageLP;
    } else {
        Cleanup("Neither CovBand nor ImageBand were specified");
    }

    // generate CovDirName
    sprintf(CovDirName, "%s,%-d-%-dHz", ParmName, (int)HPFreq, (int)LPFreq);

    // read targets
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("reading target voxel list");
        fflush(stdout);
    }
    GetMRIPath(fpath, sizeof(fpath), &Params, Params.Target, TRUE);
    GetTargets(fpath, &Target, &N);
    if(vflg) {
        printf(" - found %d targets", N);
        fflush(stdout);
    }

    // allocate memory for arrays
    if(vflg) {
        printf(" - done\n");
        printf("allocating memory");
        fflush(stdout);
    }
    if((L = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
        allocfailed("L[]");
    for(i=0; i<M; i++)
        if((L[i] = (double *)malloc((size_t)(N * vectors) * sizeof(double))) == NULL)
            allocfailed("L[][]");
    if((Bin = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
        allocfailed("Bin[]");
    for(i=0; i<M; i++)
        if((Bin[i] = (double *)malloc((size_t)N * sizeof(double))) == NULL)
            allocfailed("Bin[][]");
    if((Bout = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
        allocfailed("Bout[]");
    for(i=0; i<M; i++)
        if((Bout[i] = (double *)malloc((size_t)N * sizeof(double))) == NULL)
            allocfailed("Bout[][]");
    if((W = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
        allocfailed("W[]");
    for(i=0; i<M; i++)
        if((W[i] = (double *)malloc((size_t)N * sizeof(double))) == NULL)
            allocfailed("W[][]");

    // read required covariance files
    if(vflg) {
        printf(" - done\n");
        printf("reading covariance file(s)");
        fflush(stdout);
    }
    sprintf(fpath, "%s/%s/Global.cov", SAMpath, CovDirName);
    GetCov(fpath, &WgtCovHdr, &ChanIndex, &Cf);
    sprintf(fpath, "%s/%s/Orient.cov", SAMpath, CovDirName);
    GetCov(fpath, &OriCovHdr, &ChanIndex, &Co);

    // announce regularization used
    if(vflg == TRUE && Params.Operator != -1) {
        printf("regularizing covariance matrices by ");
        switch(Params.Operator) {
            case ADD_MU:
                printf("adding %f fT/root-Hz to diagonal\n", Params.Mu);
                break;
            case MPY_MU:
                printf("multiplying diagonal by %f\n", Params.Mu);
                break;
            default:
                break;
        }
        fflush(stdout);
    }

    // if req'd, regularize
    switch(Params.Operator) {
        case ADD_MU:
            tmp = 1.0e-30 * Params.Mu * Params.Mu;      // convert fT/root-Hz to T^2/Hz
            Mu_f = WgtCovHdr.BWFreq * tmp;
            Mu_o = OriCovHdr.BWFreq * tmp;
            for(m=0; m<M; m++) {
                Cf[m][m] += Mu_f;
                Co[m][m] += Mu_o;
            }
            Noise += Mu_f;
            break;
        case MPY_MU:
            for(m=0; m<M; m++) {
                Cf[m][m] *= Params.Mu;
                Co[m][m] *= Params.Mu;
            }
            Noise *= Params.Mu;
            break;
        default:
            break;  // do nothing
    }

    // apply simultaneous weight solution
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("solving simultaneous beamformer weights");
        fflush(stdout);
    }
    switch(Params.Model) {
        case SSPHERE: NLeadField2D(Target, L, &Header, Channel, N); break;
        case MSPHERE: NLeadField3D(Target, L, &Header, Channel, N); break;
        case NOLTE: NLeadField(Target, L, &Header, Channel, &Hull, N); break;
        default: break;
    }
    NSolveFwd(Target, L, Bin, Co, N, M, vectors);
    NSolveWts(Bin, Cf, W, N, M);

    // if req'd, normalize weights
    if(zflg == TRUE)
        for(i=0; i<N; i++) {
            for(j=0, tmp=0.; j<M; j++)
                tmp += W[j][i] * Noise * W[j][i];
            Dtmp = sqrt(tmp);
            for(j=0; j<M; j++)
                W[j][i] /= Dtmp;
        }

    // fill in SAM weight header
    SAMHeader.Version = bswap_32(SAM_REV);          // file version number
    sprintf(SAMHeader.SetName, "%s", DSName);       // name of parent dataset
    SAMHeader.NumChans = bswap_32(M);               // number of channels used by SAM
    Dtmp = WgtCovHdr.HPFreq;
    SAMHeader.HPFreq = Dflip(&Dtmp);                // highpass frequency (Hz)
    Dtmp = WgtCovHdr.LPFreq;
    SAMHeader.LPFreq = Dflip(&Dtmp);                // lowpass frequency (Hz)
    Dtmp = WgtCovHdr.BWFreq;
    SAMHeader.BWFreq = Dflip(&Dtmp);                // bandwidth of filters (Hz)
    Dtmp = Noise;
    SAMHeader.MeanNoise = Dflip(&Dtmp);             // mean primary sensor noise (T^2)
    sprintf(SAMHeader.MriName, "NONE");             // a place-holder until we decide whether to read the default.hdm file
    for(v=S_; v<=A_; v++) {
        SAMHeader.Nasion[v] = bswap_32(0);
        SAMHeader.RightPA[v] = bswap_32(0);
        SAMHeader.LeftPA[v] = bswap_32(0);
    }
    Dtmp = 0.;
    SAMHeader.XStart = Dflip(&Dtmp);                // x-start coordinate (m)
    Dtmp = 0.;
    SAMHeader.XEnd = Dflip(&Dtmp);                  // x-end coordinate (m)
    Dtmp = 0.;
    SAMHeader.YStart = Dflip(&Dtmp);                // y-start coordinate (m)
    Dtmp = 0.;
    SAMHeader.YEnd = Dflip(&Dtmp);                  // y-end coordinate (m)
    Dtmp = 0.;
    SAMHeader.ZStart = Dflip(&Dtmp);                // z-start coordinate (m)
    Dtmp = 0.;
    SAMHeader.ZEnd = Dflip(&Dtmp);                  // z-end coordinate (m)
    Dtmp = 0.;
    SAMHeader.StepSize = Dflip(&Dtmp);              // voxel step size (m)
    SAMHeader.NumWeights = bswap_32(N);             // number of SAM virtual channels
    SAMHeader.SAMType = bswap_32(SAM_TYPE_WT_LIST);
    SAMHeader.SAMUnit = bswap_32(SAM_UNIT_COEFF);   // redundant, but -- so what?

    // create weight file
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("writing SAM legacy weight file");
        fflush(stdout);
    }
    sprintf(fpath, "%s/%s,%s.wts", SAMpath, ParmName, Params.Target);
    if((sp = fopen(fpath, "w")) == NULL)
        Cleanup("can't open image file for write");
    if(fwrite((void *)CoeffID, 8, 1, sp) != 1)
        Cleanup("can't write ID to image file");
    if(fwrite((void *)&SAMHeader, sizeof(SAM_HDR), 1, sp) != 1)
        Cleanup("can't write header to image file");
    for(i=0; i<M; i++) {
        j = bswap_32(ChanIndex[i]);
        if(fwrite((void *)&j, sizeof(int), 1, sp) != 1)
            Cleanup("can't write channel index to weight file");
    }

    // write legacy weights (needed for DataEditor)
    for(d=0; d<N; d++)
        for(i=0; i<M; i++) {    // write weight
            Dtmp = Dflip(&W[i][d]);
            if(fwrite((void *)&Dtmp, sizeof(double), 1, sp) != 1)
                Cleanup("can't write weight to file");
        }
    fclose(sp);                 // close weight file

    // now write NIFTI weight file
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("writing NIFTI weight file");
        fflush(stdout);
    }
    Params.SAMStart[X_] = 0.05;
    Params.SAMEnd[X_] = 0.05;
    Params.SAMStart[Y_] = 0.;
    Params.SAMEnd[Y_] = 0.;
    Params.SAMStart[Z_] = 0.001;
    Params.SAMEnd[Z_] = (double)N * 0.001;
    Params.SAMStep = 0.001;
    sprintf(fpath, "%s/%s,%-d-%-dHz/Global.nii", SAMpath, ParmName, (int)HPFreq, (int)LPFreq);
    PutNIFTIWts(fpath, &Params, W ,Noise, M);

    // finit
    if(vflg) {
        printf("'%s' done\n", Progname);
    }
    exit(0);
}
