// sam_entropy -- synthetic aperture magnetometry imaging for single-state
//  signals. sam_entropy reads SAM weights for volume image, & computes
//  the source waveform for each voxel. sam_entropy computes 3D + time
//  images of rank-vector entropy. The output is the change in entropy
//  relative to the mean entropy.
//
//  flags:
//      -r <dataset name>   -- MEG dataset name
//      -d <data file name> -- MEG data file name (4D, only)
//      -m <analysis name>  -- name of analysis file
//      -t                  -- total entropy
//      -n                  -- normalize by variance
//      -v                  -- verbose mode
//
//  Author: Stephen E. Robinson
//      MEG Core Facility
//      NIMH
//

// include...
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <geoms.h>
#include <byteswap.h>
#include <samlib.h>
#include <siglib.h>
#include <rvelib.h>
#include <DataFiles.h>
#include <SAMfiles.h>
#include <filters.h>
#include <TimeSegs.h>
#include <version.h>
#include "samutil.h"
#include "sam_parse.h"

// internal constants
#define MINOR_REV   7
#define SIG_ORDER   4

int main(
     int    argc,
     char   **argv
) {
    IIRSPEC         DataIIR;            // IIR data filter structure
    IIRSPEC         SmoothIIR;          // IIR smoothing filter structure
    FFTSPEC         DataFFT;            // FFT data filter structure
    FFTSPEC         SmoothFFT;          // FFT smoothing filter structure
    FFTSPEC         NotchFFT;           // FFT notch filter structure (sum of notches)
    PARMINFO        Params;             // analysis parameters
    PARM            *p;                 // pointer into active parameter list
    HeaderInfo      Header;             // MEG data header
    ChannelInfo     *Channel;           // MEG channel info
    EpochInfo       *Epoch;             // MEG epoch info
    nifti_1_header  NiiHdr;             // NIFTI header
    unsigned char   *ExtHdr;            // NiiHdr[N] -- extended NIFTI header
    double          **x;                // x[TT][M] -- filtered multisensor data
    double          **Wgt;              // Wgt[V][M] -- active SAM weights
    double          *In;                // In[TT] -- input time-series
    double          *Out;               // Out[TT] -- output time-series
    double          *dd;                // tmp pointer
    double          scale;              // bits --> Tesla conversion
    double          BWFreq;             // effective bandwidth
    double          percent;            // percent done
    double          mean;               // mean of RVE time-series
    double          time;               // time
    double          var;                // variance
    float           **Image;            // Image[D][V] -- image voxels
    short           *Sym;               // Sym[TT] -- symbolic time-series
    register double tmp;                // what else?
    int             E;                  // number of good epochs
    int             M;                  // number of MEG or reference channels used
    int             N;                  // size of extended header
    int             T;                  // number of samples
    int             TE;                 // samples used
    int             TT;                 // total samples E*T
    int             V;                  // number of voxels for active state weight file
    int             c;                  // input character & virtual channel
    int             pct;                // integer percentage
    int             m;                  // channel index
    int             n;                  // alternate channel index
    int             d;                  // SAM image index
    int             e;                  // epoch index
    int             f;                  // 1st primary sensor index
    int             t;                  // sample index
    int             tt;                 // alternate sample index
    int             i;                  // matrix index
    int             NumChans;           // number of channels in weight file
    int             NumImg;             // number of SAM images
    static int      eflg = FALSE;       // command-line error flag
    static int      mflg = FALSE;       // analysis parameter file flag
    static int      nflg = FALSE;       // normalize by variance flag
    static int      rflg = FALSE;       // dataset name flag
    static int      vflg = FALSE;       // verbose mode flag
    char            fpath[256];         // general path name
    char            *DSName = NULL;     // MEG dataset name
    char            DSpath[256];        // MEG dataset path
    char            Prefix[64];         // dataset name prefix
    char            SAMpath[256];       // SAM subdirectory path
    char            *ParmName = NULL;   // parameter file name
    char            FilterName[8];      // filter name -- IIR | FFT
    char            extension[4];       // extension
    FILE            *fp;                // data output file pointer array

    // system-dependent variables
#if BTI
    static int      dflg;           // pdf file name flag
    char            *PDFName;       // name of data file
#else
    unsigned char   **Bad;          // Bad[E][T] -- flags for bad samples
#endif

    Progname = argv[0];
    reg_usage(MINOR_REV, __DATE__);

    // Register the parameters used by this program.

    reg_std_parm();

    reg_parm("NumMarkers");
    reg_parm("Marker");

    reg_parm("CovType");
    set_parm_arg_help("CovType", "GLOBAL");
    reg_parm("CovBand");
    reg_parm("ImageBand");
//    reg_parm("NoiseBand");
    reg_parm("SmoothBand");
    reg_parm("FilterType");

    reg_parm("Normalize");
    set_parm_help("Normalize", "normalize the output envelope by its variance");
    reg_parm("Baseline");
    reg_parm("TimeStep");

    reg_parm("ImageFormat");
    reg_parm("ImageMetric");
    set_parm_arg_help("ImageMetric", "RankVectorEntropy");
    reg_parm("ImageDirectory");
    reg_parm("PrefixLength");
    reg_parm("Notch");
    reg_parm("Hz");

    // Parse the arguments, environment variables, and parameter files.

    do_parse_args(argc, argv);

    // Extract specified parameters.

    eflg = get_params(&Params);

    p = get_parm("DataSet");
    if (p->set) {
        rflg = TRUE;
        DSName = Params.DataSetName;
    }
#if BTI
    p = get_parm("PDFName");
    if (p->set) {
        dflg = TRUE;
        PDFName = (char *)p->ptr;
    }
#endif
    p = get_parm("verbose");
    vflg = p->set;
    p = get_parm("param");
    mflg = p->set;
    if (mflg) ParmName = Params.ParmName;

#if BTI
    if(eflg || !rflg || !dflg || !mflg) {
#else
    if(eflg || !rflg || !mflg) {
#endif
        msg("dataset (-r) and parameter file (-m) are required\n");
        do_help();  // doesn't return
    }

    p = get_parm("Normalize");
    nflg = p->set;

    // get dataset information structures
    if(vflg) {
        printf("S A M   E N T R O P Y   I M A G I N G   O F   S O U R C E   T I M E   S E R I E S\t%s\n", PRG_REV);
        printf("opening data file");
        fflush(stdout);
    }

    // get data with sensor structures
#if BTI
    if(GetMagnesInfo(DSName, PDFName, &Header, &Channel, &Epoch) == -1)
        Cleanup("can't read Magnes information");
    sprintf(DSpath, "%s/%s", Header.DsPath, Header.SetName);
#else
    GetDsInfo(DSName, &Header, &Channel, &Epoch, &Bad, TRUE);
    sprintf(DSpath, "%s/%s.ds", Header.DsPath, Header.SetName);
#endif
    sprintf(SAMpath, "%s/SAM", DSpath);

    // announce
    if(vflg) {
        printf(" - done\n");
        printf("computing raw images of rank vector entropy\n");
        fflush(stdout);
    }

    // set constants
    M = Header.NumPri;
    E = Header.NumEpochs;
    T = Header.MaxSamples;
    TT = E * T;
    f = Header.PsIndex[0];      // 1st MEG primary sensor channel index

    // check to see if this is a continuous dataset
    if(E != 1)
        msg(" (reminder: multi-epoch data must be contiguous)");

    // check imaging parameters
    if(Params.CovHP == -999. || Params.CovLP == -999.)
        Cleanup("CovBand parameters required");
    if(Params.ImageHP == -999. || Params.ImageLP == -999.)
        Cleanup("ImageBand parameters required");
    if(Params.ImageHP < Channel[f].AnalogHpFreq)
        Cleanup("ImageBand highpass cutoff frequency out of recorded range");
    if(Params.ImageLP > Channel[f].AnalogLpFreq)
        Cleanup("ImageBand lowpass cutoff frequency out of recorded range");
    if(Params.SmoothHP == -999. || Params.SmoothLP == -999.)
        Cleanup("SmoothBand specifications required");
    if(Params.SmoothHP < Channel[f].AnalogHpFreq)
        Cleanup("SmoothBand highpass cutoff frequency out of recorded range");
    if(Params.SmoothLP > Channel[f].AnalogLpFreq)
        Cleanup("SmoothBand lowpass cutoff frequency out of recorded range");
    if(Params.FilterType == IIR)
        strcpy(FilterName, "IIR");
    else
        strcpy(FilterName, "FFT");
    if(Params.ImageMetric != RV_ENTROPY)
        Cleanup("ImageMetric must specify RankVectorEntropy");
    if(Params.Tau < 0.)
        Cleanup("RVE tau must be greater than zero");
    if(Params.Dims < 4 || Params.Dims > 5)
        Cleanup("RVE dimensions must be 4 or 5");
    if(Params.TimeStep == -999.)
        Cleanup("TimeStep specification missing from parameter file");
    if(Params.CovType != GLOBAL_)
        Cleanup("CovType must be GLOBAL");
    if(vflg == TRUE) {
        printf(" - done\n");
        switch(Params.RemoveBaseline) {
            case NONE: printf("output RVE images without baseline removal\n"); break;
            case BYVOXEL: printf("output RVE images with mean removal by voxel\n"); break;
            case GLOBAL: printf("output RVE images with global mean removed from each voxel\n"); break;
            default: break;
        }
        fflush(stdout);
    }

    // set prefix to NumPrefix characters (default is 8-character dataset hashcode)
    memset(Prefix, 0, sizeof(Prefix));
#if BTI
    memcpy(Prefix, "BIOMAG4D", 8);  // 4D datasets just have serial numbers
#else
    memcpy(Prefix, DSName, Params.NumPrefix);
#endif

    // get weight name from parameters & read SAM weights
    if(vflg == TRUE) {
        printf("reading Global SAM weights");
        fflush(stdout);
    }
    if(Params.ImageFormat == TLRC)
        sprintf(fpath, "%s/%s,%-d-%-dHz/Global_at.nii", SAMpath, ParmName, (int)Params.CovHP, (int)Params.CovLP);
    else
        sprintf(fpath, "%s/%s,%-d-%-dHz/Global.nii", SAMpath, ParmName, (int)Params.CovHP, (int)Params.CovLP);
    GetNIFTIWts(fpath, &Wgt, &NumChans, &V, &NiiHdr, extension, &ExtHdr);
    if(NumChans != M)
        Cleanup("number of channels in dataset does not match number in weight file");

    // if req'd, summarize epoch dimensions
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("dataset '%s' has %d epochs\n", DSName, E);
        printf("\tduration %6.3f seconds/epoch\n", (double)T/Header.SampleRate);
        fflush(stdout);
    }

    // set up 3d+time imaging constants)
    time = (double)TT / Header.SampleRate - 2.; // total duration (seconds) -- less 1 second at start & end
    NumImg = (int)rint(time / Params.TimeStep); // number of 3d images

    // allocate memory
    if(vflg == TRUE) {
        printf("solving for %d SAM images\n", NumImg);
        printf("allocating memory");
        fflush(stdout);
    }
    if((x = (double **)malloc((size_t)TT * sizeof(double *))) == NULL)
        allocfailed("x[]");
    for(t=0; t<TT; t++)
        if((x[t] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
            allocfailed("x[][]");
    if((In = (double *)malloc((size_t)TT * sizeof(double))) == NULL)
        allocfailed("In[]");
    if((Out = (double *)malloc((size_t)TT * sizeof(double))) == NULL)
        allocfailed("Out[]");
    if((Sym = (short *)malloc((size_t)TT * sizeof(short))) == NULL)
        allocfailed("Sym[]");
    if((Image = (float **)malloc((size_t)NumImg * sizeof(float *))) == NULL)
        allocfailed("Image[]");
    for(d=0; d<NumImg; d++)
        if((Image[d] = (float *)malloc((size_t)V * sizeof(float))) == NULL)
            allocfailed("Image[][]");

    // compute data filter
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("making %6.3f to %6.3f Hz %s data filter", Params.ImageHP, Params.ImageLP, FilterName);
        fflush(stdout);
    }
    if(Params.FilterType == IIR)
        mkfilter(&DataIIR, SIG_ORDER, Header.SampleRate, Params.ImageHP, Params.ImageLP, &BWFreq);
    else
        Butterworth(&DataFFT, TT, Params.ImageHP, Params.ImageLP, Header.SampleRate, BANDPASS, MAXORDER, &BWFreq);

    // compute smoothing filter
    if(vflg == TRUE) {
        printf(" - done\n");
        //printf("making %6.3f to %6.3f Hz %s smoothing filter", Params.SmoothHP, Params.SmoothLP, FilterName);
        printf("making 0. to %6.3f Hz %s smoothing filter", Params.SmoothLP, FilterName);
        fflush(stdout);
    }
    if(Params.FilterType == IIR)
        mkfilter(&SmoothIIR, SIG_ORDER, Header.SampleRate, 0., Params.SmoothLP, &BWFreq);
    else
        Butterworth(&SmoothFFT, TT, 0., Params.SmoothLP, Header.SampleRate, LOWPASS, MAXORDER, &BWFreq);

    // compute notch filter
    if(Params.Notch == TRUE) {
        if(Params.FilterType == FFT) {
            if(vflg == TRUE) {
                printf(" - done\n");
                printf("making notch filter");
                fflush(stdout);
            }
            mknotch(&NotchFFT, Params.CovHP, Params.CovLP, Header.SampleRate, TT, Params.Hz);
        } else {
            fprintf(stderr, "\nreminder: when using IIR filters, data should be notch filtered for power mains prior to analysis\n");
        }
    }

    // read & filter MEG data
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("reading & filtering MEG data");
        fflush(stdout);
    }
    for(m=0; m<M; m++) {                            // read & process data in channel order
        n = Header.PsIndex[m];
        scale = Channel[n].Scale;
        for(e=0; e<E; e++) {                        // multi-epoch data is presumed to be contiguous!
#if BTI
            if(GetMagnesChan(DSpath, PDFName, &Header, Channel, e, m, In) == -1)
                Cleanup("'GetMagnesChan()' failed");
#else
            GetDsData(&Header, e, n, scale, In);    // get one epoch of data for this channel -- length T samples
#endif
            for(t=0; t<T; t++) {
                tt = e * T + t;
                Out[tt] = In[t];
            }
        }
        demean(Out, TT);
        dd = In;
        if(Params.FilterType == IIR) {
            bdiir(Out, In, TT, &DataIIR);
        } else {
            FFTfilter(Out, In, &DataFFT, TT);
            if(Params.Notch == TRUE) {
                dd = Out;
                FFTfilter(In, Out, &NotchFFT, TT);
            }
        }
        for(t=0; t<TT; t++)
            x[t][m] = dd[t];

        // heartbeat
        if(vflg) {
            printf(".");
            fflush(stdout);
        }
    }

    // for each latency step, compute one volumetric SAM image...
    if(vflg) {
        printf(" - done\n");
        printf("computing SAM images for each time step:\n");
        fflush(stdout);
    }

    // for each voxel...
    for(c=0, pct=(-1); c<V; c++) {

        // test weight vector to determine if within head boundary
        for(m=0, tmp=0; m<M; m++)
            tmp += fabs(Wgt[c][m]);

        if(tmp > 0.) {      // compute image only for non-zero weights

            // compute virtual sensor time-series
            for(t=0; t<TT; t++) {
                for(m=0, tmp=0.; m<M; m++)
                    tmp += x[t][m] * Wgt[c][m];
                In[t] = tmp;        // In[] is the virtual sensor time-series
            }

            // compute instantaneous rank vector entropy
            Data2SymbolArray(In, Sym, TT, &TE, Params.Dims, Params.ImageLP, Header.SampleRate);
            Sym2RVE(Sym, Out, TE, Params.Dims);
            for(t=0; t<TE; t++)
                Out[t] *= Header.SampleRate;        // output is bits/second
            for(t=TE; t<TT; t++)
                Out[t] = 0.;

            // smooth instantaneous RVE time-series with a lowpass filter
            if(Params.FilterType == IIR)
                bdiir(Out, In, TT, &SmoothIIR);     // lowpass to smooth metric
            else
                FFTfilter(Out, In, &SmoothFFT, TT); // lowpass to smooth metric

            // if req'd, normalize by variance
            if(nflg) {
                for(t=0, tmp=0.; t<TE; t++)
                    tmp += In[t] * In[t];       // @@@ doesn't remove mean
                var = tmp / (double)(TE - 1);
                for(t=0; t<TE; t++)
                    In[t] /= var;
            }

            // downsample envelope of power
            for(d=0, time=1.; d<NumImg; d++, time+=Params.TimeStep) {
                t = (int)rint(time * Header.SampleRate);
                Image[d][c] = (float)In[t];
            }

        } else {
            for(d=0; d<NumImg; d++)
                Image[d][c] = 0.;
        }

        // percent done (counting completed SAM images)
        if(vflg) {
            percent = 100. * (double)c / (double)V;
            if(rint(percent) > (double)pct) {
                pct = (int)rint(percent);
                printf("\r\t%d percent done", pct);
                fflush(stdout);
            }
        }
    }

    // if req'd, remove voxel baselines
    switch(Params.RemoveBaseline) {
        case NONE:      // no baseline removal
            break;

        case BYVOXEL:   // remove baseline for each voxel

            // for each voxel...
            for(c=0; c<V; c++) {
                for(d=0, mean=0.; d<NumImg; d++)
                    mean += (double)Image[d][c];
                mean /= (double)NumImg;
                for(d=0; d<NumImg; d++)
                    Image[d][c] -= (float)mean;
            }
            break;

        case GLOBAL:    // remove global mean baseline from each voxel

            // find mean global entropy
            for(d=i=0, mean=0.; d<NumImg; d++)
                for(c=0; c<V; c++)
                    if(Image[d][c] > 0.) {  // non-zero voxels, only!
                        mean += (double)Image[d][c];
                        i++;                // count non-zero voxels!
                    }
            mean /= (double)(i * NumImg);

            // subtract mean
            for(d=0; d<NumImg; d++)
                for(c=0; c<V; c++)
                    if(Image[d][c] > 0.)
                        Image[d][c] -= (float)mean;
            break;


        default:
            break;
    }

    // write image
    if(vflg) {
        printf("\nwriting RVE");
        fflush(stdout);
    }

    // create NIFTI file name
    if(!strncmp(Params.DirName, "NULL", 4))
        sprintf(fpath, "%s/%s,%s,%-d-%-dHz,RVE.nii", SAMpath, Prefix, ParmName, (int)Params.ImageHP, (int)Params.ImageLP);
    else
        sprintf(fpath, "%s/%s,%s,%-d-%-dHz,RVE.nii", Params.DirName, Prefix, ParmName, (int)Params.ImageHP, (int)Params.ImageLP);
    if((fp = fopen(fpath, "w")) == NULL)
        Cleanup("can't open .nii file for write");

    // modify NIFTI header
    memset(extension, 0, 4);                                // for now, ignore the fact that we may have an extended header
    NiiHdr.dim[0] = 4;                                      // 3d+time
    NiiHdr.dim[4] = NumImg;                                 // number of 3d images
    NiiHdr.toffset = 0.;                                    // time offset shift
    NiiHdr.slice_duration = (float)Params.TimeStep;         // time step size
    NiiHdr.pixdim[4] = (float)Params.TimeStep;
    NiiHdr.slice_code = NIFTI_SLICE_SEQ_INC;
    NiiHdr.xyzt_units = (NIFTI_UNITS_MM | NIFTI_UNITS_SEC); // 3d+time units
    memset(NiiHdr.intent_name, 0, 16);
    memcpy(NiiHdr.intent_name, "RVE", 3);

    // write header & extension
    if(fwrite((void *)&NiiHdr, sizeof(nifti_1_header), 1, fp) != 1)
        Cleanup("can't write nifti header to .nii file");
    if(fwrite((void *)extension, 4, 1, fp) != 1)
        Cleanup("can't write 'extension' to .nii file");
    if((int)NiiHdr.vox_offset != 352 && ExtHdr != NULL) {
        N = (int)NiiHdr.vox_offset - 352;
        if(fwrite((void *)ExtHdr, N, 1, fp) != 1)
            Cleanup("can't write extended header to .nii file");
    }
    for(d=0; d<NumImg; d++)
        if(fwrite((void *)Image[d], sizeof(float), V, fp) != V)
            Cleanup("can't write data image file");
    fclose(fp);

    // that's all, folks!
    if(vflg == TRUE) {
        msg(" - done\n'%s' done\n", Progname);
    }
    exit(0);
}
