// sam_epi -- synthetic aperture magnetometry for imaging epiletic activity.
//  sam_epi reads SAM weights for volume image, & computes the source time-series
//  for each voxel.
//
//  Unlike prior versions, the excess kurtosis is computed for short overlapping
//  segments.
//
//  This algorithm is dedicated to the memory of Max Burbank. Max had
//  encouraged development of the original SAM(g2) algorithm as a means
//  of enabling automated studies at clinical labs everywhere in the
//  world.
//
//  Yuval Harpaz (BIU) suggested counting spikes. A compromise to this is to sum
//  g2 from each segment to the voxel. Unlike Yuval's method, I compute the median
//  of all non-zero segments so that the length of the dataset and number of
//  segments is not an issue for comparison of cases.
//
//  Version 2 introduces multi-processing via OMP directives.
//
//  flags:
//      -r <run name>           -- MEG run name
//      -d <data name>          -- MEG data file name (4D, only)
//      -m <name>               -- parameter file name
//      -v                      -- verbose mode
//
//  Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <getopt.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <geoms.h>
#include <samlib.h>
#include <siglib.h>
#include <DataFiles.h>
#include <SAMfiles.h>
#include <filters.h>
#include <nifti1.h>
#include <version.h>
#include "samutil.h"
#include "sam_parse.h"

#define MINOR_REV   6
#define SIG_ORDER   4

int main(
    int     argc,
    char    **argv
) {
    IIRSPEC         DataIIR;                // IIR data filter structure
    FFTSPEC         DataFFT;                // FFT data filter structure
    PARMINFO        Params;                 // analysis parameters
    PARM            *p;                     // pointer into active parameter list
    HeaderInfo      Header;                 // MEG data header
    ChannelInfo     *Channel;               // MEG channel info
    EpochInfo       *Epoch;                 // MEG epoch info
    nifti_1_header  NiiHdr;                 // NIFTI header
    unsigned char   *ExtHdr;                // NiiHdr[N] -- extended NIFTI header
    double          **x;                    // x[TT][M] -- multi-sensor signal
    double          **Wgt;                  // Wgt[V][M] -- SAM coefficients
    double          *In;                    // In[TT] -- input time-series
    double          *Out;                   // Out[TT] -- output time-series
    double          *p2;                    // p2[TT] -- sam^2
    double          *p4;                    // p4[TT] -- sam^4
    double          *Seg;                   // Seg[D] -- g2 value for each segment
    double          var;                    // 2nd moment (variance)
    double          sx2;                    // sum of 2nd power
    double          sx4;                    // sum of 4th power
    double          g2;                     // excess kurtosis
    double          BWFreq;                 // filter bandwidth
    double          tw;                     // total samples in Window
    double          scale;                  // bits-->T conversion
    double          percent;                // percent done
    register double tmp;                    // what else?
    float           *Image;                 // Image[V] -- SAM(g2) image
    int             D;                      // number of segments
    int             E;                      // number of good epochs
    int             M;                      // number of MEG or reference channels used
    int             N;                      // number of characters in extended NIFTI header
    int             T;                      // number of samples
    int             TW;                     // window segment size
    int             TT;                     // total sample size
    int             V;                      // number of voxels
    int             c;                      // input character & virtual channel
    int             dd;                     // number of segments used for g2
    int             pct;                    // integer percentage
    int             m;                      // channel index
    int             n;                      // alternate channel index
    int             d;                      // segment index
    int             e;                      // epoch index
    int             f;                      // 1st primary channel index
    int             t;                      // sample index
    int             ts;                     // starting sample for segment
    int             tt;                     // alternate time index
    int             NumChans;               // number of channels
    static int      eflg = FALSE;           // command-line error flag
    static int      rflg = FALSE;           // dataset name flag
    static int      mflg = FALSE;           // parameter file specification flag
    static int      vflg = FALSE;           // verbose mode flag
    char            fpath[256];             // general path name
    char            *DSName = NULL;         // dataset name
    char            DSpath[256];            // MEG path name
    char            SAMpath[256];           // SAM subdirectory path
    char            ImgPath[256];           // path for output NIFTI image
    char            *ParmName = NULL;       // name of marker parameter file
    char            Prefix[64];             // image file prefix
    char            FilterName[8];          // filter name -- IIR | FIR
    char            extension[4];           // extension
    FILE            *fp;                    // output file pointer (to SAM image)

    // system-dependent variables
#if BTI
    static int      dflg = FALSE;   // pdf file name flag
    char            *PDFName;       // name of data file
#else
    unsigned char   **Bad;          // Bad[E][T] -- flags for bad samples
#endif

    Progname = argv[0];
    reg_usage(MINOR_REV, __DATE__);

    // Register the parameters used by this program.

    reg_std_parm();

    reg_parm("NumMarkers");
    reg_parm("Marker");

    reg_parm("CovType");
    set_parm_arg_help("CovType", "GLOBAL");
    set_parm_help("CovType", "GLOBAL is always used");

    reg_parm("CovBand");
    reg_parm("ImageBand");
    reg_parm("FilterType");
    reg_parm("TimeInt");
    reg_parm("ImageDirectory");
    reg_parm("PrefixLength");

    reg_parm("ImageMetric");
    set_parm_arg_help("ImageMetric", "Kurtosis");
    set_parm_help("ImageMetric", "Kurtosis is always used");

    // Parse the arguments, environment variables, and parameter files.

    do_parse_args(argc, argv);

    // Extract specified parameters.

    eflg = get_params(&Params);

    p = get_parm("DataSet");
    if (p->set) {
        rflg = TRUE;
        DSName = Params.DataSetName;
    }
#if BTI
    p = get_parm("PDFName");
    if (p->set) {
        dflg = TRUE;
        PDFName = (char *)p->ptr;
    }
#endif
    p = get_parm("verbose");
    vflg = p->set;
    p = get_parm("param");
    mflg = p->set;

#if BTI
    if(!rflg || !dflg || !mflg || eflg) {
#else
    if(!rflg || !mflg || eflg) {
#endif
        msg("dataset (-r) and parameter file (-m) are required\n");
        do_help();  // doesn't return
    }

    ParmName = Params.ParmName;
    Params.CovType = GLOBAL;        // global is the only choice
    Params.ImageMetric = KURTOSIS;  // kurtosis is the only choice

    // get dataset information structures
    if(vflg == TRUE) {
        printf("S A M   S P I K E   I M A G I N G\t%s\n", PRG_REV);
        printf("opening data file");
        fflush(stdout);
    }

#if BTI
    // get data with sensor structures
    if(GetMagnesInfo(DSName, PDFName, &Header, &Channel, &Epoch) == -1)
        Cleanup("can't read Magnes information");
    sprintf(DSpath, "%s/%s", Header.DsPath, Header.SetName);
#else
    // get data with sensor structures
    GetDsInfo(DSName, &Header, &Channel, &Epoch, &Bad, TRUE);
    sprintf(DSpath, "%s/%s.ds", Header.DsPath, Header.SetName);
#endif
    sprintf(SAMpath, "%s/SAM", DSpath);

    // establish epoch, channel, & sample count
    E = Header.NumEpochs;
    T = Header.MaxSamples;
    TT = E * T;
    M = Header.NumPri;
    f = Header.PsIndex[0];      // 1st MEG primary sensor channel index

    // check to see if this is a continuous dataset
    if(E != 1)
        fprintf(stderr, " (reminder: multi-epoch data must be contiguous)");

    // check if this is the right program for you!
    if(Params.CovHP == -999. || Params.CovLP == -999.)
        Cleanup("CovBand parameters required");
    if(Params.ImageHP == -999. || Params.ImageLP == -999.)
        Cleanup("ImageBand parameters required");
    if(Params.ImageHP < Channel[f].AnalogHpFreq)
        Cleanup("ImageBand highpass cutoff frequency out of recorded range");
    if(Params.ImageLP > Channel[f].AnalogLpFreq)
        Cleanup("ImageBand lowpass cutoff frequency out of recorded range");
    if(Params.TimeInt == -999.)
        Cleanup("TimeInt specification missing from parameter file");
    if(Params.FilterType == IIR)
        strcpy(FilterName, "IIR");
    else
        strcpy(FilterName, "FFT");

    // set prefix to NumPrefix characters (default is 8-character dataset hashcode)
    memset(Prefix, 0, sizeof(Prefix));
#if BTI
    memcpy(Prefix, "BIOMAG4D", 8);  // 4D datasets just have serial numbers
#else
    memcpy(Prefix, DSName, Params.NumPrefix);
#endif

    // set ImgPath
    if(!strncmp(Params.DirName, "NULL", 4))
        sprintf(ImgPath, "%s/%s,%s", SAMpath, Prefix, ParmName);
    else
        sprintf(ImgPath, "%s/%s,%s", Params.DirName, Prefix, ParmName);

    // get weight name from parameters & read SAM weights
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("reading Global SAM weights");
        fflush(stdout);
    }
    sprintf(fpath, "%s/%s,%-d-%-dHz/Global.nii", SAMpath, ParmName, (int)Params.CovHP, (int)Params.CovLP);
    GetNIFTIWts(fpath, &Wgt, &NumChans, &V, &NiiHdr, extension, &ExtHdr);
    if(NumChans != M)
        Cleanup("number of channels in dataset does not match number in weight file");

    // set up imaging constants
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("determining imaging constants:\n");
        fflush(stdout);
    }
    TW = (int)floor(Params.TimeInt * Header.SampleRate);        // length of each g2 integration segment
    D = 2 * TT / TW - 1;                                        // number of overlapping integration segments
    if(vflg == TRUE) {
        printf("\tepochs=%d, samples/epoch=%d (total %d)\n", E, T, TT);
        printf("\tintegration window=%d samples\n", TW);
        printf("\tnumber of overlapping windows=%d\n", D);
        fflush(stdout);
    }

    // allocate memory for arrays
    if(vflg == TRUE) {
        printf("allocating memory");
        fflush(stdout);
    }
    if((x = (double **)malloc((size_t)TT * sizeof(double *))) == NULL)
        allocfailed("x[]");
    for(t=0; t<TT; t++)
        if((x[t] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
                allocfailed("x[][]");
    if((p2 = (double *)malloc((size_t)TT * sizeof(double))) == NULL)
        allocfailed("p2[]");
    if((p4 = (double *)malloc((size_t)TT * sizeof(double))) == NULL)
        allocfailed("p4[]");
    if((Seg = (double *)malloc((size_t)D * sizeof(double))) == NULL)
        allocfailed("Seg[]");
    if((In = (double *)malloc((size_t)TT * sizeof(double))) == NULL)
        allocfailed("In[]");
    if((Out = (double *)malloc((size_t)TT * sizeof(double))) == NULL)
        allocfailed("Out[]");
    if((Image = (float *)malloc((size_t)V * sizeof(float))) == NULL)
        allocfailed("Image[]");

    // generate Butterworth FFT filter
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("making %6.3f to %6.3f Hz %s filter for image bandpass", Params.ImageHP, Params.ImageLP, FilterName);
        fflush(stdout);
    }
    if(Params.FilterType == IIR)
        mkfilter(&DataIIR, SIG_ORDER, Header.SampleRate, Params.ImageHP, Params.ImageLP, &BWFreq);
    else
        Butterworth(&DataFFT, TT, Params.ImageHP, Params.ImageLP, Header.SampleRate, BANDPASS, MAXORDER, &BWFreq);

    // read & filter MEG data
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("reading & filtering MEG data");
        fflush(stdout);
    }
    for(m=0; m<M; m++) {                                        // read & process data in channel order
        n = Header.PsIndex[m];
        scale = Channel[n].Scale;
        for(e=0; e<E; e++) {                                    // multi-epoch data is presumed to be contiguous!
#if BTI
            if(GetMagnesChan(DSpath, PDFName, &Header, Channel, e, m, In) == -1)
                Cleanup("'GetMagnesChan()' failed");
#else
            GetDsData(&Header, e, n, scale, In);                // get one epoch of data for this channel -- length T samples
#endif
            for(t=0; t<T; t++) {
                tt = e * T + t;
                Out[tt] = In[t];
            }
        }
        demean(Out, TT);                                        // remove the mean
        if(Params.FilterType == IIR)
            bdiir(Out, In, TT, &DataIIR);
        else
            FFTfilter(Out, In, &DataFFT, TT);
        for(t=0; t<TT; t++)                                     // move to 'x[][]' array
            x[t][m] = In[t];

        // heartbeat
        if(vflg == TRUE) {
            printf(".");
            fflush(stdout);
        }
    }

    // now, step through all coordinates
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("stepping through voxels to determine g2:\n");
        fflush(stdout);
    }
    for(c=0, pct=(-1); c<V; c++) {
        for(m=0, tmp=0.; m<M; m++)
            tmp += Wgt[c][m] * Wgt[c][m];
        if(tmp > 0.) {

            // (1) compute the entire virtual sensor time-series, p2, & p4
#pragma omp parallel for shared(TT,M,Wgt,x,p2,p4) private(t,m,tmp)
            for(t=0; t<TT; t++) {
                for(m=0, tmp=0.; m<M; m++)
                    tmp += Wgt[c][m] * x[t][m];
                p2[t] = tmp * tmp;
                p4[t] = p2[t] * p2[t];
            }

            // (2) for each segment & compute sx2, sx4, & var
            for(d=dd=0; d<D; d++) {
                ts = d * TW / 2;
                for(t=0, tw=sx2=sx4=0.; t<TW; t++) {
                    tt = ts + t;
#if BTI
                    // no bad channel file in BTi datasets
                    sx2 += p2[tt];
                    sx4 += p4[tt];
                    tw += 1.;
#else
                    // CTF/VSM dataset may have a bad channel file
                    if(Bad[0][tt] == GOOD) {
                        sx2 += p2[tt];
                        sx4 += p4[tt];
                        tw += 1.;
                    }
#endif
                }

                // (3) compute var & g2
                if(tw > 12.) {
                    var = sx2 / (tw - 1.);
                    g2 = sx4 / (tw * var * var) - 3.;           // excess kurtosis of this segment
                    if(g2 > 0.) {                               // save non-zero g2 segments
                        Seg[dd] = g2;
                        dd++;
                    }
                }
            }
            Image[c] = (float)rms(Seg, dd);                     // rms value of all non-zero segments
        } else {
            Image[c] = 0.;
        }

        // percent done
        if(vflg == TRUE) {
            percent = 100. * (double)c / (double)V;
            if(rint(percent) > (double)pct) {
                pct = (int)rint(percent);
                printf("\r\t%d percent done", pct);
                fflush(stdout);
            }
        }
    }

    // write out SAM(g2) image
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("writing SAM(Epi) image");
        fflush(stdout);
    }
    sprintf(fpath, "%s,Epi.nii", ImgPath);
    if((fp = fopen(fpath, "w")) == NULL)
        Cleanup("can't open .nii file for write");

    // modify NIFTI header
    memset(extension, 0, 4);                                    // for now, ignore the fact that we may have an extended header
    NiiHdr.dim[0] = 3;                                          // 3d volume
    NiiHdr.dim[4] = 0;                                          // number of 3d images
    NiiHdr.toffset = 0.;                                        // time offset shift
    NiiHdr.slice_duration = 1.;
    NiiHdr.slice_code = NIFTI_SLICE_UNKNOWN;
    NiiHdr.xyzt_units = NIFTI_UNITS_MM;                         // units
    memset(NiiHdr.intent_name, 0, 16);
    memcpy(NiiHdr.intent_name, "Epilepsy Spike", 14);

    // write header & extension
    if(fwrite((void *)&NiiHdr, sizeof(nifti_1_header), 1, fp) != 1)
        Cleanup("can't write nifti header to .nii file");
    if(fwrite((void *)extension, 4, 1, fp) != 1)
        Cleanup("can't write 'extension' to .nii file");
    if((int)NiiHdr.vox_offset != 352 && ExtHdr != NULL) {
        N = (int)NiiHdr.vox_offset - 352;
        if(fwrite((void *)ExtHdr, N, 1, fp) != 1)
            Cleanup("can't write extended header to .nii file");
    }
    if(fwrite((void *)Image, sizeof(float), V, fp) != V)
        Cleanup("can't write voxel values to .nii file");
    fclose(fp);

    // that's all, folks!
    if(vflg == TRUE) {
        msg(" - done\n'%s' done\n", Progname);
    }
    exit(0);
}
