// sam_ers -- synthetic aperture magnetometry imaging for source event related
//      signals. sam_ers reads SAM weights for volume image, & computes the source
//      averagetime-series for each voxel. After subtraction of the baselines,
//      the ratio of the absolute value of the area under the curve for the
//      average is mapped as a 3d+time image (.nii) file
//
//      flags:
//      -r <dataset name>           -- MEG run name
//      -d <data file name>         -- MEG data file name (4D, only)
//      -m <parameter file name>    -- name of trigger marker to average
//      -z <snr threshold>          -- threshold for SNR of virtual sensor
//      -v                          -- verbose mode
//
//      Author: Stephen E. Robinson
//              MEG Core Facility
//              NIMH
//


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <float.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <geoms.h>
#include <samlib.h>
#include <siglib.h>
#include <DataFiles.h>
#include <SAMfiles.h>
#include <filters.h>
#include <nifti1.h>
#include <version.h>
#include "samutil.h"
#include "sam_parse.h"

#define MINOR_REV       1
#define STATE_START     0
#define STATE_END       1
#define SIG_ORDER       4
#define MAX_MARKS       32768

typedef struct {                    // structure for listing signal averaging trials
    int     Epoch;                  // epoch number
    int     TS;                     // average starting sample
    int     TE;                     // average ending sample
    int     Marker;                 // either Marker1 (or, optionally, Marker2) -- value is 1 or 2
} AVG_TRIAL;

int main(
    int     argc,
    char    **argv
) {
    IIRSPEC         IIRFilter;      // IIR filter structure
    FFTSPEC         FFTFilter;      // FFT filter gain coefficients
    PARMINFO        Params;         // analysis parameters
    PARM            *p;             // pointer into active parameter list
    AVG_TRIAL       *Trial;         // Trial[N]
    SAM_MARKS       *Marker;        // Marker[N] -- array of time-ordered markers
    HeaderInfo      Header;         // MEG data header
    nifti_1_header  NiiHdr;         // NIFTI header
    unsigned char   *ExtHdr;        // NiiHdr[N] -- extended NIFTI header
    ChannelInfo     *Channel;       // MEG channel info
    EpochInfo       *Epoch;         // MEG epoch info
    double          **x;            // x[M][T] -- single-epoch of multisensor data
    double          **xi;           // xi[M][T] -- imaginary part after Hilbert transform
    double          **avg1;         // avg1[T][M] -- signal average of all channels for each sample of marker1
    double          **avg1i;        // avg1i[T][M] -- imaginary part
    double          **avg2;         // avg2[T][M] -- signal average of all channels for each sample of marker2
    double          **avg2i;        // avg2i[T][M] -- imaginary part
    double          **Wgt;          // Wgt[V][M] -- SAM weights
    double          *SAM_avg1;      // SAM_avg1[T] -- SAM average time-series for marker1
    double          *SAM_avg2;      // SAM_avg2[T] -- SAM average time-series for marker2
    double          *In;            // In[T] -- input time-series
    double          *Out;           // Out[T] -- output time-series
    double          baseline;       // baseline for SAM average
    double          rms_noise;      // rms noise of primary sensors projected through SAM weights
    double          BWFreq;         // effective bandwidth
    double          sigma2;         // mean sensor noise power (T^2)
    double          ta;             // number of samples in average (float)
    double          tb;             // number of samples in baseline (float)
    double          tr;             // number of samples in response window (float)
    double          sign;           // polarity of response window
    double          percent;        // percent done
    double          tmp;            // what else?
    double          tmp2;           // for imaginary part
    float           **Image;        // Image[D][V] -- image voxels for each latency
    int             M;              // number of MEG or reference channels used
    int             N;              // number of characters in extended NIFTI header
    int             T;              // number of samples
    int             TA;             // number of samples in average
    int             AS;             // average starting sample
    int             AE;             // average end sample
    int             TB;             // number of samples in baseline
    int             TR;             // number of samples in response window
    int             BS;             // baseline starting sample
    int             BE;             // baseline ending sample
    int             RS;             // response starting sample
    int             RE;             // response ending sample
    int             PS;             // polarity starting time
    int             PE;             // polarity ending time
    int             US;             // area under the curve starting time for response
    int             UE;             // area under the curve ending time for response
    int             V;              // number of voxels
    int             c;              // input character & virtual channel
    int             pct;            // integer percentage
    int             m;              // channel index
    int             n;              // trigger marker index
    int             d;              // image latency index
    int             e;              // epoch index
    int             f;              // 1st primary channel index
    int             t;              // sample index
    int             tt;             // alternate sample index
    int             w;              // boxcar averager index
    int             a;              // actual boxcar count
    int             k;              // k-index
    int             sample;         // sample index
    int             NumMarkers;     // number of markers
    int             NumTriggers;    // number of triggers matching value specified in command line
    int             TotTrials;      // total number of trials for marker 1 & optionally marker 2
    int             NumTrials1;     // number of trials in average 1
    int             NumTrials2;     // number of trials in average 2
    int             NumChans;       // number of channels in weight file
    int             NumStep;        // number samples per image
    int             NumImg;         // number of images
    int             boxwidth;       // boxcar averager width (samples)
    TIMERANGE       *outrange;      // output only times in this range
    static int      aflg = FALSE;   // take absolute value of voxels
    static int      bflg = FALSE;   // baseline removal flag
    static int      eflg = FALSE;   // command-line error flag
    static int      mflg = FALSE;   // parameter file name flag
    static int      pflg = FALSE;   // polarity correction flag
    static int      rflg = FALSE;   // dataset name flag
    static int      vflg = FALSE;   // verbose mode flag
    static int      uflg = FALSE;   // area under the curve flag
    char            fpath[256];     // general path name
    char            *DSName = NULL; // dataset name
    char            DSpath[256];    // MEG dataset path
    char            SAMpath[256];   // SAM subdirectory path
    char            ImgPath[256];   // image path
    char            Prefix[64];     // output file prefix
    char            Suffix[10];     // output file suffix
    char            *ParmName = NULL; // name of parameter file
    char            WgtDir[256];    // weight file prefix
    char            extension[4];   // extension
#if BTI
    static int      dflg = FALSE;   // pdf file name flag
    char            *PDFName;       // name of data file
#else
    double          scale;          // bits-->Tesla conversion
    unsigned char   **Bad;          // Bad[E][T] -- flags for bad samples   @@@ ??
#endif
    FILE            *fp;            // output file pointer

    Progname = argv[0];
    reg_usage(MINOR_REV, __DATE__);

    // Register the parameters used by this program.

    reg_std_parm();

    reg_parm("NumMarkers");
    reg_parm("Marker");

    reg_parm("CovType");
    set_parm_arg_help("CovType", "GLOBAL|SUM");
    reg_parm("CovBand");
    reg_parm("ImageBand");
    reg_parm("FilterType");

    reg_parm("DataSegment");
    reg_parm("SignSegment");
    reg_parm("Absolute");
    reg_parm("Baseline");
    reg_parm("TimeStep");
    reg_parm("BoxWidth");

    reg_parm("ImageMetric");
    set_parm_arg_help("ImageMetric", "Signal|Hilbert");
    reg_parm("ImageDirectory");
    reg_parm("PrefixLength");
    reg_parm("OutName");
    reg_parm("ImageFormat");

    reg_parm("AUC");
    reg_parm("OutRange");

    // Parse the arguments, environment variables, and parameter files.

    do_parse_args(argc, argv);

    // Extract specified parameters.

    eflg = get_params(&Params);

    p = get_parm("DataSet");
    if (p->set) {
        rflg = TRUE;
        DSName = Params.DataSetName;
    }
#if BTI
    p = get_parm("PDFName");
    if (p->set) {
        dflg = TRUE;
        PDFName = (char *)p->ptr;
    }
#endif
    p = get_parm("verbose");
    vflg = p->set;
    p = get_parm("param");
    mflg = p->set;
    p = get_parm("AUC");            // -u, area under curve
    uflg = p->set;
    p = get_parm("Absolute");       // -a
    aflg = p->set;

#if BTI
    if (!rflg || !dflg || !mflg || eflg) {
#else
    if (!rflg || !mflg || eflg) {
#endif
        msg("dataset (-r) and parameter file (-m) are required\n");
        do_help();  // doesn't return
    }

    ParmName = Params.ParmName;

    // get dataset information structures
    if (vflg) {
        printf("S A M   E V E N T - R E L A T E D   S O U R C E   I M A G I N G\t%s\n", PRG_REV);
        printf("opening data file");
        fflush(stdout);
    }

#if BTI
    // get data with sensor structures
    if (GetMagnesInfo(DSName, PDFName, &Header, &Channel, &Epoch) == -1)
        Cleanup("can't read Magnes information");
    sprintf(DSpath, "%s/%s", Header.DsPath, Header.SetName);
#else
    // get data with sensor structures
    GetDsInfo(DSName, &Header, &Channel, &Epoch, &Bad, TRUE);
    sprintf(DSpath, "%s/%s.ds", Header.DsPath, Header.SetName);
#endif
    sprintf(SAMpath, "%s/SAM", DSpath);

    // extract constants
    M = Header.NumPri;
    T = Header.MaxSamples;
    f = Header.PsIndex[0];      // 1st MEG primary sensor channel index

    if (vflg) {
        printf(" - done\n");
        if (Params.ImageMetric == HILBERT)
            printf("imaging Hilbert envelope");
        else
            printf("imaging signal");
    }
    if (Params.SignSegStart != -999 && Params.SignSegEnd != -999.) {
        if (vflg) {
            printf(" with polarity correction\n");
        }
        pflg = TRUE;
    } else {
        if (vflg) {
            printf("\n");
        }
    }

    // sanity checks
    if (Params.DataSegStart == -999. && Params.DataSegEnd == -999.)
        Cleanup("DataSegment req'd in parameter file");
    if (Params.BaseStart != -999. && Params.BaseEnd != -999.)
        bflg = TRUE;
    if (Params.CovHP == -999. || Params.CovLP == -999.)
        Cleanup("CovBand parameters req'd");
    if (Params.ImageHP == -999. || Params.ImageLP == -999.)
        Cleanup("ImageBand parameters req'd");
    if (Params.ImageHP < Channel[f].AnalogHpFreq)
        Cleanup("ImageBand highpass cutoff frequency out of recorded range");
    if (Params.ImageLP > Channel[f].AnalogLpFreq)
        Cleanup("ImageBand lowpass cutoff frequency out of recorded range");
    switch(Params.NumMark) {
        case 1:
            if (!strncmp("NULL", Params.Marker[0].MarkName, 4))
                Cleanup("Marker1 name req'd");
            break;
        case 2:
            if (!strncmp("NULL", Params.Marker[0].MarkName, 4) || !strncmp("NULL", Params.Marker[1].MarkName, 4))
                Cleanup("both Marker1 & Marker2 names req'd");
            break;
        default:
            break;
    }
    if (Params.TimeStep < 0.)
        Cleanup("TimeStep must be specified");

    // set prefix to NumPrefix characters (default is 8-character dataset hashcode)
    memset(Prefix, 0, sizeof(Prefix));
#if BTI
    memcpy(Prefix, "BIOMAG4D", 8);              // 4D datasets just have serial numbers
#else
    memcpy(Prefix, DSName, Params.NumPrefix);   // default: copy the eight-letter NIMH hash code in dataset name
#endif
    memset(Suffix, 0, sizeof(Suffix));
    switch(Params.ImageMetric) {
        case MOMENT: memcpy(Suffix, "MOM", 3); break;
//        case POWER: memcpy(Suffix, "PWR", 3); break;
        case HILBERT: memcpy(Suffix, "ENV", 3); break;
        default: Cleanup("'ImageMetric' must be Signal or Hilbert");
    }

    // read markers
#if BTI                                         // parse trigger channel for markers
    if (vflg) {
        printf("parsing trigger markers");
    }
    if (GetMagnesMarkers(DSpath, PDFName, &Header, Channel, &Marker, &NumMarkers) == -1)
        Cleanup("cannot get trigger markers");
#else                                           // try to read 'MarkerFile.mrk'
    if (vflg) {
        printf("reading 'MarkerFile.mrk'");
    }
    sprintf(fpath, "%s/MarkerFile.mrk", DSpath);
    GetMarkers(fpath, &Marker, &NumMarkers);
#endif

    // summarize the number of epochs matching the marker name
    for (n=NumTriggers=0; n<NumMarkers; n++) {
        if (!strcmp(Marker[n].Name, Params.Marker[0].MarkName))
            NumTriggers++;
        if (Params.NumMark == 2)
            if (!strcmp(Marker[n].Name, Params.Marker[1].MarkName))
                NumTriggers++;
    }
    if (NumTriggers == 0)
        Cleanup("specified trigger marker(s) not found -- use valid trigger marker name");

    // set average & baseline window size
    if (vflg) {
        printf(" - done\n");
        printf("%d markers matched the designated trigger marker(s)\n", NumTriggers);
        printf("parsing time windows");
    }
    if (bflg) {
        BS = (int)floor(Params.BaseStart * Header.SampleRate);  // baseline starting sample relative to trigger marker (can be negative!)
        BE = (int)floor(Params.BaseEnd * Header.SampleRate);    // baseline ending sample relative to trigger marker (can be negative!)
        TB = BE - BS + 1;
        tb = (double)TB;
    } else {
        BS = 0;
        BE = 0;
        TB = 0;
        tb = 0.;
    }
    RS = (int)floor(Params.DataSegStart * Header.SampleRate);   // response starting sample relative to trigger marker (can be negative!)
    RE = (int)floor(Params.DataSegEnd * Header.SampleRate);     // response ending sample relative to trigger marker (can be negative!)
    TR = RE - RS + 1;
    tr = (double)TR;
    AS = (BS < RS)? BS: RS;                                     // average starts at the earliest sample - either baseline or response window
    AE = (RE > BE)? RE: BE;                                     //
    TA = AE - AS + 1;                                           // length of average, in samples
    ta = (double)TA;
    Params.DataSegStart = (double)AS / Header.SampleRate;       // save for later
    NumStep = (int)floor(Params.TimeStep * Header.SampleRate);  // number of samples per time step
    Params.TimeStep = NumStep / Header.SampleRate;
    if (NumStep == 0)
        NumStep = 1;                                            // at least 1 sample!!!
    NumImg = TA / NumStep;
    if (uflg) {
        NumImg = 1;
    }
    if (pflg) {                                                 // optionally, determine time window for polarity
        PS = (int)floor(Params.SignSegStart * Header.SampleRate) - AS;
        PE = (int)floor(Params.SignSegEnd * Header.SampleRate) - AS;
    } else {
        PS = 0;
        PE = 0;
    }
    if (uflg) {
        US = (int)floor(Params.Marker[0].MarkStart * Header.SampleRate) - AS;
        UE = (int)floor(Params.Marker[0].MarkEnd * Header.SampleRate) - AS;
    } else {
        US = 0;
        UE = 0;
    }
    // read trigger markers & parse
    if (vflg) {
        printf(" - done\n");
        printf("New TimeStep (integer number of samples): %f\n", Params.TimeStep);
        printf("parsing trials");
        fflush(stdout);
    }
    if ((Trial = (AVG_TRIAL *)malloc((size_t)(NumTriggers * sizeof(AVG_TRIAL)))) == NULL)
        Cleanup("'Trial[N]' allocation failed");
    for (n=k=NumTrials1=NumTrials2=0; n<NumMarkers; n++) {
        if (!strcmp(Marker[n].Name, Params.Marker[0].MarkName)) {
            sample = (int)(Marker[n].Latency * Header.SampleRate);  // sample for this trigger marker
            Trial[k].Epoch = Marker[n].Epoch;
            Trial[k].TS = sample + AS;                      // offset trial start time relative to sample marker
            Trial[k].TE = sample + AE;                      // offset trial end time relative to sample marker
            if (Trial[k].TS >= 0 || Trial[k].TE < T) {      // increment if this trial is within the epoch samples
                Trial[k].Marker = 1;                        // 1st marker used for both ERS & DERS modes
                NumTrials1++;
                k++;
            }
        }
        if (Params.NumMark == 2) {
            if (!strcmp(Marker[n].Name, Params.Marker[1].MarkName)) {   // for DERS, read control marker name
                sample = (int)(Marker[n].Latency * Header.SampleRate);  // sample for this trigger marker
                Trial[k].Epoch = Marker[n].Epoch;
                Trial[k].TS = sample + AS;                  // offset trial start time relative to sample marker
                Trial[k].TE = sample + AE;                  // offset trial end time relative to sample marker
                if (Trial[k].TS >= 0 || Trial[k].TE < T) {  // increment if this trial is within the epoch samples
                    Trial[k].Marker = 2;                    // 1st marker used for DERS modes
                    NumTrials2++;
                    k++;
                }
            }
        }
    }
    TotTrials = k;
    if (TotTrials < 4)
        Cleanup("insufficient trials - check trigger & time windows");
    if (vflg) {
        printf(" - done\n");
        printf("found %d trials for Marker1", NumTrials1);
        if (Params.NumMark == 2)
            printf(" & %d trials for Marker2", NumTrials2);
        printf("\n");
    }

    // revise start & end times relative to start of average so that we can deal with averaged data...
    BS -= AS;
    BE -= AS;
    RS -= AS;
    RE -= AS;
    AE -= AS;
    AS = 0;                         // average window now starts at sample zero!
    if (vflg) {
        printf("average: sample %d to %d (%d samples, %5.3f sec)\n", AS, AE, TA, ta/Header.SampleRate);
        printf("response: sample %d to %d (%d samples, %5.3f sec)\n", RS, RE, TR, tr/Header.SampleRate);
        if (bflg) {
            printf("baseline: sample %d to %d (%d samples, %5.3f sec)\n", BS, BE, TB, tb/Header.SampleRate);
        }
        if (uflg) {
            printf("area under curve interval: sample %d to %d\n", US, UE);
        }
        if (pflg) {
            printf("sign: sample %d to %d\n", PS, PE);
        }
    }

    // weight name must be extend using the params...
    if (vflg) {
        printf("reading ");
        fflush(stdout);
    }
    sprintf(WgtDir, "%s/%s,%-d-%-dHz", SAMpath, ParmName, (int)Params.CovHP, (int)Params.CovLP);
    sigma2 = 0.;
    switch(Params.CovType) {
        case GLOBAL_:       // read Global weights & Global_Noise
            if (vflg) {
                printf("'Global'");
                fflush(stdout);
            }
            if (Params.ImageFormat == TLRC)
                sprintf(fpath, "%s/Global_at.nii", WgtDir);
            else
                sprintf(fpath, "%s/Global.nii", WgtDir);
            GetNIFTIWts(fpath, &Wgt, &NumChans, &V, &NiiHdr, extension, &ExtHdr);
            if (NumChans != M)
                Cleanup("number of channels in dataset does not match number in weight file");
            sigma2 = (double)NiiHdr.noise;
            break;
        case SUM_:      // 'Sum' weights
            if (vflg) {
                printf("'Sum'");
                fflush(stdout);
            }
            if (Params.ImageFormat == TLRC)
                sprintf(fpath, "%s/Sum_at.nii", WgtDir);
            else
                sprintf(fpath, "%s/Sum.nii", WgtDir);
            GetNIFTIWts(fpath, &Wgt, &NumChans, &V, &NiiHdr, extension, &ExtHdr);
            if (NumChans != M)
                Cleanup("number of channels in dataset does not match number in weight file");
            sigma2 = (NiiHdr.noise == 0.)? 2.5e-27: 1.0e-28 * (double)NiiHdr.noise;
            break;
        default:
            Cleanup("CovType must be either GLOBAL or SUM");
            break;
    }
    if (vflg) {
        printf(" SAM weights");
        fflush(stdout);
    }

    // set up boxcar averager for smoothing output
    boxwidth = 2 * (int)(Header.SampleRate / Params.ImageLP);
    p = get_parm("BoxWidth");
    if (p->set) {
        n = *(int *)p->ptr;
        if (n < 0 || n >= T) {
            fatalerr("improper box width %d", n);
        }
        n /= 2;             // make sure it's even
        boxwidth = n * 2;
    }

    if (vflg) {
        printf(" - done\n");
        if (boxwidth > 0) {
            printf("using %d sample length boxcar filter for smoothing output", boxwidth);
        } else {
            printf("no smoothing applied");
        }
        fflush(stdout);
    }

    // allocate memory for arrays
    if (vflg) {
        printf(" - done\n");
        printf("allocating memory");
        fflush(stdout);
    }
    if ((x = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
        allocfailed("x[]");
    for(m=0; m<M; m++)
        if ((x[m] = (double *)malloc((size_t)T * sizeof(double))) == NULL)
            allocfailed("x[][]");
    if ((xi = (double **)malloc((size_t)M * sizeof(double *))) == NULL)
        allocfailed("xi[]");
    for(m=0; m<M; m++)
        if ((xi[m] = (double *)malloc((size_t)T * sizeof(double))) == NULL)
            allocfailed("xi[][]");
    if ((In = (double *)malloc((size_t)T * sizeof(double))) == NULL)
            allocfailed("In[]");
    if ((Out = (double *)malloc((size_t)T * sizeof(double))) == NULL)
            allocfailed("Out[]");
    if ((avg1 = (double **)malloc((size_t)TA * sizeof(double *))) == NULL)
        allocfailed("avg1[]");
    for(t=0; t<TA; t++)
        if ((avg1[t] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
            allocfailed("'avg1[][]");
    if ((avg1i = (double **)malloc((size_t)TA * sizeof(double *))) == NULL)
        allocfailed("avg1i[]");
    for(t=0; t<TA; t++)
        if ((avg1i[t] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
            allocfailed("'avg1i[][]");
    if ((avg2 = (double **)malloc((size_t)TA * sizeof(double *))) == NULL)
        allocfailed("avg2[]");
    for(t=0; t<TA; t++)
        if ((avg2[t] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
            allocfailed("avg2[][]");
    if ((avg2i = (double **)malloc((size_t)TA * sizeof(double *))) == NULL)
        allocfailed("avg2i[]");
    for(t=0; t<TA; t++)
        if ((avg2i[t] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
            allocfailed("avg2i[][]");
    if ((SAM_avg1 = (double *)malloc((size_t)TA * sizeof(double))) == NULL)
            allocfailed("SAM_avg1[]");
    if ((SAM_avg2 = (double *)malloc((size_t)TA * sizeof(double))) == NULL)
            allocfailed("SAM_avg2[]");
    if ((Image = (float **)malloc((size_t)NumImg * sizeof(float *))) == NULL)
        allocfailed("Image[]");
    for(d=0; d<NumImg; d++)
        if ((Image[d] = (float *)malloc((size_t)V * sizeof(float))) == NULL)
            allocfailed("Image[][]");

    // determine filter type, & determine band-limited noise
    if (vflg) {
        printf(" - done\n");
        printf("making %6.3f to %6.3f Hz filter", Params.ImageHP, Params.ImageLP);
        fflush(stdout);
    }
    if (Params.FilterType == IIR) {
        mkfilter(&IIRFilter, SIG_ORDER, Header.SampleRate, Params.ImageHP, Params.ImageLP, &BWFreq);
    } else {
        Butterworth(&FFTFilter, T, Params.ImageHP, Params.ImageLP, Header.SampleRate, BANDPASS, MAXORDER, &BWFreq);
    }

    // zero average buffers
    for(t=0; t<TA; t++)
        for(m=0; m<M; m++) {
            avg1[t][m] = 0.;
            avg2[t][m] = 0.;
            avg1i[t][m] = 0.;
            avg2i[t][m] = 0.;
        }

    // read in data, & apply bandpass filter
    if (vflg) {
        printf(" - done\n");
        printf("reading data & accumulating sums:\n");
        fflush(stdout);
    }
    for(k=0, e=-1; k<TotTrials; k++) {      // set 'e' negative so first epoch called will be read & filtered

        // if req'd read in data from a new epoch, & filter (see? I told ya so...)
        if (e != Trial[k].Epoch) {
            e = Trial[k].Epoch;
            if (vflg) {
                printf("epoch %d:\n", e);
                fflush(stdout);
            }
            if (vflg) {
                printf("\treading & filtering");
                fflush(stdout);
            }
            for(m=0; m<M; m++) {
#if BTI
                if (GetMagnesChan(DSName, PDFName, &Header, Channel, e, m, In) == -1)
                    Cleanup("'GetMagnesChan()' failed");
#else
                n = Header.PsIndex[m];
                scale = Channel[n].Scale;
                GetDsData(&Header, e, n, scale, Out);
#endif
                demean(Out, T);
                if (Params.FilterType == IIR) {
                    bdiir(Out, In, T, &IIRFilter);
                } else {
                    FFTfilter(Out, In, &FFTFilter, T);
                }

                // if req'd, Hilbert transform
                if (Params.ImageMetric == HILBERT) {
                    hilbert(T, In, Out);    // Out[] is the phase-shifted VS time-series
                    for(t=0; t<T; t++) {
                        x[m][t] = In[t];
                        xi[m][t] = Out[t];
                    }
                } else {
                    for(t=0; t<T; t++)
                        x[m][t] = In[t];
                }
            }

            // end epoch read
            if (vflg) {
                printf(".");
                fflush(stdout);
            }
        }

        // update average(s)
        if (Params.ImageMetric == HILBERT) {
            switch(Trial[k].Marker) {
                case 1:
                    for(t=0, tt=Trial[k].TS; t<TA; t++, tt++)
                        for(m=0; m<M; m++) {
                            avg1[t][m] += x[m][tt];
                            avg1i[t][m] += xi[m][tt];
                        }
                    break;
                case 2:
                    for(t=0, tt=Trial[k].TS; t<TA; t++, tt++)
                        for(m=0; m<M; m++) {
                            avg2[t][m] += x[m][tt];
                            avg2i[t][m] += xi[m][tt];
                        }
                    break;
                default:
                    break;
            }
        } else {
            switch(Trial[k].Marker) {
                case 1:
                    for(t=0, tt=Trial[k].TS; t<TA; t++, tt++)
                        for(m=0; m<M; m++)
                            avg1[t][m] += x[m][tt];
                    break;
                case 2:
                    for(t=0, tt=Trial[k].TS; t<TA; t++, tt++)
                        for(m=0; m<M; m++)
                            avg2[t][m] += x[m][tt];
                    break;
                default:
                    break;
            }
        }
    }               // end segment loop

    // compute average
    if (vflg) {
        printf(" - done\n");
        printf("computing averages");
        fflush(stdout);
    }
    for(m=0; m<M; m++)
        for(t=0; t<TA; t++) {
            avg1[t][m] /= (double)NumTrials1;
            if (Params.NumMark == 2) {
                avg2[t][m] /= (double)NumTrials2;
            }
            if (Params.ImageMetric == HILBERT) {
                avg1i[t][m] /= (double)NumTrials1;
                if (Params.NumMark == 2) {
                    avg2i[t][m] /= (double)NumTrials2;
                }
            }
        }

    // loop on all coordinates in the ROI
    if (vflg) {
        printf(" - done\n");
        printf("computing image voxels:\n");
        fflush(stdout);
    }
    for(c=0, pct=(-1); c<V; c++) {

        // read weight vector & determine if within head boundary
        for(m=0, tmp=0; m<M; m++)
            tmp += fabs(Wgt[c][m]);
        if (tmp > 0.) {

            // compute band-limited noise of this virtual sensor
            for(m=0, tmp=0.; m<M; m++)                      // compute noise of unaveraged primary sensors
                tmp += Wgt[c][m] * sigma2 * Wgt[c][m];
            rms_noise = sqrt(tmp);                          // rms noise estimate

            // compute SAM time-series for average & remove baseline
            switch(Params.NumMark) {

                case 1: // compute average for marker1

                    // compute SAM projection of avg1
                    if (Params.ImageMetric == HILBERT) {
                        for(t=0; t<TA; t++) {
                            for(m=0, tmp=0., tmp2=0.; m<M; m++) {
                                tmp += avg1[t][m] * Wgt[c][m];
                                tmp2 += avg1i[t][m] * Wgt[c][m];
                            }
                            // calculate the envelope of the average
                            tmp = sqrt(tmp * tmp + tmp2 * tmp2);
                            In[t] = tmp / rms_noise;
                        }
                    } else {
                        // compute SAM projection of avg1
                        for(t=0; t<TA; t++) {
                            for(m=0, tmp=0.; m<M; m++)
                                tmp += avg1[t][m] * Wgt[c][m];
                            In[t] = tmp / rms_noise;
                        }
                    }

                    // apply smoothing to avg1
                    for(t=0; t<TA; t++) {
                        if (boxwidth > 0) {
                            for(w=(-boxwidth/2), tmp=0., a=0; w<=boxwidth/2; w++) {
                                tt = t + w;
                                if (tt >=0 && tt < (TA-1)) {
                                    tmp += In[tt];
                                    a++;
                                }
                            }
                            Out[t] = tmp / (double)a;
                        } else {
                            Out[t] = In[t];
                        }
                    }

                    // if req'd, compute mean baseline of avg1
                    if (bflg) {
                        for(t=BS, baseline=0.; t<=BE; t++)
                            baseline += Out[t];
                        baseline /= tb;
                    } else {
                        baseline = 0.;
                    }

                    // subtract mean baseline from avg1
                    for(t=0; t<TA; t++)
                        SAM_avg1[t] = Out[t] - baseline;

                    break;

                case 2: // compute average for markers1 & 2 & remove baseline

                    // compute SAM projection of avg1
                    if (Params.ImageMetric == HILBERT) {
                        for(t=0; t<TA; t++) {
                            for(m=0, tmp=0., tmp2=0.; m<M; m++) {
                                tmp += avg1[t][m] * Wgt[c][m];
                                tmp2 += avg1i[t][m] * Wgt[c][m];
                            }
                            // calculate the envelope of the average
                            tmp = sqrt(tmp * tmp + tmp2 * tmp2);
                            In[t] = tmp / rms_noise;
                        }
                    } else {
                        for(t=0; t<TA; t++) {
                            for(m=0, tmp=0.; m<M; m++)
                                tmp += avg1[t][m] * Wgt[c][m];
                            In[t] = tmp / rms_noise;
                        }
                    }

                    // apply smoothing to avg1
                    for(t=0; t<TA; t++) {
                        if (boxwidth > 0) {
                            for(w=(-boxwidth/2), tmp=0., a=0; w<=boxwidth/2; w++) {
                                tt = t + w;
                                if(tt >=0 && tt < (TA-1)) {
                                    tmp += In[tt];
                                    a++;
                                }
                            }
                            Out[t] = tmp / (double)a;
                        } else {
                            Out[t] = In[t];
                        }
                    }

                    // if req'd, compute mean baseline of avg1
                    if (bflg) {
                        for(t=BS, baseline=0.; t<=BE; t++)
                            baseline += Out[t];
                        baseline /= tb;
                    } else {
                        baseline = 0.;
                    }

                    // subtract mean baseline from avg1
                    for(t=0; t<TA; t++)
                        SAM_avg1[t] = Out[t] - baseline;

                    // compute SAM projection of avg2
                    if (Params.ImageMetric == HILBERT) {
                        for(t=0; t<TA; t++) {
                            for(m=0, tmp=0., tmp2=0.; m<M; m++) {
                                tmp += avg2[t][m] * Wgt[c][m];
                                tmp2 += avg2i[t][m] * Wgt[c][m];
                            }
                            // calculate the envelope of the average
                            tmp = sqrt(tmp * tmp + tmp2 * tmp2);
                            In[t] = tmp / rms_noise;
                        }
                    } else {
                        for(t=0; t<TA; t++) {
                            for(m=0, tmp=0.; m<M; m++)
                                tmp += avg2[t][m] * Wgt[c][m];
                            In[t] = tmp / rms_noise;
                        }
                    }

                    // apply smoothing to avg2
                    for(t=0; t<TA; t++) {
                        if (boxwidth > 0) {
                            for(w=(-boxwidth/2), tmp=0., a=0; w<=boxwidth/2; w++) {
                                tt = t + w;
                                if (tt >=0 && tt < (TA-1)) {
                                    tmp += In[tt];
                                    a++;
                                }
                            }
                            Out[t] = tmp / (double)a;
                        } else {
                            Out[t] = In[t];
                        }
                    }

                    // if req'd, compute mean baseline of avg2
                    if (bflg) {
                        for(t=BS, baseline=0.; t<=BE; t++)
                            baseline += Out[t];
                        baseline /= tb;
                    } else {
                        bflg = 0.;
                    }

                    // subtract mean baseline from avg2
                    for(t=0; t<TA; t++)
                        SAM_avg2[t] = Out[t] - baseline;

                    // compute difference with noise normalization
                    for(t=0; t<TA; t++)
                        SAM_avg1[t] = (SAM_avg1[t] - SAM_avg2[t]);
                    break;

                default:
                    break;
            }

            // if req'd, estimate polarity
            sign = 1.;
            if (pflg) {
                for (t=PS, tmp=0.; t<=PE; t++)
                    tmp += SAM_avg1[t];
                sign = (tmp > 0.) ? 1. : -1.;
            }
            if (uflg) {
                for (t=US, tmp=0.; t<=UE; t++)
                    tmp += SAM_avg1[t];
                Image[0][c] = (float)fabs(tmp);
            } else {
                // for each latency...
                for (d=0; d<NumImg; d++) {
                    t = d * NumStep;
                    if (aflg)
                        Image[d][c] = (float)fabs(SAM_avg1[t]);
                    else
                        Image[d][c] = (float)(SAM_avg1[t] * sign);
                }
            }
        } else {
            for (d=0; d<NumImg; d++)
                Image[d][c] = 0.;
        }

        // percent done
        if (vflg) {
            percent = 100. * (double)c / (double)V;
            if (rint(percent) > (double)pct) {
                pct = (int)rint(percent);
                printf("\r\t%d percent done", pct);
                fflush(stdout);
            }
        }
    }

    // open output SAM image file, & write header
    if (vflg) {
        printf("\nSAM(ers) image done\n");
        printf("writing SAM(ers) images");
        fflush(stdout);
    }
    p = get_parm("OutName");
    if (p->set) {
        ParmName = (char *)p->ptr;
    }
    if (!strncmp(Params.DirName, "NULL", 4))
        sprintf(ImgPath, "%s/%s,%s,", SAMpath, Prefix, ParmName);
    else
        sprintf(ImgPath, "%s/%s,%s,", Params.DirName, Prefix, ParmName);
    switch (Params.NumMark) {
        case 1: sprintf(fpath, "%s%s,%s,ERS.nii", ImgPath, Params.Marker[0].MarkName, Suffix); break;
        case 2: sprintf(fpath, "%s%s-%s,%s,dERS.nii", ImgPath, Params.Marker[0].MarkName, Params.Marker[1].MarkName, Suffix); break;
        default: break;
    }
    if ((fp = fopen(fpath, "w")) == NULL)
        Cleanup("can't open .nii file for write");

    // Set output range (default all)
    RS = 0;
    RE = NumImg;
    p = get_parm("OutRange");
    if (p->set) {
        outrange = (TIMERANGE *)p->ptr;
        tmp = outrange->t0 - Params.DataSegStart;
        tmp2 = NumImg * Params.TimeStep;
        if (tmp < 0. || tmp > tmp2) {               // ignore it if it's out of range
            tmp = 0.;
        }
        RS = (int)floor(tmp / Params.TimeStep);
        tmp = outrange->t1 - Params.DataSegStart;
        if (tmp < 0. || tmp > tmp2) {
            tmp = tmp2;
        }
        RE = (int)floor(tmp / Params.TimeStep);

        Params.DataSegStart += RS * Params.TimeStep;

        if (vflg) {
            tmp = Params.DataSegStart + (RE - RS) * Params.TimeStep;
            printf(" -- writing from %g to %g", Params.DataSegStart, tmp);
        }
    }

    // modify NIFTI header
    memset(extension, 0, 4);                                    // for now, ignore the fact that we may have an extended header
    NiiHdr.dim[0] = 4;                                          // 3d+time
    NiiHdr.dim[4] = RE - RS;                                    // number of 3d images

    NiiHdr.toffset = (float)Params.DataSegStart;    // time offset shift (markers 1 & 2 must be identical)
    NiiHdr.slice_duration = Params.TimeStep;                    // time step size --- unused?
    NiiHdr.pixdim[4] = Params.TimeStep;                         // TR is really stored here
    NiiHdr.slice_code = NIFTI_SLICE_SEQ_INC;
    NiiHdr.xyzt_units = (NIFTI_UNITS_MM | NIFTI_UNITS_SEC);     // 3d+time units
    memset(NiiHdr.intent_name, 0, 16);
    switch (Params.NumMark) {
        case 1: memcpy(NiiHdr.intent_name, "ERS", 3); break;
        case 2: memcpy(NiiHdr.intent_name, "Differential ERS", 16); break;
        default: break;
    }

    // write header & extension
    if (fwrite((void *)&NiiHdr, sizeof(nifti_1_header), 1, fp) != 1)
        Cleanup("can't write nifti header to .nii file");
    if (fwrite((void *)extension, 4, 1, fp) != 1)
        Cleanup("can't write 'extension' to .nii file");
    if ((int)NiiHdr.vox_offset != 352 && ExtHdr != NULL) {
        N = (int)NiiHdr.vox_offset - 352;
        if (fwrite((void *)ExtHdr, N, 1, fp) != 1)
            Cleanup("can't write extended header to .nii file");
    }
    for (d=RS; d<RE; d++)   // write time as BRIKs
        if (fwrite((void *)Image[d], sizeof(float), V, fp) != V)
            Cleanup("can't write data image file");
    fclose(fp);

    // we're done now!
    if (vflg) {
        printf(" - done\n'%s' done\n", Progname);
    }
    exit(0);
}
