// sam_hfo -- synthetic aperture magnetometry for imaging epiletic activity.
//  sam_hfo reads SAM weights for volume image, & computes the source time-series
//  for each voxel.
//
//  A bandpass filter for HFOs is applied to the data. For each voxel, the virtual
//  sensor time-series is computed & an FFT applied to each data segment of specified
//  duration. A search for the largest power among segments & frequencies is performed,
//  and the excess kurtosis of the segments at the best frequency is used to localize
//  short bursts of HFOs.
//
//  flags:
//      -r <run name>       -- MEG run name
//      -d <data name>      -- MEG data file name (4D, only)
//      -m <name>           -- parameter file name
//      -v                  -- verbose mode
//
//  Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <getopt.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <geoms.h>
#include <samlib.h>
#include <siglib.h>
#include <DataFiles.h>
#include <SAMfiles.h>
#include <filters.h>
#include <nifti1.h>
#include <version.h>
#include "samutil.h"
#include "sam_parse.h"

#define MINOR_REV   5
#define SIG_ORDER   4

int main(
    int     argc,
    char    **argv
) {
    IIRSPEC         HFO_IIR;        // IIR HFO data filter structure
    FFTSPEC         HFO_FFT;        // HFOFilter -- FFT filter structure for HFOs
    FFTSPEC         NotchFFT;       // FFT notch filter structure (sum of notches)
    PARMINFO        Params;         // analysis parameters
    PARM            *p;             // pointer into active parameter list
    HeaderInfo      Header;         // MEG data header
    ChannelInfo     *Channel;       // MEG channel info
    EpochInfo       *Epoch;         // MEG epoch info
    nifti_1_header  NiiHdr;         // NIFTI header
    unsigned char   *ExtHdr;        // NiiHdr[N] -- extended NIFTI header
    double          **x;            // x[TT][M] -- multi-sensor signal
    double          **Wgt;          // Wgt[V][M] -- SAM coefficients
    double          *spect;         // spect[F] -- real & complex spectrum
    double          *In;            // In[TT] -- input time-series
    double          *Out;           // Out[TT] -- output time-series
    double          *seg;           // seg[D] -- data segment
    double          max;            // maximum power
    double          mean;           // mean power
    double          rms;            // rms power
    double          sx;             // sum of x
    double          sx2;            // sum of x^2
    double          sx4;            // sum of x^4
    double          p2;             // p^2
    double          p4;             // p^4
    double          var;            // 2nd moment (variance)
    double          BWFreq;         // filter bandwidth
    double          tw;             // total samples in Window
    double          scale;          // bits-->T conversion
    double          band;           // frequency band at optimal power
    double          percent;        // percent done
    double          tmp;            // what else?
    float           ***pwr;         // pwr[V][F][D] -- power in each band/segment
    float           *Image;         // Image[V] -- SAM(g2) image
    int             NumChans;       // number of channels in weight file
    int             NumSegs;        // number of segments
    int             D;              // duration of segment (samples) & length of spectrum
    int             E;              // number of good epochs
    int             F;              // number of bands (fhi-flo+1)
    int             M;              // number of MEG or reference channels used
    int             N;              // number of characters in extended NIFTI header
    int             T;              // number of samples
    int             TT;             // total sample size
    int             V;              // number of voxels
    int             pct;            // integer percentage
    int             m;              // channel index
    int             n;              // alternate channel index
    int             e;              // epoch index
    int             f;              // 1st primary channel index
    int             i;              // i-index
    int             t;              // sample index
    int             tt;             // alternate time index
    int             v;              // voxel index
    int             freq;           // alternate frequency index
    int             flo;            // index of lowest frequency bin
    int             fhi;            // index of highest frequency bin
    static int      eflg = FALSE;   // command-line error flag
    static int      rflg = FALSE;   // dataset name flag
    static int      mflg = FALSE;   // parameter file flag
    static int      vflg = FALSE;   // verbose mode flag
    char            fpath[256];     // general path name
    char            *DSName = NULL; // MEG dataset name
    char            DSpath[256];    // MEG path name
    char            SAMpath[256];   // SAM subdirectory path
    char            ImgPath[256];   // output path for writing image
    char            *ParmName = NULL; // parameter file name
    char            Prefix[64];     // image file prefix
    char            extension[4];   // extension
    FILE            *fp;            // output file pointer (to SAM image)

    // system-dependent variables
#if BTI
    static int      dflg;           // pdf file name flag
    char            *PDFName;       // name of data file
#else
    unsigned char   **Bad;          // Bad[E][T] -- flags for bad samples   @@@
    unsigned char   *Flags;         // Flags[TT] -- transformed flags
#endif

    Progname = argv[0];
    reg_usage(MINOR_REV, __DATE__);

    // Register the parameters used by this program.

    reg_std_parm();

    reg_parm("CovType");
    set_parm_arg_help("CovType", "GLOBAL");
    set_parm_help("CovType", "GLOBAL is always used");

    reg_parm("CovBand");
    reg_parm("ImageBand");
    reg_parm("FilterType");
    reg_parm("Notch");
    reg_parm("Hz");

    reg_parm("TimeInt");

    reg_parm("ImageMetric");
    set_parm_arg_help("ImageMetric", "Kurtosis");
    set_parm_help("ImageMetric", "Kurtosis is always used");

    reg_parm("ImageDirectory");
    reg_parm("PrefixLength");
    reg_parm("OutName");

    // Parse the arguments, environment variables, and parameter files.

    do_parse_args(argc, argv);

    // Extract specified parameters.

    eflg = get_params(&Params);

    p = get_parm("DataSet");
    if (p->set) {
        rflg = TRUE;
        DSName = Params.DataSetName;
    }
#if BTI
    p = get_parm("PDFName");
    if (p->set) {
        dflg = TRUE;
        PDFName = (char *)p->ptr;
    }
#endif
    p = get_parm("verbose");
    vflg = p->set;
    p = get_parm("param");
    mflg = p->set;

#if BTI
    if(!rflg || !dflg || !mflg || eflg) {
#else
    if(!rflg || !mflg || eflg) {
#endif
        msg("dataset (-r) and parameter file (-m) are required\n");
        do_help();  // doesn't return
    }

    ParmName = Params.ParmName;
    Params.CovType = GLOBAL;        // global is the only choice
    Params.ImageMetric = KURTOSIS;  // kurtosis is the only choice

    // get dataset information structures
    if(vflg == TRUE) {
        printf("S A M   H F O   I M A G I N G\t%s\n", PRG_REV);
        printf("opening data file");
        fflush(stdout);
    }

#if BTI
    // get data with sensor structures
    if(GetMagnesInfo(DSName, PDFName, &Header, &Channel, &Epoch) == -1)
        Cleanup("can't read Magnes information");
    sprintf(DSpath, "%s/%s", Header.DsPath, Header.SetName);
#else
    // get data with sensor structures
    GetDsInfo(DSName, &Header, &Channel, &Epoch, &Bad, TRUE);
    sprintf(DSpath, "%s/%s.ds", Header.DsPath, Header.SetName);
#endif
    sprintf(SAMpath, "%s/SAM", DSpath);

    // establish epoch, channel, & sample count
    E = Header.NumEpochs;
    T = Header.MaxSamples;
    TT = E * T;
    M = Header.NumPri;
    f = Header.PsIndex[0];      // 1st MEG primary sensor channel index

    // check to see if this is a continuous dataset
    if(E != 1)
        msg("(reminder: multi-epoch data must be contiguous)\n");

    // check if this is the right program for you!
    if(Params.CovHP == -999. || Params.CovLP == -999.)
        Cleanup("CovBand specifications required");
    if(Params.ImageHP == -999. || Params.ImageLP == -999.)
        Cleanup("ImageBand specifications required");
    if(Params.ImageHP < Channel[f].AnalogHpFreq)
        Cleanup("ImageBand highpass cutoff frequency out of recorded range");
    if(Params.ImageLP > Channel[f].AnalogLpFreq)
        Cleanup("ImageBand lowpass cutoff frequency out of recorded range");
    if(strncmp(Params.DirName, "NULL", 4))
        sprintf(ImgPath, "%s", Params.DirName);
    else
        sprintf(ImgPath, "%s", SAMpath);
    if(Params.TimeInt == -999.)
        Cleanup("'TimeInt' req'd for FFT window duration");

    // determine FFT window
    D = (int)rint(Params.TimeInt * Header.SampleRate);
    NumSegs = TT / D;

    // determine frequency bands
    flo = st_freq(Params.ImageHP, D, Header.SampleRate);
    fhi = st_freq(Params.ImageLP, D, Header.SampleRate);
    F = fhi - flo + 1;              // number of bands

    // set prefix to NumPrefix characters (default is 8-character dataset hashcode)
    memset(Prefix, 0, sizeof(Prefix));
#if BTI
    memcpy(Prefix, "BIOMAG4D", 8);              // 4D datasets just have serial numbers
#else
    memcpy(Prefix, DSName, Params.NumPrefix);

    // copy flags
    if((Flags = (unsigned char *)malloc((size_t)TT * sizeof(unsigned char))) == NULL)
        allocfailed("Flags[]");
    for(e=0; e<E; e++)
        for(t=0; t<T; t++) {
            tt = e * T + t;
            Flags[tt] = Bad[e][t];
        }
#endif

    // get weight name from parameters & read SAM weights
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("reading Global SAM weights");
        fflush(stdout);
    }
    sprintf(fpath, "%s/%s,%-d-%-dHz/Global.nii", SAMpath, ParmName, (int)Params.CovHP, (int)Params.CovLP);
    GetNIFTIWts(fpath, &Wgt, &NumChans, &V, &NiiHdr, extension, &ExtHdr);
    if(NumChans != M)
        Cleanup("number of channels in dataset does not match number in weight file");

    // set up imaging constants
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("determining imaging constants:\n");
        fflush(stdout);
    }
    if(vflg == TRUE) {
        printf("\tsegment duration=%d samples\n", D);
        printf("\tnumber of segments=%d\n", NumSegs);
        printf("\tflo=%d, fhi=%d, F=%d\n", flo, fhi, F);
        printf("\tepochs=%d, samples/epoch=%d (total %d)\n", E, T, TT);
        fflush(stdout);
    }

    // allocate memory for arrays
    if(vflg == TRUE) {
        printf("allocating memory");
        fflush(stdout);
    }
    if((x = (double **)malloc((size_t)TT * sizeof(double *))) == NULL)
        allocfailed("x[]");
    for(t=0; t<TT; t++)
        if((x[t] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
            allocfailed("x[][]");
    if((In = (double *)malloc((size_t)TT * sizeof(double))) == NULL)
        allocfailed("In[]");
    if((Out = (double *)malloc((size_t)TT * sizeof(double))) == NULL)
        allocfailed("Out[]");
    if((Image = (float *)malloc((size_t)V * sizeof(float))) == NULL)
        allocfailed("Image[]");
    if((pwr = (float ***)malloc((size_t)V * sizeof(float **))) == NULL)
        allocfailed("pwr[]");
    for(v=0; v<V; v++) {
        if((pwr[v] = (float **)malloc((size_t)F * sizeof(float *))) == NULL)
            allocfailed("pwr[][]");
        for(f=0; f<F; f++)
            if((pwr[v][f] = (float *)malloc((size_t)NumSegs * sizeof(float))) == NULL)
                allocfailed("pwr[][][]");
    }
    if((spect = (double *)malloc((size_t)D * sizeof(double))) == NULL)
        allocfailed("spect[]");
    if((seg = (double *)malloc((size_t)D * sizeof(double))) == NULL)
        allocfailed("seg[]");

    // generate HFO filter
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("making %6.3f to %6.3f Hz filter for HFO bandpass", Params.ImageHP, Params.ImageLP);
        fflush(stdout);
    }
    if(Params.FilterType == IIR)
        mkfilter(&HFO_IIR, SIG_ORDER, Header.SampleRate, Params.ImageHP, Params.ImageLP, &BWFreq);
    else
        Butterworth(&HFO_FFT, TT, Params.ImageHP, Params.ImageLP, Header.SampleRate, BANDPASS, MAXORDER, &BWFreq);

    // compute notch filter
    if(Params.FilterType == IIR) {
        msg("\nreminder: when using IIR filters, data should be notch filtered for power mains prior to analysis\n");
    } else {
        if(Params.Notch == TRUE) {
            if(vflg == TRUE) {
                printf(" - done\n");
                printf("computing FFT notch filter");
                fflush(stdout);
            }
            mknotch(&NotchFFT, Params.CovHP, Params.CovLP, Header.SampleRate, T, Params.Hz);
        } else {
            Butterworth(&NotchFFT, T, 0., 0., Header.SampleRate, ALLPASS, MAXORDER, &tmp);          // dummy allpass filter when notch is not used
        }
    }

    // read & filter MEG data
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("reading & filtering MEG data");
        fflush(stdout);
    }
    for(m=0; m<M; m++) {                            // read & process data in channel order
        n = Header.PsIndex[m];
        scale = Channel[n].Scale;
        for(e=0; e<E; e++) {                        // multi-epoch data is presumed to be contiguous!
#if BTI
            if(GetMagnesChan(DSpath, PDFName, &Header, Channel, e, m, In) == -1)
                Cleanup("'GetMagnesChan()' failed");
#else
            GetDsData(&Header, e, n, scale, In);    // get one epoch of data for this channel -- length T samples
#endif
            // make continuous data out of multi-epochs
            for(t=0; t<T; t++) {
                tt = e * T + t;
                Out[tt] = In[t];
            }
        }
        demean(Out, TT);                            // remove the mean
        if(Params.FilterType == IIR) {
            bdiir(Out, In, TT, &HFO_IIR);           // now, filter the 'Out' data
            for(t=0; t<TT; t++)                     // move to 'x[][]' array
                x[t][m] = In[t];
        } else {
            FFTfilter(Out, In, &HFO_FFT, TT);       // now, filter the 'Out' data
            FFTfilter(In, Out, &NotchFFT, T);
            for(t=0; t<TT; t++)                     // move to 'x[][]' array
                x[t][m] = Out[t];
        }
        if(Params.FilterType == IIR)
            bdiir(Out, In, TT, &HFO_IIR);           // now, filter the 'Out' data
        else
            FFTfilter(Out, In, &HFO_FFT, TT);       // now, filter the 'Out' data
        for(t=0; t<TT; t++)                         // move to 'x[][]' array
            x[t][m] = In[t];

        // heartbeat
        if(vflg == TRUE) {
            printf(".");
            fflush(stdout);
        }
    }

    // now, step through all coordinates
    if(vflg == TRUE) {
        printf(" - done\n");
        printf("stepping through voxels to determine power in each band & segment:\n");
        fflush(stdout);
    }
    for(v=0, pct=(-1); v<V; v++) {
        for(m=0, tmp=0.; m<M; m++)
            tmp += Wgt[v][m] * Wgt[v][m];
        if(tmp > 0.) {

            // compute virtual sensor time series
            for(t=0; t<TT; t++) {
                for(m=0, tmp=0.; m<M; m++)
                    tmp += Wgt[v][m] * x[t][m];
                In[t] = tmp;
            }

            // step through segments
            for(i=0; i<NumSegs; i++) {
                for(t=0, tt=i*D; t<D; t++, tt++)
                    seg[t] = In[tt];

                // compute fft
                fft(D, seg, spect);

                // compute power in segment & accumulate stats -- spect has real & imaginary pairs
                for(f=0, freq=2*flo; f<F; f++, freq+=2)         // power = real^2 + imag^2
                    pwr[v][f][i] = (float)(spect[freq] * spect[freq] + spect[freq+1] * spect[freq+1]);
            }

        } else {
            for(f=0; f<F; f++)
                for(i=0; i<NumSegs; i++)
                    pwr[v][f][i] = 0.;
        }

        // percent done
        if(vflg == TRUE) {
            percent = 100. * (double)v / (double)V;
            if(rint(percent) > (double)pct) {
                pct = (int)rint(percent);
                printf("\r\t%d percent done", pct);
                fflush(stdout);
            }
        }
    }

    // if verbose, summarize some statistics
    if(vflg == TRUE) {
        printf("\nsummarizing hfo statistics:\n");
        for(f=0; f<F; f++) {
            band = Params.ImageLP * (double)(flo + f) / (double)fhi;
            for(v=0, max=sx=sx2=0.; v<V; v++) {
                for(i=0; i<NumSegs; i++) {
                    if(pwr[v][f][i] > max)
                        max = pwr[v][f][i];
                    sx += pwr[v][f][i];
                    sx2 += pwr[v][f][i] * pwr[v][f][i];
                }
            }
            tmp = (double)(NumSegs * V);
            mean = sx / tmp;
            rms = sqrt(sx2 / tmp);
            printf("freq=%5.1f Hz, max=%e, mean=%e, rms=%e\n", band, max, mean, rms);fflush(stdout);
        }
    }

    p = get_parm("OutName");
    if (p->set) {
        ParmName = (char *)p->ptr;
    }

    // modify NIFTI header
    memset(extension, 0, 4);                // for now, ignore the fact that we may have an extended header
    NiiHdr.dim[0] = 3;                      // 3d volume
    NiiHdr.dim[4] = 0;                      // number of 3d images
    NiiHdr.toffset = 0.;                    // time offset shift
    NiiHdr.slice_duration = 1.;
    NiiHdr.slice_code = NIFTI_SLICE_UNKNOWN;
    NiiHdr.xyzt_units = NIFTI_UNITS_MM;     // units
    memset(NiiHdr.intent_name, 0, 16);
    memcpy(NiiHdr.intent_name, "Epilepsy HFO", 12);

    // once more, step through all coordinates, but this time determine g2 for each band
    for(f=0; f<F; f++) {

        // get band frequency value
        band = Params.ImageLP * (double)(flo + f) / (double)fhi;
        if(vflg == TRUE) {
            printf("stepping through voxels to compute g2 at %5.1f Hz\n", band);
            fflush(stdout);
        }

        // step through voxels
        for(v=0, pct=(-1); v<V; v++) {

            if(pwr[v][f][0] > 0.) {

                // compute voxel g2 for this band
                for(i=0, sx2=sx4=tw=0.; i<NumSegs; i++) {
                    p2 = (double)(pwr[v][f][i] * pwr[v][f][i]);
                    p4 = p2 * p2;
                    sx2 += p2;
                    sx4 += p4;
                    tw += 1.;
                }
                var = sx2 / tw;
                Image[v] = (float)(sx4 / (tw * var * var) - 3.);                // g2 of current band!
            } else {
                Image[v] = 0.;
            }

            // percent done
            if(vflg == TRUE) {
                percent = 100. * (double)v / (double)V;
                if(rint(percent) > (double)pct) {
                    pct = (int)rint(percent);
                    printf("\r\t%d percent done", pct);
                    fflush(stdout);
                }
            }
        }

        // write out SAM(g2) image
        if(vflg == TRUE) {
            printf(" - done\n");
            printf("writing hfo image for %5.1f Hz band", band);
            fflush(stdout);
        }

        sprintf(fpath, "%s/%s,%s,%05.1fHz,HFO.nii", ImgPath, Prefix, ParmName, band);
        if((fp = fopen(fpath, "w")) == NULL)
            Cleanup("can't open .nii file for write");

        if(fwrite((void *)&NiiHdr, sizeof(nifti_1_header), 1, fp) != 1)
            Cleanup("can't write nifti header to .nii file");
        if(fwrite((void *)extension, 4, 1, fp) != 1)
            Cleanup("can't write 'extension' to .nii file");
        if((int)NiiHdr.vox_offset != 352 && ExtHdr != NULL) {
            N = (int)NiiHdr.vox_offset - 352;
            if(fwrite((void *)ExtHdr, N, 1, fp) != 1)
                Cleanup("can't write extended header to .nii file");
        }
        if(fwrite((void *)Image, sizeof(float), V, fp) != V)
            Cleanup("can't write voxel values to .nii file");
        fclose(fp);
    }

    // that's all, folks!
    if(vflg) {
        printf(" - done\n'%s' done\n", Progname);
    }
    exit(0);
}
