// sam_wts -- linear synthetic aperture magnetometry coefficient solution.
//  Linear solution for the Cartesian moment vector, using multiple
//  local spheres. Orientation vector is determined from a separate
//  covariance matrix.
//
// Features of this version include: reading a cortical atlas file, reading
//  a coordinate transform, & generating target weights using atlas nearest
//  neighbor coordinates. Note that the presence of an atlas over-rides the
//  Talairach transformed weights.
//
// Additional changes require that the MRI directory pathname is provided
//  so that hull.shape & atlas mesh files can be found. This enforces the
//  directory hierarchy where all MRIs are under a single directory and
//  identified by a prefix such as the 8-character hashcode.
//
//  flags:
//      -r <run name>       -- MEG dataset name
//      -d <data file name> -- MEG data file name (4D, only)
//      -m <parameters>     -- parameter file name
//      -B                  -- output primary sensor forward solutions for target list (along with weights)
//      -n                  -- output SAM weights for normalized (SNR) source projection
//      -v                  -- Verbose mode
//
//  Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <model.h>
#include <voxel.h>
#include <samlib.h>
#include <siglib.h>
#include <lfulib.h>
#include <DataFiles.h>
#include <SAMfiles.h>
#include <version.h>
#include "samutil.h"
#include "sam_parse.h"

#define MINOR_REV       10
#define ETA             1.0e-12     // a small fraction of a step

typedef struct {
    double      **Cov;              // Cov[M][M] -- covariance matrix
    double      **Cinv;             // Cinv[M][M] -- inverse covariance matrix
    double      Noise;              // noise power/variance -- including optional regularization
    double      **Wd;               // Wd[V][M] -- primary sensor weight vector
    char        Name[64];           // short name
    char        CovFileName[256];   // covariance pathname
    char        WgtFileName[256];   // weight pathname
    char        NiiName[64];        // nifti weight file name
    COV_HDR     WgtCovHdr;          // weight covariance header
    SAM_HDR     SAMHeader;          // weight file header
    int         Used;               // flag for computing weights
} COV_WGTS;

static int vflg = FALSE;            // verbose mode flag (default: silent)

static void fixBounds(PARMINFO *p);

extern int CoeffFlag;               // initialization for Nolte solution

int main(
    int             argc,
    char            **argv
) {
    COV_WGTS        *Stats;             // analysis parameters for covariance files
    PARMINFO        Params;             // analysis parameters
    PARM            *p;                 // pointer into active parameter list
    HULL            Hull;               // cortical hull model -- vertices & normals
    HeaderInfo      Header;             // MEG data header
    ChannelInfo     *Channel;           // MEG channel info
    ChannelInfo     *InChannel;         // original MEG channel info
    EpochInfo       *Epoch;             // MEG epoch info
    VOXELINFO       *Voxel;             // Voxel[V] -- structure for voxel info (including targets)
    VOXELINFO       *Atlas;             // Atlas[A] -- structure for atlas voxel info
    VOXELINFO       *Target;            // Target[V] -- structure for reading targets
    VOXELINFO       VoxDelta;           // interpolated voxel
    double          *SortW;             // SortW[M] -- sorted singular values (don't confuse these with weights!)
    double          *Bd;                // Bd[M] -- primary sensor lead field for ECD solution, only
    double          *WgtDelta;          // WgtDelta[M] -- weights for each interpolation
    double          HPFreq = 0.;        // highpass frequency
    double          LPFreq = 0.;        // lowpass frequency
    double          Src;                // voxel source power
    double          Nse;                // voxel noise power
    double          Mu;                 // regularization
    double          percent;            // floating percentage
    double          span;               // singular value span for solution
    double          min;                // minimum distance
    double          r2;                 // distance^2
    int             ndim = 0.;          // tolerance for pinv()
    register double tmp;                // what else?
    int             *ChanIndex;         // ChanIndex[M] -- index of channels used for covariance
    int             A;                  // number of atlas voxels
    int             M;                  // number of MEG or reference channels used
    int             R;                  // number of physical reference SQUID channels
    int             S;                  // number of physical SQUID channels
    int             T;                  // number of target voxels
    int             V;                  // number of voxels
    int             c;                  // virtual channel
    int             m;                  // channel index
    int             n;                  // covariance type index
    int             v;                  // vector index
    int             i;                  // matrix index
    int             j;                  // matrix index
    int             ix;                 // x-interpolation index
    int             iy;                 // y-interpolation index
    int             iz;                 // z-interpolation index
    int             index = 0;          // index for nearest neighbor search
    int             pct;                // integer percentage
    int             a;                  // atlas index
    int             t;                  // target index
    int             NumCov;             // how many covariance matrices we make
    static int      Order = 2;          // coil integration order -- default 2nd-order
    static int      eflg = FALSE;       // command-line error flag
    static int      aflg = FALSE;       // atlas flag
    static int      bflg = FALSE;       // forward solution flag
    static int      iflg = FALSE;       // interpolation flag
    static int      mflg = FALSE;       // parameter file flag
    static int      rflg = FALSE;       // dataset name flag
    static int      tflg = FALSE;       // target file flag
    static int      wflg = FALSE;       // normalize
    static int      pflg = FALSE;       // use pseudo-inverse
    static int      xflg = FALSE;       // atlas transform flag
    char            *s;                 // string pointer
    char            fpath[256];         // general path name
    char            *DSName = NULL;     // MEG dataset name
    char            DSpath[256];        // MEG dataset path
    char            SAMpath[256];       // SAM subdirectory path
    char            HullPath[256];      // hull.shape path
    char            AtlasPath[256];     // atlas path
    char            TargetPath[256];    // target path
    char            TransformPath[256]; // transform path
    char            TlrcPath[256];      // orth+tlrc path
    char            CovDirName[256];    // covariance subdirectory name
    char            *ParmName = NULL;   // parameter file name
    char            CurrentDir[256];    // current directory
    char            ByYourCommand[256]; // string for system call
    char            FwdID[] =           // SAM fwd solution ID
        "SAMFWDSL";
    FILE            *fp;
    FILE            *np;
    FILE            *bp = NULL;         // forward solution pointer

    // system-dependent variables
#if BTI
    static int      dflg;               // pdf file name flag
    char            *PDFName;           // name of data file
#else
    HEAD_MODEL      Model;              // model structure for .hdm file
    int             found;              // flag for finding sensor name matches
    int             k;                  // k-index
#endif

    Progname = argv[0];
    reg_usage(MINOR_REV, __DATE__);

    // Register the parameters used by this program.

    reg_std_parm();

    reg_parm("NumMarkers");
    reg_parm("Marker");

    reg_parm("CovBand");
    reg_parm("NoiseBand");
    reg_parm("Mu");
    reg_parm("PropMu");

    reg_parm("XBounds");
    reg_parm("YBounds");
    reg_parm("ZBounds");
    reg_parm("ImageStep");

    reg_parm("MRIDirectory");
    reg_parm("MRIPattern");
    reg_parm("PrefixLength");
    reg_parm("HullName");

    reg_parm("Normalize");
    set_parm_help("Normalize", "normalize the SAM weights by sqrt(noise)");
    reg_parm("Field");

    reg_parm("Pinv");

    reg_parm("Model");
    reg_parm("doMultiSphere");  // since Model isn't allowed on the command line
    reg_parm("doNolte");        // allow --domulti or --donolte

    reg_parm("AtlasName");
    reg_parm("TargetName");
    reg_parm("Transform");
    reg_parm("ImageFormat");
    reg_parm("Interpolate");

    // Parse the arguments, environment variables, and parameter files.

    do_parse_args(argc, argv);

    // Extract specified parameters.

    eflg = get_params(&Params);

    p = get_parm("DataSet");
    if (p->set) {
        rflg = TRUE;
        DSName = Params.DataSetName;
    }
#if BTI
    p = get_parm("PDFName");
    if (p->set) {
        dflg = TRUE;
        PDFName = (char *)p->ptr;
    }
#endif
    p = get_parm("verbose");
    vflg = p->set;
    p = get_parm("param");
    mflg = p->set;

    p = get_parm("Normalize");      // -n
    wflg = p->set;
    p = get_parm("Field");          // -B
    bflg = p->set;
    p = get_parm("Pinv");           // -p
    pflg = p->set;
    if (pflg) {
        ndim = *(int *)p->ptr;
    }

#if BTI
    if (eflg || !rflg || !dflg || !mflg) {
#else
    if (eflg || !rflg || !mflg) {
#endif
        msg("dataset (-r) and parameter file (-m) are required\n");
        do_help();  // doesn't return
    }

    ParmName = Params.ParmName;

    // check for model override
    j = 0;
    p = get_parm("doMultiSphere");
    if (p->set) {
        Params.Model = MSPHERE;
        j++;
    }
    p = get_parm("doNolte");
    if (p->set) {
        Params.Model = NOLTE;
        j++;
    }
    if (j == 2) {
        Cleanup("can't specify both MultiSphere and Nolte");
    }

    // announce
    if (vflg) {
        printf("S A M   C O E F F I C I E N T   G E N E R A T O R\t%s\n", PRG_REV);
        printf("opening data file");
        fflush(stdout);
    }

#if BTI
    // get data with sensor structures
    if (GetMagnesInfo(DSName, PDFName, &Header, &Channel, &Epoch) == -1)
        Cleanup("can't read Magnes information");
    sprintf(DSpath, "%s/%s", Header.DsPath, Header.SetName);
#else
    // get data with sensor structures
    GetDsInfo(DSName, &Header, &Channel, &Epoch, NULL, TRUE);
    sprintf(DSpath, "%s/%s.ds", Header.DsPath, Header.SetName);
#endif
    sprintf(SAMpath, "%s/SAM", DSpath);

    // set channel & epoch count constants
    M = Header.NumPri;
    R = Header.NumRef;
    S = Header.NumPri + Header.NumRef;

    // test for MRI directory pathname for use with hull.shape, cortical atlas, & Talairach transform
    p = get_parm("MRIDirectory");
    if (!p->set) {
        if (Params.Model != SSPHERE)        // MRI not req'd for single sphere model
            fatalerr("MRIDirectory root req'd for use with Nolte or multi-sphere model");
    } else {
        p = get_parm("AtlasName");
        if (p->set) {
            GetMRIPath(AtlasPath, sizeof(AtlasPath), &Params, (char *)p->ptr, TRUE);
            Params.ImageFormat = ORIG;      // @@@ force ORIG if atlas is present
            aflg = TRUE;
        }
        p = get_parm("TargetName");
        if (p->set) {
            GetMRIPath(TargetPath, sizeof(TargetPath), &Params, (char *)p->ptr, TRUE);
            tflg = TRUE;
        }
        p = get_parm("Transform");
        if (p->set) {
            GetMRIPath(TransformPath, sizeof(TransformPath), &Params, (char *)p->ptr, TRUE);
            xflg = TRUE;
        }
        if (Params.Model == NOLTE || Params.Model == MSPHERE) {
            p = get_parm("HullName");
            if (p->set) {
                GetMRIPath(HullPath, sizeof(HullPath), &Params, (char *)p->ptr, TRUE);
            } else {
                GetMRIPath(HullPath, sizeof(HullPath), &Params, "hull.shape", TRUE);
            }
        }
        if (Params.ImageFormat == TLRC) {           // make sure that we can access brain+tlrc for @auto_tlrc step
            GetMRIPath(TlrcPath, sizeof(TlrcPath), &Params, "brain+tlrc.HEAD", TRUE);
            if (Params.ImageRes < (1000. * Params.SAMStep) || Params.ImageRes > 20.)
                Cleanup("Talairach resolution must be >= ortho image step-size & < 20 mm");
        }
    }

    // determine weight interpolation step
    if (Params.Interpolate > 0.) {
        if (Params.Interpolate <= 0.5 * Params.SAMStep)
            iflg = TRUE;
        else
            Cleanup("'Interpolate' step size must be <= 'SAMstep'");
    }

    if (vflg) {
        printf(" - done\n");
    }

    // Make sure that the ROI is an integral number of voxels.

    if (get_parm("XBounds")->set && get_parm("YBounds")->set &&
        get_parm("ZBounds")->set && get_parm("ImageStep")->set) {
        fixBounds(&Params);
    }

    // define forward solution model
    if (Params.Model == -1)
        Cleanup("'Model' specification missing from parameter file");
    if (vflg) {
        printf("modeling forward solution using ");
    }
    switch (Params.Model) {

        // single sphere origin
        case SSPHERE:
            if (vflg) {
                printf("single local sphere origin: { %6.3f, %6.3f, %6.3f } (cm)", 100. * Params.Sp[X_], 100. * Params.Sp[Y_], 100. * Params.Sp[Z_]);
                fflush(stdout);
            }
            model = SPHERE;
            for (i=0; i<M; i++) {
                j = Header.PsIndex[i];
                for (v=X_; v<=Z_; v++)
                    Channel[j].Geom.MEGSensor.LocalSphere[v] = Params.Sp[v];
            }
            for (i=0; i<R; i++) {
                j = Header.RsIndex[i];
                for (v=X_; v<=Z_; v++)
                    Channel[j].Geom.MEGSensor.LocalSphere[v] = Params.Sp[v];
            }
            Hull.nv = 1;
            if ((Hull.vertex = (FIELD *)malloc((size_t)2 * sizeof(FIELD))) == NULL)
                Cleanup("'Hull.vertex[1]' allocation failed");
            for (v=X_; v<=Z_; v++)
                Hull.Vo[v] = Params.Sp[v];
            break;

        // multiple local spheres
        case MSPHERE:
            if (vflg) {
                printf("multiple local spheres");
            }
#if BTI
            // read hs_file into 'Model'
            HStoMS(DSpath, &Header, Channel);
#else
            // read .hdm file into 'Model'
            sprintf(fpath, "%s/default.hdm", DSpath);
            GetHDM(fpath, &Model);

            // get the hull for defining the ROI
            GetHull(HullPath, &Hull);       // read hull model

            // test if the model matches the number of sensors
            if (Model.NumSensors != S)
                Cleanup("number of sensors in '.hdm' file does not match that of the dataset");

            // set multiple sphere origins
            for (i=0; i<S; i++) {
                k = (int)strlen(Model.MultiSphere[i].ChanName) - 1;
                for (j=0, found=FALSE; j<Header.NumChannels; j++)
                    if (!strncmp(Model.MultiSphere[i].ChanName, Channel[j].ChannelName, k)) {
                        for (v=X_; v<=Z_; v++)
                            Channel[j].Geom.MEGSensor.LocalSphere[v] = Model.MultiSphere[i].Origin[v];
                        found = TRUE;
                    }
                if (!found)
                    Cleanup("could not find matching channel name in '.HDM' file");
            }
#endif
            break;

        // Nolte realistic head model
        case NOLTE:
            if (vflg) {
                printf("'%s'", HullPath);
            }
            GetHull(HullPath, &Hull);       // read hull model
            CoeffFlag = TRUE;
            break;

        default:
            break;
    }

    // check for atlas
    if (aflg) {
        if (vflg) {
            printf(" - done\n");
            printf("reading '%s' atlas file instead of using ROI", Params.Atlas);
            fflush(stdout);
        }
        GetVoxelsFS(AtlasPath, &Atlas, &A, Params.SAMStart, Params.SAMEnd, &Params.SAMStep);
        if (vflg) {
            printf("\n%d voxels\n", A);
            printf("X: %g to %g cm\n", 100.*Params.SAMStart[X_], 100.*Params.SAMEnd[X_]);
            printf("Y: %g to %g cm\n", 100.*Params.SAMStart[Y_], 100.*Params.SAMEnd[Y_]);
            printf("Z: %g to %g cm\n", 100.*Params.SAMStart[Z_], 100.*Params.SAMEnd[Z_]);
            printf("Step %g cm\n", 100.*Params.SAMStep);
        }
        if (!tflg) {
            Voxel = Atlas;
            V = A;
        }

        // modify atlas frame if transform is present
        if (xflg) {
            if (vflg) {
                printf("reading '%s' sensor transform file", Params.Transform);
            }
            GetDsInfo(DSName, &Header, &InChannel, &Epoch, NULL, TRUE);
            MoveFrame(&Header, InChannel, Channel, TransformPath);
        }
    }

    // allocate voxels with boundary flags
    if (!strncmp(Params.Target, "NULL", 4)) {       // no target files specified in parameter file
        if (vflg) {
            printf(" - done\n");
            printf("generating ROI voxel array");
            fflush(stdout);
        }
        if (aflg) {
            printf(" using '%s' atlas", Params.Atlas);
            fflush(stdout);
        } else {                                    // if there NO Target file and NO Atlas found, 'Voxel' is allocated!
            if (Params.SAMStart[X_] == -999. || Params.SAMStart[Y_] == -999. || Params.SAMStart[Z_] == -999.)       // no need to test SAMend...
                Cleanup("ROI 'XBounds/YBounds/ZBounds' specifications required");
            if (Params.SAMStep == -999.)
                Cleanup("ROI 'ImageStep' specification required");
#if BTI
            GetVoxelsHS(&Voxel, &V, DSpath, Params.SAMStart, Params.SAMEnd, Params.SAMStep);
#else
            GetVoxels(&Voxel, &V, &Hull, Params.SAMStart, Params.SAMEnd, Params.SAMStep);
#endif
        }
        bflg = FALSE;   // force bflg false

    } else {                                        // a Target file is present!
        if (vflg) {
            printf(" - done\n");
            printf("reading target list file");
            fflush(stdout);
        }

        // atlas present -- find nearest neighbors, & set orientations
        //  NOTE:   'Target' is an array of target coordinates obtained by 'GetTargets()' where T is the number of target coordinates
        //          'Atlas' is an array of atlas coordinates obtained by 'GetVoxelsFS()' where A is the number of atlas coordinates
        if (aflg) {

            GetTargets(TargetPath, &Target, &T);
            if (vflg) {
                printf(" - read %d targets", T);
                fflush(stdout);
            }

            // for each target voxel...
            for (t=0; t<T; t++) {

                // scan atlas for nearest neighbor
                for (a=0, min=HUGE; a<A; a++) {
                    for (v=X_, r2=0.; v<=Z_; v++) {
                        tmp = Atlas[a].p[v] - Target[t].p[v];
                        r2 += tmp * tmp;
                    }
                    if (r2 < min) {
                        min = r2;
                        index = a;
                    }
                }
                if (sqrt(min) > 0.02)
                    msg("\nWarning: target %d was more than 20 mm from nearest atlas voxel\n", t);
                for (v=X_; v<=Z_; v++) {        // set the target position & moment vectors to nearest neighbor in atlas
                    Target[t].p[v] = Atlas[index].p[v];
                    Target[t].v[v] = Atlas[index].v[v];
                }
            }
            Voxel = Target;     // set Voxel pointer to Target
            V = T;
        } else {
            GetTargets(TargetPath, &Voxel, &V);
            if (vflg) {
                printf(" - read %d targets", V);
                fflush(stdout);
            }
        }
    }

    // always use CovBand frequencies
    if (Params.CovHP >= 0. && Params.CovLP >= 0.) {
        HPFreq = Params.CovHP;
        LPFreq = Params.CovLP;
    } else {
        Cleanup("CovBand not specified!");
    }

    // initialize flags
    NumCov = Params.NumMark + 4;
    Stats = new_array(COV_WGTS, NumCov);
    for (n=0; n<NumCov; n++)
        Stats[n].Used = FALSE;

    // generate covariance subdirectory full pathname string by concatenating SAMpath, ParmName, & bandpass
    sprintf(CovDirName, "%s/%s,%-d-%-dHz", SAMpath, ParmName, (int)HPFreq, (int)LPFreq);

    // count actual number of covariance matrices
    if (vflg) {
        printf(" - done\n");
        printf("determining covariance & weight names");
        fflush(stdout);
    }

    // populate full pathnames for Global
    sprintf(Stats[GLOBAL_].Name, "Global");
    sprintf(Stats[GLOBAL_].CovFileName,"%s/Global.cov", CovDirName);
    sprintf(Stats[GLOBAL_].WgtFileName, "%s/Global.nii", CovDirName);
    sprintf(Stats[GLOBAL_].NiiName, "Global.nii");
    Stats[GLOBAL_].Used = TRUE;

    // populate full pathname for Orient
    sprintf(Stats[ORIENT_].Name, "Orient");
    sprintf(Stats[ORIENT_].CovFileName,"%s/Orient.cov", CovDirName);
    sprintf(Stats[ORIENT_].WgtFileName, "NULL");
    sprintf(Stats[ORIENT_].NiiName, "NULL");
    Stats[ORIENT_].Used = TRUE;

    // check for Noise covariance
    if (Params.NoiseHP != -999. && Params.NoiseLP != -999.) {
        sprintf(Stats[NOISE_].Name, "Noise");
        sprintf(Stats[NOISE_].CovFileName,"%s/Noise.cov", CovDirName);
        sprintf(Stats[NOISE_].WgtFileName, "%s/Noise.nii", CovDirName);
        sprintf(Stats[NOISE_].NiiName, "Noise.nii");
        Stats[NOISE_].Used = TRUE;
    }

    // now check markers
    if (Params.NumMark > 0) {

        // include Sum covariance
        sprintf(Stats[SUM_].Name, "Sum");
        sprintf(Stats[SUM_].CovFileName,"%s/Sum.cov", CovDirName);
        sprintf(Stats[SUM_].WgtFileName, "%s/Sum.nii", CovDirName);
        sprintf(Stats[SUM_].NiiName, "Sum.nii");
        Stats[SUM_].Used = TRUE;

        for (i=0; i<Params.NumMark; i++) {
            if (Params.Marker[i].MarkName2) {
                s = Params.Marker[i].MarkName2;
            } else {
                s = Params.Marker[i].MarkName;
            }
            j = MARKER1_ + i;
            sprintf(Stats[j].Name, "%s", s);
            sprintf(Stats[j].CovFileName,"%s/%s.cov", CovDirName, s);
            sprintf(Stats[j].WgtFileName, "%s/%s.nii", CovDirName, s);
            sprintf(Stats[j].NiiName, "%s.nii", s);
            Stats[j].Used = TRUE;
        }
    }

    // allocate memory for arrays
    if (vflg) {
        printf(" - done\n");
        printf("allocating memory");
        fflush(stdout);
    }
    if ((Bd = (double *)malloc((size_t)M * sizeof(double))) == NULL)
        allocfailed("Bd[]");
    if ((SortW = (double *)malloc((size_t)M * sizeof(double))) == NULL)
        allocfailed("SortW[]");
    if ((WgtDelta = (double *)malloc((size_t)M * sizeof(double))) == NULL)
        allocfailed("WgtDelta[]");
    for (n=0; n<NumCov; n++)
        if (Stats[n].Used == TRUE && n != ORIENT_) {
            if ((Stats[n].Wd = (double **)malloc((size_t)V * sizeof(double *))) == NULL)
                allocfailed("Stats[].Wd[]");
            for (c=0; c<V; c++)
                if ((Stats[n].Wd[c] = (double *)malloc((size_t)M * sizeof(double))) == NULL)
                    allocfailed("Stats[].Wd[][]");
        }
    for (n=0; n<NumCov; n++)
        if (Stats[n].Used) {
            Stats[n].Cinv = new_matrixE(M, M, "inverse covariance matrix");
        }

    // read covariance matrices
    if (vflg) {
        printf(" - done\n");
        printf("reading covariance files:\n");
    }

    for (n=0; n<NumCov; n++)
        if (Stats[n].Used) {

            // read covariance matrix
            if (vflg) {
                printf("\t'%s'", Stats[n].Name);
                fflush(stdout);
            }
            GetCov(Stats[n].CovFileName, &Stats[n].WgtCovHdr, &ChanIndex, &Stats[n].Cov);

            // read noise files
            sprintf(fpath, "%s/%s_Noise", CovDirName, Stats[n].Name);
            if ((np = fopen(fpath, "r")) == NULL)
                Cleanup("can't open noise file");
            if (fscanf(np, "%le", &Stats[n].Noise) != 1)
                Cleanup("can't read noise file");
            fclose(np);

            // if req'd, regularize covariance matrix
            switch (Params.Operator) {
                case ADD_MU:    // add scaled noise power to diagonal of covariance matrix
                    tmp = 1.0e-30 * Params.Mu * Params.Mu;      // convert fT/root-Hz to T^2/Hz
                    Mu = Stats[n].WgtCovHdr.BWFreq * tmp;       // multiply by covariance bandwidth to get T^2
                    for (m=0; m<M; m++)
                        Stats[n].Cov[m][m] += Mu;
                    Stats[n].Noise += Mu;
                    break;
                case MPY_MU:    // multiply covariance diagonal by mu
                    for (m=0; m<M; m++)
                        Stats[n].Cov[m][m] *= Params.Mu;
                    Stats[n].Noise *= Params.Mu;
                    break;
                default:        // no regularizarion -- do nothing
                    break;
            }

            if (Params.Mu >= 0. && Params.PropMu >= 0.)
                Cleanup("Illegal specification of both Mu and PropMu");

            if (Params.PropMu >= 0.) {
                Mu = Stats[n].Noise * Params.PropMu;
                for (m=0; m<M; m++)
                    Stats[n].Cov[m][m] += Mu;
                Stats[n].Noise += Mu;
            }

            if (vflg) {
                printf(" - done\n");
            }

            // write new noise files after regularization
            if (Params.Operator != -1 || Params.PropMu >= 0.) {
                sprintf(fpath, "%s/%s_MuNoise", CovDirName, Stats[n].Name);
                if ((np = fopen(fpath, "w")) == NULL)
                    Cleanup("can't write noise file");
                fprintf(np, "%e\n", Stats[n].Noise);
                fclose(np);
            }
        }

    if (wflg && vflg) {     // normalize
        printf("normalizing weights by sqrt(regularized noise power)\n");
        for (n=0; n<NumCov; n++) {
            if (Stats[n].Used) {
                printf("normalizing %s weights by %g\n", Stats[n].Name, sqrt(Stats[n].Noise));
            }
        }
    }

    // announce regularization used
    if (Params.Operator != -1 || Params.PropMu >= 0.)
        if (vflg) {
            printf("regularizing covariance matrices by ");
            if (Params.Operator != -1) {
                switch (Params.Operator) {
                    case ADD_MU:
                        printf("adding %6.3f fT/root-Hz to diagonal\n", Params.Mu);
                        break;
                    case MPY_MU:
                        printf("multiplying diagonal by %6.3f\n", Params.Mu);
                        break;
                    default:
                        break;
                }
            } else {
                printf("adding %6.3f times the noise to diagonal\n", Params.PropMu);
            }
        }

    // Invert covariance matrices.

    for (n=0; n<NumCov; n++) {
        if (Stats[n].Used) {
            for (i=0; i<M; i++)
                for (j=0; j<M; j++)
                    Stats[n].Cinv[i][j] = Stats[n].Cov[i][j];
            if (pflg) {
                pinv(Stats[n].Cinv, M, ndim);
            } else {
                invert(Stats[n].Cinv, M);
            }
        }
    }

    // create shape file (used for checking that subject's head is fully inside helmet)
    if (vflg) {
        printf("saving '%s.shape'", Header.SetName);
        fflush(stdout);
    }
    sprintf(fpath, "%s/%s.shape", DSpath, Header.SetName);
    if ((fp = fopen(fpath, "w")) == NULL)
        Cleanup("can't open 'shape' file for write");
    fprintf(fp, "%d\n", M);
    for (m=0; m<M; m++) {
        j = ChanIndex[m];
        fprintf(fp, "%7.3f\t%7.3f\t%7.3f\n",
          100. * Channel[j].Geom.MEGSensor.origin.p[X_], 100. * Channel[j].Geom.MEGSensor.origin.p[Y_], 100. * Channel[j].Geom.MEGSensor.origin.p[Z_]);
    }
    fclose(fp);

    // if writing legacy weights, fill in SAM weight header
    if (tflg) {
        if (vflg) {
            printf(" - done\n");
            printf("populating weight headers");
            fflush(stdout);
        }
        for (n=0; n<NumCov; n++)
            if (Stats[n].Used) {
                Stats[n].SAMHeader.Version = SAM_REV;                   // file version number
                sprintf(Stats[n].SAMHeader.SetName, "%s", DSName);      // name of parent dataset
                Stats[n].SAMHeader.NumChans = M;                        // number of channels used by SAM
                Stats[n].SAMHeader.HPFreq = Stats[n].WgtCovHdr.HPFreq;  // highpass frequency (Hz)
                Stats[n].SAMHeader.LPFreq = Stats[n].WgtCovHdr.LPFreq;  // lowpass frequency (Hz)
                Stats[n].SAMHeader.BWFreq = Stats[n].WgtCovHdr.BWFreq;  // bandwidth of filters (Hz)
                Stats[n].SAMHeader.MeanNoise = Stats[n].Noise;          // mean primary sensor noise (T^2)
                sprintf(Stats[n].SAMHeader.MriName, "NONE");            // a place-holder until we decide whether to read the default.hdm file
                for (v=S_; v<=A_; v++) {
                    Stats[n].SAMHeader.Nasion[v] = 0;
                    Stats[n].SAMHeader.RightPA[v] = 0;
                    Stats[n].SAMHeader.LeftPA[v] = 0;
                }
                Stats[n].SAMHeader.XStart = 0.;                         // x-start coordinate (m)
                Stats[n].SAMHeader.XEnd = 0.;                           // x-end coordinate (m)
                Stats[n].SAMHeader.YStart = 0.;                         // y-start coordinate (m)
                Stats[n].SAMHeader.YEnd = 0.;                           // y-end coordinate (m)
                Stats[n].SAMHeader.ZStart = 0.;                         // z-start coordinate (m)
                Stats[n].SAMHeader.ZEnd = 0.;                           // z-end coordinate (m)
                Stats[n].SAMHeader.StepSize = 0.;                       // voxel step size (m)
                Stats[n].SAMHeader.NumWeights = V;                      // number of SAM virtual sensors
                Stats[n].SAMHeader.SAMType = SAM_TYPE_WT_LIST;
                Stats[n].SAMHeader.SAMUnit = (wflg == TRUE)? SAM_UNIT_NORM: SAM_UNIT_COEFF; // redundant, but -- so what?
            }
    }

    // if req'd, open SAM forward solution files & write header
    if (bflg) {
        if (vflg) {
            printf(" - done\n");
            printf("opening forward solution file for write");
            fflush(stdout);
        }
        sprintf(fpath, "%s/%s.fwd", SAMpath, Params.TargetName);
        bp = fileopen(fpath, "w");
        if (fwrite((void *)FwdID, 8, 1, bp) != 1)
            Cleanup("can't write ID to forward solution file");
        if (fwrite((void *)&Stats[GLOBAL_].SAMHeader, sizeof(SAM_HDR), 1, bp) != 1)
            Cleanup("can't write header to forward solution file");
        if (fwrite((void *)ChanIndex, sizeof(int), M, bp) != M)
            Cleanup("can't write channel index to forward solution file");
    }

    // compute beamformer coefficients for each marker
    if (vflg) {
        printf(" - done\n");
        printf("computing beamformer coefficients for:\n");
    }
    for (n=0; n<NumCov; n++) {      // loop on each valid marker

        if (Stats[n].Used && n != ORIENT_) {

            if (vflg)
                printf("\t%s:\n", Stats[n].Name);

            // loop on voxels
            for (c=0, pct=(-1); c<V; c++) {

                // estimate SAM voxel only if ROI flag is true
                if (Voxel[c].ROI) {

                    // set voxel coordinate & solve for optimal beamformer orientation using orientation covariance
                    if (!iflg) {

                        // no interpolation
                        switch (Params.Model) {
                            case SSPHERE:
                                DIPSolve2D(&Voxel[c], Bd, Stats[ORIENT_].Cinv, &Header, Channel, Order, &span);
                                break;
                            case MSPHERE:
                                DIPSolve3D(&Voxel[c], Bd, Stats[ORIENT_].Cinv, &Header, Channel, Order, &span);
                                break;
                            case NOLTE:
                                ECDSolve(&Voxel[c], Bd, Stats[ORIENT_].Cinv, &Header, Channel, &Hull, &span);
                                break;
                            default:
                                break;
                        }

                        // solve for weights
                        SAMsolve2(Stats[n].Cov, Bd, Stats[n].Wd[c], Stats[n].Noise, M, &Src, &Nse);
                        if (wflg) {             // noise normalized weights
                            tmp = sqrt(Stats[n].Noise);
                            for (m=0; m<M; m++)
                                Stats[n].Wd[c][m] /= tmp;
                        }

                    } else {    // else interpolate

                        // copy orientation if using Atlas & set flags
                        if (aflg) {
                            for (v=X_; v<=Z_; v++)
                                VoxDelta.v[v] = Voxel[c].v[v];
                            VoxDelta.Solve = FALSE;
                        } else {
                            VoxDelta.Solve = TRUE;
                        }
                        VoxDelta.ROI = TRUE;

                        // zero weight vector
                        for (m=0; m<M; m++)
                            Stats[n].Wd[c][m] = 0.;

                        // loop on interpolations
                        for (ix=(-1); ix<=1; ix++) {
                            VoxDelta.p[X_] = Voxel[c].p[X_] + (double)ix * Params.Interpolate;
                            for (iy=(-1); iy<=1; iy++) {
                                VoxDelta.p[Y_] = Voxel[c].p[Y_] + (double)iy * Params.Interpolate;
                                for (iz=(-1); iz<=1; iz++) {
                                    VoxDelta.p[Z_] = Voxel[c].p[Z_] + (double)iz * Params.Interpolate;

                                    switch (Params.Model) {
                                        case SSPHERE:
                                            DIPSolve2D(&VoxDelta, Bd, Stats[ORIENT_].Cov, &Header, Channel, Order, &span);
                                            break;
                                        case MSPHERE:
                                            DIPSolve3D(&VoxDelta, Bd, Stats[ORIENT_].Cov, &Header, Channel, Order, &span);
                                            break;
                                        case NOLTE:
                                            ECDSolve(&VoxDelta, Bd, Stats[ORIENT_].Cov, &Header, Channel, &Hull, &span);
                                            break;
                                        default:
                                            break;
                                    }

                                    // solve for weights
                                    SAMsolve2(Stats[n].Cov, Bd, WgtDelta, Stats[n].Noise, M, &Src, &Nse);
                                    if (wflg) {         // noise normalized weights
                                        tmp = sqrt(Stats[n].Noise);
                                        for (m=0; m<M; m++)
                                            WgtDelta[m] /= tmp;
                                    }

                                    // sum weights
                                    for (m=0; m<M; m++)
                                        Stats[n].Wd[c][m] += WgtDelta[m];

                                }   // end iz loop
                            }       // end iy loop
                        }           // end ix loop

                        // average weights
                        for (m=0; m<M; m++)
                            Stats[n].Wd[c][m] /= 27.;

                    }   // end iflg

                // if outside bounds
                } else {
                    for (m=0; m<M; m++) {       // zero forward solution
                        Bd[m] = 0.;
                        Stats[n].Wd[c][m] = 0.;
                    }
                }

                // if req'd, write weight (or forward solution) vector
                if (bflg) {
                    if (fwrite((void *)Bd, sizeof(double), M, bp) != M)
                        Cleanup("can't write weight to file");
                }

                // percent done
                if (vflg) {
                    percent = 100. * (double)c / (double)V;
                    if (rint(percent) > (double)pct) {
                        pct = (int)rint(percent);
                        printf("\r\t\t%d percent done", pct);
                        fflush(stdout);
                    }
                }
            }       // end c-loop
            if (vflg) {
                printf("\n");
            }
        }           // end if stats TRUE
    }

    // finit
    if (!bflg) {
        if (!tflg) {            // write NIFTI weights for ROI
            if (vflg) {
                printf("writing NIFTI weights for ROI");
                fflush(stdout);
            }
            for (n=0; n<NumCov; n++)
                if (Stats[n].Used && n != ORIENT_)
                    PutNIFTIWts(Stats[n].WgtFileName, &Params, Stats[n].Wd, Stats[n].Noise, M); // write the beamformer coefficients to the NIFTI file
        } else {                // else write NIFTI + legacy weights for target
            if (vflg) {
                printf("writing NIFTI weights for target list");
                fflush(stdout);
            }
            for (n=0; n<NumCov; n++)
                if (Stats[n].Used && n != ORIENT_)
                    PutNIFTITargets(Stats[n].WgtFileName, Stats[n].Wd, Stats[n].Noise, M, V);

            // write legacy weight file to SAM subdirectory
            if (vflg) {
                printf(" - done\n");
                printf("writing legacy weights");
                fflush(stdout);
            }
            for (n=0; n<NumCov; n++) {
                if (Stats[n].Used && n != ORIENT_) {
                    if (strncmp(Params.Target, "NULL", 4)) {
                        sprintf(fpath, "%s/%s,%s,%s.wts", SAMpath, ParmName, Params.TargetName, Stats[n].Name);
                    } else {
                        sprintf(fpath, "%s/%s,%s.wts", SAMpath, ParmName, Stats[n].Name);
                    }
                    for (c=0; c<V; c++) {
                        PutWts(fpath, Stats[n].Wd[c], Header, Stats[n].WgtCovHdr, V);
                    }
                    PutWts(NULL, Stats[0].Wd[c], Header, Stats[0].WgtCovHdr, 0);
                }
            }
        }
        if (vflg) {
            printf(" - done\n");
        }
    } else {
        fclose(bp);
    }

    // if TLRC, execute system call for @auto_tlrc
    if (!tflg && Params.ImageFormat == TLRC) {
        if (vflg) {
            printf("computing Talairach transform of weights at %4.1f mm resolution:\n", Params.ImageRes);
        }
        if (getcwd(CurrentDir, sizeof(CurrentDir)) == NULL)
            Cleanup("can't get current working directory");
        if (chdir(CovDirName) == -1)
            Cleanup("can't change to covariance directory");
        for (n=0; n<NumCov; n++) {
            if (Stats[n].Used && n != ORIENT_) {    // note that the verbose output of @auto_tlrc is now eliminated
                if (vflg) {
                    printf("\t%s", Stats[n].Name);
                    fflush(stdout);
                }
                sprintf(ByYourCommand, "@auto_tlrc -rmode NN -dxyz %4.1f -apar %s -input %s -overwrite > /dev/null 2>&1", Params.ImageRes, TlrcPath, Stats[n].NiiName);
                if (system(ByYourCommand) == -1)
                    Cleanup("system call to '@auto_tlrc' failed");
                if (vflg) {
                    printf(" - done\n");
                }
            }
        }
        if (chdir(CurrentDir) == -1)
            Cleanup("can't change back to current directory");
    }

    // all done!
    if (vflg) {
        printf("'%s' done\n", Progname);
    }
    exit(0);
}

/* Convert ROI bounds so that there are always an integral number of steps
and the computed ROI encloses the given one. This way the voxel boundaries
line up exactly with the coordinate axes. (The origin is at the corner of
a voxel.) lowhigh is -1 or 1 for low or high bounds, resp. */

static double fixBound(double b, double step, int lowhigh)
{
    double t, r;

    if (lowhigh < 0) {
        /* lower bound */
        t = floor(b / step);
        r = (t + .5) * step;
        if (r > b) {
            r = (t - .5) * step;
        }
    } else {
        /* higher bound */
        t = ceil(b / step);
        r = (t - .5) * step;
        if (r < b) {
            r = (t + .5) * step;
        }
    }
    return r;
}

static void fixBounds(PARMINFO *p)
{
    int i, j;
    double tmp;

    j = 0;
    for (i = 0; i < 3; i++) {
        tmp = fixBound(p->SAMStart[i], p->SAMStep, -1);
        if (!j && fabs(tmp - p->SAMStart[i]) > ETA) {
            j = 1;
        }
        p->SAMStart[i] = tmp;
        tmp = fixBound(p->SAMEnd[i], p->SAMStep, 1);
        if (!j && fabs(tmp - p->SAMEnd[i]) > ETA) {
            j = 1;
        }
        p->SAMEnd[i] = tmp;
    }

    if (vflg) {
        if (j) {
            printf("\nROI bounds adjusted to accommodate step size of %g\n\n", p->SAMStep * 100.);
            printf("XBounds %g %g\n", p->SAMStart[0] * 100., p->SAMEnd[0] * 100.);
            printf("YBounds %g %g\n", p->SAMStart[1] * 100., p->SAMEnd[1] * 100.);
            printf("ZBounds %g %g\n\n", p->SAMStart[2] * 100., p->SAMEnd[2] * 100.);
        }
    }
}
